/*
 * Copyright (c) 2022 NetEase, Inc. All rights reserved.
 * Use of this source code is governed by a MIT license that can be
 * found in the LICENSE file.
 */

package com.netease.yunxin.flutter.plugins.roomkit

import com.netease.yunxin.flutter.plugins.roomkit.pigeon.Pigeon
import com.netease.yunxin.kit.roomkit.api.NERoleParams
import com.netease.yunxin.kit.roomkit.api.NERoomChatCustomMessage
import com.netease.yunxin.kit.roomkit.api.NERoomChatFileMessage
import com.netease.yunxin.kit.roomkit.api.NERoomChatImageMessage
import com.netease.yunxin.kit.roomkit.api.NERoomChatMessage
import com.netease.yunxin.kit.roomkit.api.NERoomChatTextMessage
import com.netease.yunxin.kit.roomkit.api.NERoomEndReason
import com.netease.yunxin.kit.roomkit.api.NERoomMember
import com.netease.yunxin.kit.roomkit.api.NERoomRole

fun NERoomRole.mapToRoomRole() =
    Pigeon.RoomRole.Builder()
        .setLimit(limit.toLong())
        .setName(name)
        .setParams(params?.mapToRoleParams())
        .build()

fun NERoleParams.mapToRoleParams() =
    Pigeon.RoleParams.Builder()
        .setAudio(
            if (audio != null) {
                Pigeon.RoleAudioParams.Builder()
                    .setProfile(audio!!.profile)
                    .setScenario(audio!!.scenario)
                    .build()
            } else {
                null
            }
        )
        .setVideo(
            if (video != null) {
                Pigeon.RoleVideoParams.Builder()
                    .setWidth(video!!.width.toLong())
                    .setHeight(video!!.height.toLong())
                    .setFps(video!!.fps.toLong())
                    .build()
            } else {
                null
            }
        )
        .setScreen(
            if (screen != null) {
                Pigeon.RoleScreenParams.Builder()
                    .setVideoProfile(screen!!.videoProfile)
                    .build()
            } else {
                null
            }
        )
        .build()

fun NERoomMember.mapToRoomMember() =
    Pigeon.RoomMember.Builder()
        .setUuid(uuid)
        .setName(name)
        .setAvatar(avatar)
        .setRole(role.mapToRoomRole())
        .setIsAudioOn(isAudioOn)
        .setIsVideoOn(isVideoOn)
        .setIsInChatroom(isInChatroom)
        .setIsInRtcChannel(isInRtcChannel)
        .setIsSharingScreen(isSharingScreen)
        .setIsSharingWhiteboard(isSharingWhiteboard)
        .setProperties(properties)
        .setClientType(clientType.name)
        .build()

fun stringifyRoomEndReason(reason: NERoomEndReason) =
    when (reason) {
        NERoomEndReason.LEAVE_BY_SELF -> "leaveBySelf"
        NERoomEndReason.SYNC_DATA_ERROR -> "syncDataError"
        NERoomEndReason.KICK_OUT -> "kickOut"
        NERoomEndReason.KICK_BY_SELF -> "selfKick"
        NERoomEndReason.CLOSE_BY_MEMBER -> "closeByMember"
        NERoomEndReason.END_OF_LIFE -> "endOfLife"
        NERoomEndReason.ALL_MEMBERS_OUT -> "allMemberOut"
        NERoomEndReason.CLOSE_BY_BACKEND -> "closeByBackend"
        NERoomEndReason.LOGIN_STATE_ERROR -> "loginStateError"
        NERoomEndReason.END_OF_RTC -> "endOfRtc"
        NERoomEndReason.UNKNOWN -> "unknown"
    }

fun NERoomChatMessage.mapToRoomMessage(): Pigeon.RoomMessage {
    val builder = Pigeon.RoomMessage.Builder()
        .setUuid(messageUuid)
        .setType(messageType.name)
        .setToAccounts(toUserUuidList)
        .setFromAccount(fromUserUuid)
        .setFromNick(fromNick)
        .setTime(time)
    when (this) {
        is NERoomChatTextMessage -> {
            builder.setType("text").setContent(text)
        }
        is NERoomChatCustomMessage -> {
            builder.setType("custom").setContent(attachStr)
        }
        is NERoomChatImageMessage -> {
            builder.setType("image").setAttachment(buildMessageAttachment(this))
        }
        is NERoomChatFileMessage -> {
            builder.setType("file").setAttachment(buildMessageAttachment(this))
        }
    }
    return builder.build()
}

fun buildMessageAttachment(msg: NERoomChatFileMessage): Pigeon.RoomMessageAttachment {
    val attachmentBuilder = Pigeon.RoomMessageAttachment.Builder()
        .setDisplayName(msg.displayName)
        .setExtension(msg.extension)
        .setPath(msg.path)
        .setThumbPath(msg.thumbPath)
        .setMd5(msg.md5)
        .setUrl(msg.url)
        .setSize(msg.size)
    if (msg is NERoomChatImageMessage) {
        attachmentBuilder
            .setWidth(msg.width.toLong())
            .setHeight(msg.height.toLong())
    }
    return attachmentBuilder.build()
}
