/*
 * Copyright (c) 2022 NetEase, Inc. All rights reserved.
 * Use of this source code is governed by a MIT license that can be
 * found in the LICENSE file.
 */

package com.netease.yunxin.flutter.plugins.roomkit

import android.app.Activity
import android.content.Context
import android.content.Intent
import android.media.projection.MediaProjection
import android.media.projection.MediaProjectionManager
import android.net.Uri
import android.opengl.EGLContext
import android.text.TextUtils
import android.util.Log
import com.netease.lava.webrtc.EglBase
import com.netease.lava.webrtc.EglBase10
import com.netease.lava.webrtc.EglBase14
import com.netease.yunxin.flutter.plugins.roomkit.pigeon.Pigeon
import com.netease.yunxin.flutter.plugins.roomkit.rtc.FlutterVideoRenderer
import com.netease.yunxin.flutter.plugins.roomkit.whiteboard.WhiteboardViewFactory
import com.netease.yunxin.kit.roomkit.api.NECallback
import com.netease.yunxin.kit.roomkit.api.NEChatroomMemberQueryType
import com.netease.yunxin.kit.roomkit.api.NEErrorCode
import com.netease.yunxin.kit.roomkit.api.NEPreviewRoomContext
import com.netease.yunxin.kit.roomkit.api.NEPreviewRoomListenerAdapter
import com.netease.yunxin.kit.roomkit.api.NERoomBeautyEffectType
import com.netease.yunxin.kit.roomkit.api.NERoomChatMessage
import com.netease.yunxin.kit.roomkit.api.NERoomContext
import com.netease.yunxin.kit.roomkit.api.NERoomEndReason
import com.netease.yunxin.kit.roomkit.api.NERoomKit
import com.netease.yunxin.kit.roomkit.api.NERoomKitOptions
import com.netease.yunxin.kit.roomkit.api.NERoomLanguage
import com.netease.yunxin.kit.roomkit.api.NERoomListenerAdapter
import com.netease.yunxin.kit.roomkit.api.NERoomLiveConfig
import com.netease.yunxin.kit.roomkit.api.NERoomLiveInfo
import com.netease.yunxin.kit.roomkit.api.NERoomLiveRequest
import com.netease.yunxin.kit.roomkit.api.NERoomLiveState
import com.netease.yunxin.kit.roomkit.api.NERoomLiveStreamAudioCodecProfile
import com.netease.yunxin.kit.roomkit.api.NERoomLiveStreamAudioSampleRate
import com.netease.yunxin.kit.roomkit.api.NERoomLiveStreamImageInfo
import com.netease.yunxin.kit.roomkit.api.NERoomLiveStreamLayout
import com.netease.yunxin.kit.roomkit.api.NERoomLiveStreamMode
import com.netease.yunxin.kit.roomkit.api.NERoomLiveStreamTaskInfo
import com.netease.yunxin.kit.roomkit.api.NERoomLiveStreamUserTranscoding
import com.netease.yunxin.kit.roomkit.api.NERoomLiveStreamVideoScaleMode
import com.netease.yunxin.kit.roomkit.api.NERoomMember
import com.netease.yunxin.kit.roomkit.api.NERoomRole
import com.netease.yunxin.kit.roomkit.api.NERoomRtcStatsListenerAdapter
import com.netease.yunxin.kit.roomkit.api.NEValueCallback
import com.netease.yunxin.kit.roomkit.api.mapToLiveLayout
import com.netease.yunxin.kit.roomkit.api.model.NEAudioDumpType
import com.netease.yunxin.kit.roomkit.api.model.NEAudioOutputDevice
import com.netease.yunxin.kit.roomkit.api.model.NEChatroomMember
import com.netease.yunxin.kit.roomkit.api.model.NEMemberVolumeInfo
import com.netease.yunxin.kit.roomkit.api.model.NERoomCreateAudioEffectOption
import com.netease.yunxin.kit.roomkit.api.model.NERoomCreateAudioMixingOption
import com.netease.yunxin.kit.roomkit.api.model.NERoomRtcAudioStreamType
import com.netease.yunxin.kit.roomkit.api.model.NERoomRtcNetworkQualityInfo
import com.netease.yunxin.kit.roomkit.api.model.NERoomRtcStats
import com.netease.yunxin.kit.roomkit.api.model.NERoomVirtualBackgroundSource
import com.netease.yunxin.kit.roomkit.api.model.NEServerConfig
import com.netease.yunxin.kit.roomkit.api.model.NEVideoStreamType
import com.netease.yunxin.kit.roomkit.api.service.NEAuthEvent
import com.netease.yunxin.kit.roomkit.api.service.NEAuthListener
import com.netease.yunxin.kit.roomkit.api.service.NEAuthService
import com.netease.yunxin.kit.roomkit.api.service.NECreateRoomOptions
import com.netease.yunxin.kit.roomkit.api.service.NECreateRoomParams
import com.netease.yunxin.kit.roomkit.api.service.NECrossAppAuthorization
import com.netease.yunxin.kit.roomkit.api.service.NECustomMessage
import com.netease.yunxin.kit.roomkit.api.service.NEJoinRoomOptions
import com.netease.yunxin.kit.roomkit.api.service.NEJoinRoomParams
import com.netease.yunxin.kit.roomkit.api.service.NEMessageChannelListener
import com.netease.yunxin.kit.roomkit.api.service.NEMessageChannelService
import com.netease.yunxin.kit.roomkit.api.service.NEPreviewRoomOptions
import com.netease.yunxin.kit.roomkit.api.service.NEPreviewRoomParams
import com.netease.yunxin.kit.roomkit.api.service.NERoomService
import com.netease.yunxin.kit.roomkit.api.service.NESeatEventListener
import com.netease.yunxin.kit.roomkit.api.service.NESeatInfo
import com.netease.yunxin.kit.roomkit.api.service.NESeatInitParams
import com.netease.yunxin.kit.roomkit.api.service.NESeatInvitationItem
import com.netease.yunxin.kit.roomkit.api.service.NESeatItem
import com.netease.yunxin.kit.roomkit.api.service.NESeatRequestItem
import io.flutter.embedding.engine.plugins.FlutterPlugin
import io.flutter.plugin.common.BinaryMessenger
import io.flutter.plugin.common.PluginRegistry
import io.flutter.view.TextureRegistry
import java.io.File

class FlutterResultCallback<U, T>(
    private val flutterResult: Pigeon.Result<U>,
    private val mapper: (code: Int, message: String?, data: T?) -> U
) : NECallback<T> {

    override fun onResult(code: Int, message: String?, data: T?) {
        runCatching { mapper(code, message, data) }
            .onSuccess { value -> flutterResult.success(value) }
            .onFailure { throwable -> flutterResult.error(throwable) }
    }
}

fun Pigeon.Result<Pigeon.SimpleResponse>.simpleCallback(onSuccess: (() -> Unit)? = null) =
    FlutterResultCallback<Pigeon.SimpleResponse, Unit>(this) { code: Int, message: String?, _: Unit? ->
        if (code == NEErrorCode.SUCCESS) {
            onSuccess?.invoke()
        }
        Pigeon.SimpleResponse.Builder()
            .setCode(code.toLong())
            .setMsg(message)
            .build()
    }

val roomContextNotFound =
    Pigeon.SimpleResponse.Builder()
        .setCode(-1)
        .setMsg("Room context not found")
        .build()

fun consequence(code: Int = NEErrorCode.SUCCESS, msg: String? = null) = Pigeon.SimpleResponse.Builder()
    .setCode(code.toLong())
    .setMsg(msg)
    .build()

interface IPlatform {

    fun onAttachedToEngine(binding: FlutterPlugin.FlutterPluginBinding)

    fun onDetachedFromEngine(binding: FlutterPlugin.FlutterPluginBinding)
}

class RoomKitPlatform : Pigeon.RoomKitApi, IPlatform {

    private lateinit var applicationContext: Context
    override fun onAttachedToEngine(binding: FlutterPlugin.FlutterPluginBinding) {
        applicationContext = binding.applicationContext
        Pigeon.RoomKitApi.setup(binding.binaryMessenger, this)
    }

    override fun onDetachedFromEngine(binding: FlutterPlugin.FlutterPluginBinding) {
        Pigeon.RoomKitApi.setup(binding.binaryMessenger, null)
    }

    override fun initialize(
        options: Pigeon.InitializeRequest,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        NERoomKit.instance.initialize(
            applicationContext,
            NERoomKitOptions(
                appKey = options.appKey,
                // reuseIM = options.reuseIM,
                useAssetServerConfig = options.useAssetServerConfig,
                serverConfig = NEServerConfig.fromJson(options.serverConfigJson),
                serverUrl = options.serverUrl,
                extras = options.extras ?: emptyMap()
            ),
            result.simpleCallback()
        )
    }

    override fun switchLanguage(languageCode: String): Long {
        val lang = when (languageCode) {
            "zh" -> NERoomLanguage.CHINESE
            "en" -> NERoomLanguage.ENGLISH
            "ja" -> NERoomLanguage.JAPANESE
            else -> NERoomLanguage.AUTOMATIC
        }
        return NERoomKit.instance.switchLanguage(lang).toLong()
    }

    override fun sdkVersions(): Pigeon.SDKVersions {
        return NERoomKit.instance.sdkVersions.run {
            Pigeon.SDKVersions.Builder()
                .setImVersion(imVersion)
                .setRtcVersion(rtcVersion)
                .setWhiteboardVersion(whiteboardVersion)
                .setRoomKitVersion(roomKitVersion)
                .build()
        }
    }
}

class AuthServicePlatform : Pigeon.AuthServiceApi, IPlatform {

    private val authService by lazy {
        NERoomKit.instance.getService(NEAuthService::class.java)
    }

    private var authStatusEventSink: Pigeon.AuthStatusEventSink? = null

    private val authEventListener by lazy {
        object : NEAuthListener {
            override fun onAuthEvent(evt: NEAuthEvent) {
                authStatusEventSink?.onAuthStatusChanged(
                    Pigeon.AuthStatusEvent.Builder().setStatus(
                        when (evt) {
                            NEAuthEvent.KICK_OUT -> "KickOut"
                            NEAuthEvent.TOKEN_EXPIRED -> "TokenExpired"
                            NEAuthEvent.INCORRECT_TOKEN -> "IncorrectToken"
                            NEAuthEvent.FORBIDDEN -> "Forbidden"
                            NEAuthEvent.ACCOUNT_TOKEN_ERROR -> "AccountTokenError"
                            NEAuthEvent.LOGGED_IN -> "LoggedIn"
                            NEAuthEvent.LOGGED_OUT -> "LoggedOut"
                        }
                    ).setIsLoggedIn(authService.isLoggedIn).build()
                ) {}
            }
        }
    }

    override fun onAttachedToEngine(binding: FlutterPlugin.FlutterPluginBinding) {
        Pigeon.AuthServiceApi.setup(binding.binaryMessenger, this)
        authStatusEventSink = Pigeon.AuthStatusEventSink(binding.binaryMessenger)
        authService.addAuthListener(authEventListener)
    }

    override fun onDetachedFromEngine(binding: FlutterPlugin.FlutterPluginBinding) {
        Pigeon.AuthServiceApi.setup(binding.binaryMessenger, null)
        authStatusEventSink = null
        authService.removeAuthListener(authEventListener)
    }

    override fun login(
        account: String,
        token: String,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        authService.login(account, token, result.simpleCallback())
    }

    override fun logout(result: Pigeon.Result<Pigeon.SimpleResponse>) {
        authService.logout(result.simpleCallback())
    }

    override fun isLoggedIn(): Boolean {
        return authService.isLoggedIn
    }
}

class MessageServicePlatform : Pigeon.MessageServiceApi, IPlatform {

    private var messageEventSink: Pigeon.MessageEventSink? = null

    private val messageService by lazy {
        NERoomKit.instance.getService(NEMessageChannelService::class.java)
    }

    // IM自定义透传消息监听
    private val messageListener by lazy {
        object : NEMessageChannelListener {
            override fun onReceiveCustomMessage(message: NECustomMessage) {
                messageEventSink?.onReceivePassThroughMessage(
                    Pigeon.PassThroughMessageEvent.Builder()
                        .setRoomUuid(message.roomUuid)
                        .setSenderUuid(message.senderUuid)
                        .setCommandId(message.commandId.toLong())
                        .setData(message.data)
                        .build()
                ) {}
            }
        }
    }

    override fun onAttachedToEngine(binding: FlutterPlugin.FlutterPluginBinding) {
        Pigeon.MessageServiceApi.setup(binding.binaryMessenger, this)
        messageEventSink = Pigeon.MessageEventSink(binding.binaryMessenger)
        messageService.addMessageChannelListener(messageListener)
    }

    override fun onDetachedFromEngine(binding: FlutterPlugin.FlutterPluginBinding) {
        messageEventSink = null
        Pigeon.MessageServiceApi.setup(binding.binaryMessenger, null)
        messageService.removeMessageChannelListener(messageListener)
    }

    override fun sendCustomMessage(
        roomUuid: String,
        userUuid: String,
        commandId: Long,
        data: String,
        crossAppAuthorization: Pigeon.CrossAppAuthorization?,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        val authorization = crossAppAuthorization?.let { NECrossAppAuthorization(it.appKey, it.user, it.token) }
        messageService.sendCustomMessage(roomUuid, userUuid, commandId.toInt(), data, authorization, result.simpleCallback())
    }
}

class RoomServicePlatform : Pigeon.RoomServiceApi, IPlatform {

    private val roomService by lazy {
        NERoomKit.instance.getService(NERoomService::class.java)
    }

    private var roomEventSink: Pigeon.RoomEventSink? = null
    private var roomRtcStatsSink: Pigeon.RoomStatsSink? = null
    private var previewRoomEventSink: Pigeon.PreviewRoomEventSink? = null
    private var seatEventSink: Pigeon.SeatEventSink? = null

    override fun onAttachedToEngine(binding: FlutterPlugin.FlutterPluginBinding) {
        Pigeon.RoomServiceApi.setup(binding.binaryMessenger, this)
        roomEventSink = Pigeon.RoomEventSink(binding.binaryMessenger)
        roomRtcStatsSink = Pigeon.RoomStatsSink(binding.binaryMessenger)
        previewRoomEventSink = Pigeon.PreviewRoomEventSink(binding.binaryMessenger)
        seatEventSink = Pigeon.SeatEventSink(binding.binaryMessenger)
    }

    override fun onDetachedFromEngine(binding: FlutterPlugin.FlutterPluginBinding) {
        Pigeon.RoomServiceApi.setup(binding.binaryMessenger, null)
        roomEventSink = null
        roomRtcStatsSink = null
        previewRoomEventSink = null
        seatEventSink = null
    }

    override fun createRoom(
        params: Pigeon.CreateRoomParams,
        options: Pigeon.CreateRoomOptions,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        var seatInitParams: NESeatInitParams? = null
        if (params.seatInitParams != null) {
            val seatInitParams1 = params.seatInitParams as Pigeon.SeatInitParams
            seatInitParams = NESeatInitParams(
                seatCount = seatInitParams1.seatCount.toInt(),
                seatRequestApprovalMode = seatInitParams1.seatRequestApprovalMode.toInt(),
                seatInvitationConfirmMode = seatInitParams1.seatInvitationConfirmMode.toInt()
            )
        }
        roomService.createRoom(
            params = NECreateRoomParams(
                roomUuid = params.roomUuid,
                roomName = params.roomName,
                templateId = params.templateId.toInt(),
                password = params.password,
                seatInitParams = seatInitParams,
                initialProperties = params.initialProperties ?: emptyMap()
            ),
            options = NECreateRoomOptions(
                enableRecord = options.enableRecord,
                enableLive = options.enableLive,
                enableWhiteboard = options.enableWhiteboard,
                enableChatroom = options.enableChatroom,
                enableRtc = options.enableRtc
//                enableSip = options.enableSip,
            ),
            result.simpleCallback()
        )
    }

    override fun joinRoom(
        params: Pigeon.JoinRoomParams,
        options: Pigeon.JoinRoomOptions,
        result: Pigeon.Result<Pigeon.JoinRoomResponse>
    ) {
        roomService.joinRoom(
            params = NEJoinRoomParams(
                roomUuid = params.roomUuid,
                userName = params.userName,
                avatar = params.avatar,
                role = params.role,
                password = params.password,
                crossAppAuthorization = params.crossAppAuthorization?.let { NECrossAppAuthorization(it.appKey, it.user, it.token) },
                initialProperties = params.initialProperties ?: emptyMap()
            ),
            options = NEJoinRoomOptions().apply {
                enableMyAudioDeviceOnJoinRtc = options.enableMyAudioDeviceOnJoinRtc
            },
            callback = FlutterResultCallback(result) { code: Int, message: String?, data: NERoomContext? ->
                if (code == NEErrorCode.SUCCESS && data != null) {
                    data.addRoomListener(RoomListenerPlatform(data.roomUuid, roomEventSink))
                    data.addRtcStatsListener(RoomRtcStatsListenerPlatform(data.roomUuid, roomRtcStatsSink))
                    data.seatController.addSeatListener(SeatListenerPlatform(seatEventSink))
                }
                Pigeon.JoinRoomResponse.Builder()
                    .setCode(code.toLong())
                    .setMsg(message)
                    .setSnapshot(
                        data?.let { ctx ->
                            Pigeon.RoomSnapshot.Builder()
                                .setRoomUuid(ctx.roomUuid)
                                .setRoomName(ctx.roomName)
                                .setPassword(ctx.password)
                                .setRtcStartTime(ctx.rtcStartTime)
                                .setIsRoomLocked(ctx.isRoomLocked)
                                .setProperties(ctx.roomProperties)
                                .setLocalMember(ctx.localMember.mapToRoomMember())
                                .setRemoteMembers(ctx.remoteMembers.map { it.mapToRoomMember() })
                                .setIsChatroomSupported(ctx.chatController.isSupported)
                                .setIsRtcSupported(ctx.rtcController.isSupported)
                                .setIsWhiteboardSupported(ctx.whiteboardController.isSupported)
                                .setIsLiveSupported(ctx.liveController.isSupported)
                                .setIsSeatSupported(ctx.seatController.isSupported)
                                .setIsCloudRecordSupported(false)
                                .setSipCid(ctx.sipCid)
                                .setRemainingSeconds(ctx.remainingSeconds)
                                .build()
                        }
                    ).build()
            }
        )
    }

    override fun previewRoom(
        params: Pigeon.PreviewRoomParams,
        options: Pigeon.PreviewRoomOptions,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.previewRoom(
            NEPreviewRoomParams(),
            NEPreviewRoomOptions(),
            FlutterResultCallback(result) { code: Int, message: String?, data: NEPreviewRoomContext? ->
                if (code == NEErrorCode.SUCCESS && data != null) {
                    data.addPreviewRoomListener(PreviewRoomListenerPlatform(previewRoomEventSink))
                }
                Pigeon.SimpleResponse.Builder()
                    .setCode(code.toLong())
                    .setMsg(message)
                    .build()
            }
        )
    }
}

class RoomChatControllerPlatform : Pigeon.RoomChatControllerApi, IPlatform {

    private val roomService by lazy {
        NERoomKit.instance.getService(NERoomService::class.java)
    }

    override fun onAttachedToEngine(binding: FlutterPlugin.FlutterPluginBinding) {
        Pigeon.RoomChatControllerApi.setup(binding.binaryMessenger, this)
    }

    override fun onDetachedFromEngine(binding: FlutterPlugin.FlutterPluginBinding) {
        Pigeon.RoomChatControllerApi.setup(binding.binaryMessenger, null)
    }

    override fun joinChatroom(roomUuid: String, result: Pigeon.Result<Pigeon.SimpleResponse>) {
        roomService.getRoomContext(roomUuid)?.chatController?.joinChatroom(result.simpleCallback())
            ?: result.success(roomContextNotFound)
    }

    override fun leaveChatroom(roomUuid: String, result: Pigeon.Result<Pigeon.SimpleResponse>) {
        roomService.getRoomContext(roomUuid)?.chatController?.leaveChatroom(result.simpleCallback())
            ?: result.success(roomContextNotFound)
    }

    override fun sendBroadcastTextMessage(
        roomUuid: String,
        message: String,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.chatController?.sendBroadcastTextMessage(message, result.simpleCallback())
            ?: result.success(roomContextNotFound)
    }

    override fun sendDirectTextMessage(
        roomUuid: String,
        userUuid: String,
        message: String,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.chatController?.sendDirectTextMessage(userUuid, message, result.simpleCallback())
            ?: result.success(roomContextNotFound)
    }

    override fun sendGroupTextMessage(
        roomUuid: String,
        userUuids: MutableList<String>,
        message: String,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.chatController?.sendGroupTextMessage(userUuids, message, result.simpleCallback())
            ?: result.success(roomContextNotFound)
    }

    override fun sendImageMessage(
        roomUuid: String,
        messageUuid: String,
        imagePath: String,
        userUuids: MutableList<String>,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.chatController?.sendImageMessage(messageUuid, imagePath, userUuids, result.simpleCallback())
            ?: result.success(roomContextNotFound)
    }

    override fun sendFileMessage(
        roomUuid: String,
        messageUuid: String,
        filePath: String,
        userUuids: MutableList<String>,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.chatController?.sendFileMessage(messageUuid, filePath, userUuids, result.simpleCallback())
            ?: result.success(roomContextNotFound)
    }

    override fun downloadAttachment(
        roomUuid: String,
        messageUuid: String,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.chatController?.downloadAttachment(messageUuid, result.simpleCallback())
            ?: result.success(roomContextNotFound)
    }

    override fun cancelDownloadAttachment(
        roomUuid: String,
        messageUuid: String,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.chatController?.cancelDownloadAttachment(messageUuid, result.simpleCallback())
            ?: result.success(roomContextNotFound)
    }

    override fun fetchChatRoomMembers(
        roomUuid: String,
        queryType: Long,
        limit: Long,
        result: Pigeon.Result<Pigeon.ChatRoomMembersResponse>
    ) {
        val type = when (queryType) {
            1L -> NEChatroomMemberQueryType.ONLINE_NORMAL
            2L -> NEChatroomMemberQueryType.GUEST_DESC
            3L -> NEChatroomMemberQueryType.GUEST_ASC
            else -> NEChatroomMemberQueryType.NORMAL
        }
        roomService.getRoomContext(roomUuid)?.chatController?.fetchChatroomMembers(
            type,
            limit.toInt(),
            object : NECallback<List<NEChatroomMember>> {
                override fun onResult(code: Int, message: String?, data: List<NEChatroomMember>?) {
                    result.success(chatRoomMembersResponse(code, message, data))
                }
            }
        )
    }

    private fun chatRoomMembersResponse(
        code: Int,
        message: String?,
        chatRoomMembers: List<NEChatroomMember>?
    ): Pigeon.ChatRoomMembersResponse {
        val members = mutableListOf<Pigeon.ChatRoomMember>()
        chatRoomMembers?.forEach {
            members.add(
                Pigeon.ChatRoomMember.Builder()
                    .setNick(it.nick)
                    .setAvatar(it.avatar)
                    .build()
            )
        }
        return Pigeon.ChatRoomMembersResponse.Builder()
            .setCode(code.toLong())
            .setMsg(message)
            .setChatRoomMembers(members).build()
    }
}

class PreviewRoomRtcControllerPlatform() : Pigeon.PreviewRoomRtcControllerApi, IPlatform {

    private lateinit var applicationContext: Context

    private val roomService by lazy {
        NERoomKit.instance.getService(NERoomService::class.java)
    }

    override fun onAttachedToEngine(binding: FlutterPlugin.FlutterPluginBinding) {
        applicationContext = binding.applicationContext
        Pigeon.PreviewRoomRtcControllerApi.setup(binding.binaryMessenger, this)
    }

    override fun onDetachedFromEngine(binding: FlutterPlugin.FlutterPluginBinding) {
        Pigeon.PreviewRoomRtcControllerApi.setup(binding.binaryMessenger, null)
    }

    override fun startPreview(result: Pigeon.Result<Pigeon.SimpleResponse>) {
        result.success(
            roomService.getPreviewRoomContext()?.previewController?.let {
                consequence(it.startPreview())
            } ?: roomContextNotFound
        )
    }

    override fun stopPreview(releaseRtc: Boolean, result: Pigeon.Result<Pigeon.SimpleResponse>) {
        result.success(
            roomService.getPreviewRoomContext()?.previewController?.let {
                consequence(it.stopPreview(releaseRtc))
            } ?: roomContextNotFound
        )
    }

    override fun switchCamera(result: Pigeon.Result<Pigeon.SimpleResponse>) {
        result.success(
            roomService.getPreviewRoomContext()?.previewController?.let {
                consequence(it.switchCamera())
            } ?: roomContextNotFound
        )
    }
    override fun addBeautyFilter(path: String, result: Pigeon.Result<Pigeon.SimpleResponse>) {
        result.success(
            roomService.getPreviewRoomContext()?.previewController?.let {
                consequence(it.addBeautyFilter(path))
            } ?: roomContextNotFound
        )
    }

    override fun addBeautySticker(path: String, result: Pigeon.Result<Pigeon.SimpleResponse>) {
        result.success(
            roomService.getPreviewRoomContext()?.previewController?.let {
                consequence(it.addBeautySticker(path))
            } ?: roomContextNotFound
        )
    }

    override fun enableBeauty(
        isOpenBeauty: Boolean,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        result.success(
            roomService.getPreviewRoomContext()?.previewController?.let {
                consequence(it.enableBeauty(isOpenBeauty))
            } ?: roomContextNotFound
        )
    }

    override fun removeBeautyFilter(result: Pigeon.Result<Pigeon.SimpleResponse>) {
        result.success(
            roomService.getPreviewRoomContext()?.previewController?.let {
                consequence(it.removeBeautyFilter())
            } ?: roomContextNotFound
        )
    }

    override fun removeBeautySticker(result: Pigeon.Result<Pigeon.SimpleResponse>) {
        result.success(
            roomService.getPreviewRoomContext()?.previewController?.let {
                consequence(it.removeBeautySticker())
            } ?: roomContextNotFound
        )
    }

    override fun startBeauty(result: Pigeon.Result<Pigeon.SimpleResponse>) {
        result.success(
            roomService.getPreviewRoomContext()?.previewController?.let {
                consequence(it.startBeauty())
            } ?: roomContextNotFound
        )
    }

    override fun stopBeauty(result: Pigeon.Result<Pigeon.SimpleResponse>) {
        result.success(
            roomService.getPreviewRoomContext()?.previewController?.let {
                consequence(it.stopBeauty())
            } ?: roomContextNotFound
        )
    }

    override fun enableVirtualBackground(
        enabled: Boolean,
        backgroundSource: Pigeon.NERoomVirtualBackgroundSource,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        result.success(
            roomService.getPreviewRoomContext()?.previewController?.let {
                consequence(
                    it.enableVirtualBackground(
                        enabled,
                        NERoomVirtualBackgroundSource(
                            backgroundSource.backgroundSourceType.toInt(),
                            backgroundSource.color.toInt(),
                            backgroundSource.source,
                            backgroundSource.blurDegree.toInt()
                        )
                    )
                )
            } ?: roomContextNotFound
        )
    }

    override fun setBeautyFilterLevel(
        level: Double,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        result.success(
            roomService.getPreviewRoomContext()?.previewController?.let {
                consequence(it.setBeautyFilterLevel(level.toFloat()))
            } ?: roomContextNotFound
        )
    }

    override fun setBeautyEffect(
        beautyType: Long,
        level: Double,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        result.success(
            roomService.getPreviewRoomContext()?.previewController?.let {
                consequence(
                    it.setBeautyEffect(
                        NERoomBeautyEffectType.values()[beautyType.toInt()],
                        level.toFloat()
                    )
                )
            } ?: roomContextNotFound
        )
    }
}

class RoomRtcControllerPlatform(
    private val activityResultLauncher: ActivityResultLauncher
) : Pigeon.RoomRtcControllerApi, IPlatform, PluginRegistry.ActivityResultListener {

    private lateinit var applicationContext: Context

    private var startScreenShareResult: Pigeon.Result<Pigeon.SimpleResponse>? = null
    private var startScreenShareRequest: Pigeon.StartScreenShareRequest? = null
    private var mediaProjectionIntent: Intent? = null
    private val mediaProjectionCallback by lazy {
        object : MediaProjection.Callback() {
            override fun onStop() {
                Log.e("RtcControllerPlatform", "MediaProjection stopped")
            }
        }
    }

    private val screenShareRequestCode = hashCode() and 0x0000FFFF

    private val roomService by lazy {
        NERoomKit.instance.getService(NERoomService::class.java)
    }

    init {
        activityResultLauncher.addActivityResultListener(this)
    }

    override fun onActivityResult(requestCode: Int, resultCode: Int, data: Intent?): Boolean {
        if (requestCode == screenShareRequestCode && startScreenShareRequest != null) {
            val request = startScreenShareRequest!!
            if (resultCode == Activity.RESULT_OK && data != null) {
                roomService.getRoomContext(request.roomUuid)?.rtcController?.startScreenShare(
                    data,
                    mediaProjectionCallback,
                    startScreenShareResult!!.simpleCallback { mediaProjectionIntent = data }
                ) ?: startScreenShareResult!!.success(roomContextNotFound)
            } else {
                startScreenShareResult!!.success(consequence(msg = "request screen capture intent error"))
            }
            startScreenShareRequest = null
            startScreenShareResult = null
            return true
        }
        return false
    }

    override fun onAttachedToEngine(binding: FlutterPlugin.FlutterPluginBinding) {
        applicationContext = binding.applicationContext
        Pigeon.RoomRtcControllerApi.setup(binding.binaryMessenger, this)
    }

    override fun onDetachedFromEngine(binding: FlutterPlugin.FlutterPluginBinding) {
        Pigeon.RoomRtcControllerApi.setup(binding.binaryMessenger, null)
    }

    override fun switchCamera(roomUuid: String, result: Pigeon.Result<Pigeon.SimpleResponse>) {
        result.success(
            roomService.getRoomContext(roomUuid)?.rtcController?.let {
                consequence(it.switchCamera())
            } ?: roomContextNotFound
        )
    }

    override fun addBeautyFilter(roomUuid: String, path: String, result: Pigeon.Result<Pigeon.SimpleResponse>) {
        result.success(
            roomService.getRoomContext(roomUuid)?.rtcController?.let {
                consequence(it.addBeautyFilter(path))
            } ?: roomContextNotFound
        )
    }

    override fun addBeautySticker(roomUuid: String, path: String, result: Pigeon.Result<Pigeon.SimpleResponse>) {
        result.success(
            roomService.getRoomContext(roomUuid)?.rtcController?.let {
                consequence(it.addBeautySticker(path))
            } ?: roomContextNotFound
        )
    }

    override fun enableBeauty(
        roomUuid: String,
        isOpenBeauty: Boolean,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        result.success(
            roomService.getRoomContext(roomUuid)?.rtcController?.let {
                consequence(it.enableBeauty(isOpenBeauty))
            } ?: roomContextNotFound
        )
    }

    override fun removeBeautyFilter(roomUuid: String, result: Pigeon.Result<Pigeon.SimpleResponse>) {
        result.success(
            roomService.getRoomContext(roomUuid)?.rtcController?.let {
                consequence(it.removeBeautyFilter())
            } ?: roomContextNotFound
        )
    }

    override fun removeBeautySticker(roomUuid: String, result: Pigeon.Result<Pigeon.SimpleResponse>) {
        result.success(
            roomService.getRoomContext(roomUuid)?.rtcController?.let {
                consequence(it.removeBeautySticker())
            } ?: roomContextNotFound
        )
    }

    override fun startBeauty(roomUuid: String, result: Pigeon.Result<Pigeon.SimpleResponse>) {
        result.success(
            roomService.getRoomContext(roomUuid)?.rtcController?.let {
                consequence(it.startBeauty())
            } ?: roomContextNotFound
        )
    }

    override fun stopBeauty(roomUuid: String, result: Pigeon.Result<Pigeon.SimpleResponse>) {
        result.success(
            roomService.getRoomContext(roomUuid)?.rtcController?.let {
                consequence(it.stopBeauty())
            } ?: roomContextNotFound
        )
    }

    override fun enableVirtualBackground(
        roomUuid: String,
        enabled: Boolean,
        backgroundSource: Pigeon.NERoomVirtualBackgroundSource,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        result.success(
            roomService.getRoomContext(roomUuid)?.rtcController?.let {
                consequence(
                    it.enableVirtualBackground(
                        enabled,
                        NERoomVirtualBackgroundSource(
                            backgroundSource.backgroundSourceType.toInt(),
                            backgroundSource.color.toInt(),
                            backgroundSource.source,
                            backgroundSource.blurDegree.toInt()
                        )
                    )
                )
            } ?: roomContextNotFound
        )
    }

    override fun setBeautyFilterLevel(
        roomUuid: String,
        level: Double,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        result.success(
            roomService.getRoomContext(roomUuid)?.rtcController?.let {
                consequence(it.setBeautyFilterLevel(level.toFloat()))
            } ?: roomContextNotFound
        )
    }

    override fun setBeautyEffect(
        roomUuid: String,
        beautyType: Long,
        level: Double,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        result.success(
            roomService.getRoomContext(roomUuid)?.rtcController?.let {
                consequence(
                    it.setBeautyEffect(
                        NERoomBeautyEffectType.values()[beautyType.toInt()],
                        level.toFloat()
                    )
                )
            } ?: roomContextNotFound
        )
    }

    override fun joinRtcChannel(roomUuid: String, result: Pigeon.Result<Pigeon.SimpleResponse>) {
        roomService.getRoomContext(roomUuid)?.rtcController?.joinRtcChannel(result.simpleCallback())
            ?: result.success(roomContextNotFound)
    }

    override fun leaveRtcChannel(roomUuid: String, result: Pigeon.Result<Pigeon.SimpleResponse>) {
        roomService.getRoomContext(roomUuid)?.rtcController?.leaveRtcChannel(result.simpleCallback())
            ?: result.success(roomContextNotFound)
    }

    override fun muteMyAudio(
        roomUuid: String,
        enableMediaPub: Boolean,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.rtcController?.muteMyAudio(enableMediaPub, result.simpleCallback())
            ?: result.success(roomContextNotFound)
    }

    override fun unmuteMyAudio(
        roomUuid: String,
        enableMediaPub: Boolean,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.rtcController?.unmuteMyAudio(enableMediaPub, result.simpleCallback())
            ?: result.success(roomContextNotFound)
    }

    override fun muteMemberAudio(
        roomUuid: String,
        uuid: String,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.rtcController?.muteMemberAudio(uuid, result.simpleCallback())
            ?: result.success(roomContextNotFound)
    }

    override fun unmuteMemberAudio(
        roomUuid: String,
        uuid: String,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.rtcController?.unmuteMemberAudio(uuid, result.simpleCallback())
            ?: result.success(roomContextNotFound)
    }

    override fun muteMemberVideo(
        roomUuid: String,
        uuid: String,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.rtcController?.muteMemberVideo(uuid, result.simpleCallback())
            ?: result.success(roomContextNotFound)
    }

    override fun unmuteMemberVideo(
        roomUuid: String,
        uuid: String,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.rtcController?.unmuteMemberVideo(uuid, result.simpleCallback())
            ?: result.success(roomContextNotFound)
    }

    override fun muteMyVideo(roomUuid: String, result: Pigeon.Result<Pigeon.SimpleResponse>) {
        roomService.getRoomContext(roomUuid)?.rtcController?.muteMyVideo(result.simpleCallback())
            ?: result.success(roomContextNotFound)
    }

    override fun unmuteMyVideo(roomUuid: String, result: Pigeon.Result<Pigeon.SimpleResponse>) {
        roomService.getRoomContext(roomUuid)?.rtcController?.unmuteMyVideo(result.simpleCallback())
            ?: result.success(roomContextNotFound)
    }

    override fun enableMediaPub(
        roomUuid: String,
        mediaType: Long,
        enable: Boolean,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        result.success(
            roomService.getRoomContext(roomUuid)?.rtcController?.let {
                consequence(it.enableMediaPub(mediaType.toInt(), enable))
            } ?: roomContextNotFound
        )
    }

    override fun pauseLocalAudioRecording(
        roomUuid: String,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        result.success(
            roomService.getRoomContext(roomUuid)?.rtcController?.let {
                consequence(it.pauseLocalAudioRecording())
            } ?: roomContextNotFound
        )
    }

    override fun resumeLocalAudioRecording(
        roomUuid: String,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        result.success(
            roomService.getRoomContext(roomUuid)?.rtcController?.let {
                consequence(it.resumeLocalAudioRecording())
            } ?: roomContextNotFound
        )
    }

    override fun pauseLocalVideoCapture(
        roomUuid: String,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        result.success(
            roomService.getRoomContext(roomUuid)?.rtcController?.let {
                consequence(it.pauseLocalVideoCapture())
            } ?: roomContextNotFound
        )
    }

    override fun resumeLocalVideoCapture(
        roomUuid: String,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        result.success(
            roomService.getRoomContext(roomUuid)?.rtcController?.let {
                consequence(it.resumeLocalVideoCapture())
            } ?: roomContextNotFound
        )
    }

    override fun startScreenShare(
        request: Pigeon.StartScreenShareRequest,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        if (startScreenShareRequest == null) {
            val mediaProjectionManager = applicationContext.getSystemService(
                Context.MEDIA_PROJECTION_SERVICE
            ) as MediaProjectionManager
            val success = activityResultLauncher.launch(
                mediaProjectionManager.createScreenCaptureIntent(),
                screenShareRequestCode
            )
            if (success) {
                startScreenShareRequest = request
                startScreenShareResult = result
            } else {
                result.success(consequence(code = NEErrorCode.FAILURE, msg = "request screen capture intent error"))
            }
        } else {
            result.success(consequence(code = NEErrorCode.FAILURE, msg = "duplicate request"))
        }
    }

    override fun stopScreenShare(roomUuid: String, result: Pigeon.Result<Pigeon.SimpleResponse>) {
        startScreenShareRequest = null
        startScreenShareResult = null
        mediaProjectionIntent = null
        roomService.getRoomContext(roomUuid)?.rtcController?.stopScreenShare(result.simpleCallback())
            ?: result.success(roomContextNotFound)
    }

    override fun stopMemberScreenShare(
        roomUuid: String,
        uuid: String,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.rtcController?.stopMemberScreenShare(uuid, result.simpleCallback())
            ?: result.success(roomContextNotFound)
    }

    override fun subscribeRemoteVideoStream(
        roomUuid: String,
        userUuid: String,
        streamType: Long,
        subscribe: Boolean,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        result.success(
            roomService.getRoomContext(roomUuid)?.rtcController?.let {
                val videoStreamType =
                    if (streamType == 0L) NEVideoStreamType.HIGH else NEVideoStreamType.LOW
                consequence(
                    if (subscribe) {
                        it.subscribeRemoteVideoStream(userUuid, videoStreamType)
                    } else {
                        it.unsubscribeRemoteVideoStream(userUuid, videoStreamType)
                    }
                )
            } ?: roomContextNotFound
        )
    }

    override fun subscribeRemoteSubStreamVideo(
        roomUuid: String,
        userUuid: String,
        subscribe: Boolean,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        result.success(
            roomService.getRoomContext(roomUuid)?.rtcController?.let {
                consequence(
                    if (subscribe) {
                        it.subscribeRemoteVideoSubStream(userUuid)
                    } else {
                        it.unsubscribeRemoteVideoSubStream(userUuid)
                    }
                )
            } ?: roomContextNotFound
        )
    }

    override fun setSpeakerphoneOn(
        roomUuid: String,
        on: Boolean,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.rtcController?.run {
            result.success(consequence(setSpeakerphoneOn(on)))
        } ?: result.success(roomContextNotFound)
    }

    override fun isSpeakerphoneOn(
        roomUuid: String,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.rtcController?.run {
            result.success(consequence(if (isSpeakerphoneOn()) 0 else 1))
        } ?: result.success(roomContextNotFound)
    }

    override fun stopAudioDump(
        roomUuid: String,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.rtcController?.run {
            result.success(consequence(stopAudioDump()))
        } ?: result.success(roomContextNotFound)
    }

    override fun startRtcChannelMediaRelay(
        roomUuid: String,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.rtcController?.startChannelMediaRelay(result.simpleCallback())
            ?: result.success(roomContextNotFound)
    }

    override fun stopRtcChannelMediaRelay(
        roomUuid: String,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        result.success(
            roomService.getRoomContext(roomUuid)?.rtcController?.let {
                consequence(it.stopChannelMediaRelay())
            } ?: roomContextNotFound
        )
    }

    override fun enableEarBack(
        roomUuid: String,
        enable: Boolean,
        volume: Long,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        result.success(
            roomService.getRoomContext(roomUuid)?.rtcController?.let {
                consequence(
                    if (enable) {
                        it.enableEarBack(volume.toInt())
                    } else {
                        it.disableEarBack()
                    }
                )
            } ?: roomContextNotFound
        )
    }

    override fun startAudioMixing(
        roomUuid: String,
        option: Pigeon.CreateAudioMixingOption,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.rtcController?.run {
            result.success(
                consequence(
                    startAudioMixing(
                        NERoomCreateAudioMixingOption(
                            path = option.path,
                            loopCount = option.loopCount.toInt(),
                            sendEnabled = option.sendEnabled,
                            sendVolume = option.sendVolume.toInt(),
                            playbackEnabled = option.playbackEnabled,
                            playbackVolume = option.playbackVolume.toInt(),
                            startTimeStamp = option.startTimeStamp,
                            sendWithAudioType = if (option.sendWithAudioType == Pigeon.AudioStreamType.AUDIO_STREAM_TYPE_MAIN) NERoomRtcAudioStreamType.NERtcAudioStreamTypeMain else NERoomRtcAudioStreamType.NERtcAudioStreamTypeSub
                        )
                    )
                )
            )
        } ?: result.success(roomContextNotFound)
    }

    override fun playEffect(
        roomUuid: String,
        id: Long,
        option: Pigeon.CreateAudioEffectOption,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.rtcController?.run {
            result.success(
                consequence(
                    playEffect(
                        id.toInt(),
                        NERoomCreateAudioEffectOption(
                            path = option.path,
                            loopCount = option.loopCount.toInt(),
                            sendEnabled = option.sendEnabled,
                            sendVolume = option.sendVolume.toInt(),
                            playbackEnabled = option.playbackEnabled,
                            playbackVolume = option.playbackVolume.toInt(),
                            startTimestamp = option.startTimeStamp,
                            progressInterval = option.progressInterval,
                            sendWithAudioType = if (option.sendWithAudioType == Pigeon.AudioStreamType.AUDIO_STREAM_TYPE_MAIN) NERoomRtcAudioStreamType.NERtcAudioStreamTypeMain else NERoomRtcAudioStreamType.NERtcAudioStreamTypeSub
                        )
                    )
                )
            )
        } ?: result.success(roomContextNotFound)
    }

    override fun stopAudioMixing(
        roomUuid: String,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.rtcController?.run {
            result.success(consequence(stopAudioMixing()))
        } ?: result.success(roomContextNotFound)
    }

    override fun stopEffect(
        roomUuid: String,
        id: Long,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.rtcController?.run {
            result.success(consequence(stopEffect(id.toInt())))
        } ?: result.success(roomContextNotFound)
    }

    override fun setAudioMixingSendVolume(
        roomUuid: String,
        volume: Long,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.rtcController?.run {
            result.success(consequence(setAudioMixingSendVolume(volume.toInt())))
        } ?: result.success(roomContextNotFound)
    }

    override fun setAudioMixingPlaybackVolume(
        roomUuid: String,
        volume: Long,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.rtcController?.run {
            result.success(consequence(setAudioMixingPlaybackVolume(volume.toInt())))
        } ?: result.success(roomContextNotFound)
    }

    override fun setEffectSendVolume(
        roomUuid: String,
        id: Long,
        volume: Long,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.rtcController?.run {
            result.success(consequence(setEffectSendVolume(id.toInt(), volume.toInt())))
        } ?: result.success(roomContextNotFound)
    }

    override fun setEffectPlaybackVolume(
        roomUuid: String,
        id: Long,
        volume: Long,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.rtcController?.run {
            result.success(consequence(setEffectPlaybackVolume(id.toInt(), volume.toInt())))
        } ?: result.success(roomContextNotFound)
    }

    override fun stopAllEffects(
        roomUuid: String,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.rtcController?.run {
            result.success(consequence(stopAllEffects()))
        } ?: result.success(roomContextNotFound)
    }

    override fun adjustUserPlaybackSignalVolume(
        roomUuid: String,
        userUuid: String,
        volume: Long,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        result.success(
            roomService.getRoomContext(roomUuid)?.rtcController?.let {
                consequence(it.adjustUserPlaybackSignalVolume(userUuid, volume.toInt()))
            } ?: roomContextNotFound
        )
    }

    override fun adjustPlaybackSignalVolume(
        roomUuid: String,
        volume: Long,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        result.success(
            roomService.getRoomContext(roomUuid)?.rtcController?.let {
                consequence(it.adjustPlaybackSignalVolume(volume.toInt()))
            } ?: roomContextNotFound
        )
    }

    override fun startAudioDump(
        roomUuid: String,
        type: Long,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.rtcController?.run {
            result.success(
                consequence(
                    startAudioDump(
                        when (type) {
                            1L -> NEAudioDumpType.ALL
                            2L -> NEAudioDumpType.WAV
                            else -> NEAudioDumpType.PCM
                        }
                    )
                )
            )
        } ?: result.success(roomContextNotFound)
    }

    override fun enableAudioVolumeIndication(
        roomUuid: String,
        enable: Boolean,
        interval: Long,
        enableVad: Boolean,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.rtcController?.run {
            result.success(consequence(enableAudioVolumeIndication(enable, interval.toInt(), enableVad)))
        } ?: result.success(roomContextNotFound)
    }

    override fun pauseAudioMixing(
        roomUuid: String,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.rtcController?.run {
            result.success(consequence(pauseAudioMixing()))
        } ?: result.success(roomContextNotFound)
    }

    override fun adjustRecordingSignalVolume(
        roomUuid: String,
        volume: Long,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.rtcController?.run {
            result.success(consequence(adjustRecordingSignalVolume(volume.toInt())))
        } ?: result.success(roomContextNotFound)
    }

    override fun resumeAudioMixing(
        roomUuid: String,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.rtcController?.run {
            result.success(consequence(resumeAudioMixing()))
        } ?: result.success(roomContextNotFound)
    }

    override fun enableLoopbackRecording(
        roomUuid: String,
        enable: Boolean,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.rtcController?.run {
            if (enable && mediaProjectionIntent == null) {
                result.success(
                    consequence(
                        NEErrorCode.FAILURE,
                        "Loopback recording is only supported when screen sharing"
                    )
                )
                return
            }
            result.success(consequence(enableLoopbackRecording(enable, mediaProjectionIntent, mediaProjectionCallback)))
        } ?: result.success(roomContextNotFound)
    }

    override fun enableAudioAINS(
        roomUuid: String,
        enable: Boolean,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.rtcController?.run {
            result.success(consequence(enableAudioAINS(enable)))
        } ?: result.success(roomContextNotFound)
    }

    override fun setRecordDeviceMute(
        roomUuid: String,
        mute: Boolean,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.rtcController?.run {
            result.success(consequence(setRecordDeviceMute(mute)))
        } ?: result.success(roomContextNotFound)
    }

    override fun setClientRole(
        roomUuid: String,
        role: Long,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.rtcController?.run {
            result.success(consequence(setClientRole(role.toInt())))
        } ?: result.success(roomContextNotFound)
    }

    override fun setLocalAudioProfile(
        roomUuid: String,
        profile: Long,
        scenario: Long,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.rtcController?.run {
            result.success(consequence(setAudioProfile(profile.toInt(), scenario.toInt())))
        } ?: result.success(roomContextNotFound)
    }

    override fun setChannelProfile(
        roomUuid: String,
        value: Long,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.rtcController?.run {
            result.success(consequence(setChannelProfile(value.toInt())))
        } ?: result.success(roomContextNotFound)
    }
}

data class RenderEntry(
    val roomUuid: String,
    val render: FlutterVideoRenderer
)

class RoomRtcRenderPlatform : Pigeon.RoomRtcVideoRender, IPlatform {
    private lateinit var registry: TextureRegistry
    private lateinit var messenger: BinaryMessenger
    private lateinit var context: Context

    private val renderers = hashMapOf<String, RenderEntry>()

    private val roomService by lazy {
        NERoomKit.instance.getService(NERoomService::class.java)
    }

    private fun ofRtcController(roomUuid: String) =
        roomService.getRoomContext(roomUuid)?.rtcController

    override fun onAttachedToEngine(binding: FlutterPlugin.FlutterPluginBinding) {
        Pigeon.RoomRtcVideoRender.setup(binding.binaryMessenger, this)
        registry = binding.textureRegistry
        messenger = binding.binaryMessenger
        context = binding.applicationContext
    }

    override fun onDetachedFromEngine(binding: FlutterPlugin.FlutterPluginBinding) {
        Pigeon.RoomRtcVideoRender.setup(binding.binaryMessenger, null)
    }

    override fun createVideoRenderer(
        roomUuid: String,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        val eglContext = if (TextUtils.isEmpty(roomUuid)) {
            roomService.getPreviewRoomContext()?.previewController?.getEglContext()
        } else {
            roomService.getRoomContext(roomUuid)?.rtcController?.getEglContext()
        }
        if (eglContext != null) {
            val entry = registry.createSurfaceTexture()
            val renderer = FlutterVideoRenderer(
                messenger,
                entry,
                getEglBaseContext(eglContext)
            )
            renderers[entry.id().toString()] = RenderEntry(roomUuid, renderer)
            result.success(
                Pigeon.SimpleResponse.Builder()
                    .setCode(entry.id())
                    .build()
            )
        } else {
            result.success(roomContextNotFound)
        }
    }

    // ////////////////////////////////////////////////////////////////////////////////////////////
    // ///////////////////////////////// VideoRendererApi /////////////////////////////////////////////
    // 支持外部输入eglContext
    private fun getEglBaseContext(eglContext: Any?): EglBase.Context? {
        if (eglContext is EGLContext) {
            return EglBase14.Context(eglContext)
        } else if (eglContext is javax.microedition.khronos.egl.EGLContext) {
            return EglBase10.Context(eglContext)
        }
        return null
    }

    // override fun setupLocalSubStreamVideoRenderer(
    //    textureIdRendererRequest: Pigeon.TextureIdRendererRequest,
    //    result: Pigeon.Result<Pigeon.SimpleResponse>,
    // ) {
    //    val code = renderers[textureIdRendererRequest.textureId.toString()]?.let {
    //        NERtcEx.getInstance().setupLocalSubStreamVideoCanvas(it)
    //    } ?: NEErrorCode.FAILURE
    //    result.success(consequence(code = code.toLong()))
    // }

    override fun setupLocalVideoRenderer(
        textureId: Long,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        renderers[textureId.toString()]?.let { entry ->
            val controller = if (TextUtils.isEmpty(entry.roomUuid)) {
                roomService.getPreviewRoomContext()?.previewController
            } else {
                ofRtcController(entry.roomUuid)
            }
            result.success(
                controller?.let {
                    entry.render.prepareToAttach()
                    consequence(it.setupLocalVideoRender(entry.render))
                } ?: roomContextNotFound
            )
        } ?: result.success(consequence(code = NEErrorCode.FAILURE, msg = "renderer not found"))
    }

    override fun setMirror(
        textureId: Long,
        mirror: Boolean,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        val code = renderers[textureId.toString()]?.let { entry ->
            entry.render.setRequestedMirror(mirror)
            NEErrorCode.SUCCESS
        } ?: NEErrorCode.FAILURE
        result.success(consequence(code = code))
    }

    override fun setupRemoteSubStreamVideoRenderer(
        userUuid: String,
        textureId: Long,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        renderers[textureId.toString()]?.let { entry ->
            val rtcController = ofRtcController(entry.roomUuid)
            result.success(
                rtcController?.let {
                    entry.render.prepareToAttach()
                    consequence(it.setupRemoteVideoSubStreamRender(entry.render, userUuid))
                } ?: roomContextNotFound
            )
        } ?: result.success(consequence(code = NEErrorCode.FAILURE, msg = "renderer not found"))
    }

    override fun setupRemoteVideoRenderer(
        userUuid: String,
        textureId: Long,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        renderers[textureId.toString()]?.let { entry ->
            val rtcController = ofRtcController(entry.roomUuid)
            result.success(
                rtcController?.let {
                    entry.render.prepareToAttach()
                    consequence(it.setupRemoteVideoRender(entry.render, userUuid))
                } ?: roomContextNotFound
            )
        } ?: result.success(consequence(code = NEErrorCode.FAILURE))
    }

    override fun detachLocalVideoRenderer(
        subStream: Boolean,
        textureId: Long,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        renderers[textureId.toString()]?.let { entry ->
            if (subStream) {
                result.success(consequence(code = NEErrorCode.FAILURE, msg = "not support"))
            } else {
                val controller = if (TextUtils.isEmpty(entry.roomUuid)) {
                    roomService.getPreviewRoomContext()?.previewController
                } else {
                    ofRtcController(entry.roomUuid)
                }
                result.success(
                    controller?.let {
                        consequence(it.setupLocalVideoRender(null))
                    } ?: roomContextNotFound
                )
            }
        } ?: result.success(consequence(code = NEErrorCode.FAILURE, msg = "renderer not found"))
    }

    override fun detachRemoteVideoRenderer(
        userUuid: String,
        subStream: Boolean,
        textureId: Long,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        renderers[textureId.toString()]?.let { entry ->
            result.success(
                ofRtcController(entry.roomUuid)?.let {
                    consequence(
                        if (subStream) {
                            it.setupRemoteVideoSubStreamRender(null, userUuid)
                        } else {
                            it.setupRemoteVideoRender(null, userUuid)
                        }
                    )
                } ?: roomContextNotFound
            )
        } ?: result.success(consequence(code = NEErrorCode.FAILURE, msg = "renderer not found"))
    }

    override fun disposeVideoRenderer(
        textureId: Long,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        val code = renderers[textureId.toString()]?.let { entry ->
            renderers.remove(textureId.toString())
            entry.render.dispose()
            NEErrorCode.SUCCESS
        } ?: NEErrorCode.FAILURE
        result.success(consequence(code = code))
    }
}

class RoomWhiteboardControllerPlatform : Pigeon.RoomWhiteboardControllerApi, IPlatform {

    private val roomService by lazy {
        NERoomKit.instance.getService(NERoomService::class.java)
    }

    override fun onAttachedToEngine(binding: FlutterPlugin.FlutterPluginBinding) {
        Pigeon.RoomWhiteboardControllerApi.setup(binding.binaryMessenger, this)
        binding.platformViewRegistry.registerViewFactory("com.netease.yunxin.flutter.plugins.roomkit/whiteboardView", WhiteboardViewFactory())
    }

    override fun onDetachedFromEngine(binding: FlutterPlugin.FlutterPluginBinding) {
        Pigeon.RoomWhiteboardControllerApi.setup(binding.binaryMessenger, null)
    }

    override fun startWhiteboardShare(
        roomUuid: String,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.whiteboardController?.startWhiteboardShare(result.simpleCallback())
            ?: result.success(roomContextNotFound)
    }

    override fun stopWhiteboardShare(
        roomUuid: String,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.whiteboardController?.stopWhiteboardShare(result.simpleCallback())
            ?: result.success(roomContextNotFound)
    }

    override fun stopMemberWhiteboardShare(
        roomUuid: String,
        userUuid: String,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.whiteboardController?.stopMemberWhiteboardShare(userUuid, result.simpleCallback())
            ?: result.success(roomContextNotFound)
    }

    override fun setEnableDraw(
        roomUuid: String,
        enable: Boolean,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.let {
            it.whiteboardController.setEnableDraw(enable)
            result.success(consequence(NEErrorCode.SUCCESS))
        } ?: result.success(roomContextNotFound)
    }

    override fun setCanvasBackgroundColor(
        roomUuid: String,
        color: String,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.let {
            result.success(consequence(it.whiteboardController.setCanvasBackgroundColor(color)))
        } ?: result.success(roomContextNotFound)
    }

    override fun lockCameraWithContent(
        roomUuid: String,
        width: Long,
        height: Long,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.let {
            result.success(consequence(it.whiteboardController.lockCameraWithContent(width.toInt(), height.toInt())))
        } ?: result.success(roomContextNotFound)
    }
}

class RoomContextPlatform : Pigeon.RoomContextApi, IPlatform {

    private val roomService by lazy {
        NERoomKit.instance.getService(NERoomService::class.java)
    }

    override fun onAttachedToEngine(binding: FlutterPlugin.FlutterPluginBinding) {
        Pigeon.RoomContextApi.setup(binding.binaryMessenger, this)
    }

    override fun onDetachedFromEngine(binding: FlutterPlugin.FlutterPluginBinding) {
        Pigeon.RoomContextApi.setup(binding.binaryMessenger, null)
    }

    override fun changeMemberRole(
        roomUuid: String,
        userUuid: String,
        role: String,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.changeMemberRole(userUuid, role, result.simpleCallback())
            ?: result.success(roomContextNotFound)
    }

    override fun changeMyName(roomUuid: String, name: String, result: Pigeon.Result<Pigeon.SimpleResponse>) {
        roomService.getRoomContext(roomUuid)?.changeMyName(name, result.simpleCallback())
            ?: result.success(roomContextNotFound)
    }

    override fun handOverMyRole(
        roomUuid: String,
        userUuid: String,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.handOverMyRole(userUuid, result.simpleCallback())
            ?: result.success(roomContextNotFound)
    }

    override fun kickMemberOut(
        roomUuid: String,
        userUuid: String,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.kickMemberOut(userUuid, result.simpleCallback())
            ?: result.success(roomContextNotFound)
    }

    override fun leaveRoom(roomUuid: String, result: Pigeon.Result<Pigeon.SimpleResponse>) {
        roomService.getRoomContext(roomUuid)?.leaveRoom(result.simpleCallback())
            ?: result.success(roomContextNotFound)
    }

    override fun endRoom(roomUuid: String, isForce: Boolean, result: Pigeon.Result<Pigeon.SimpleResponse>) {
        roomService.getRoomContext(roomUuid)?.endRoom(isForce, result.simpleCallback())
            ?: result.success(roomContextNotFound)
    }

    override fun updateRoomProperty(
        roomUuid: String,
        key: String,
        value: String,
        associatedUuid: String?,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.updateRoomProperty(key, value, associatedUuid, result.simpleCallback())
            ?: result.success(roomContextNotFound)
    }

    override fun deleteRoomProperty(
        roomUuid: String,
        key: String,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.deleteRoomProperty(key, result.simpleCallback())
            ?: result.success(roomContextNotFound)
    }

    override fun updateMemberProperty(
        roomUuid: String,
        userUuid: String,
        key: String,
        value: String,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.updateMemberProperty(userUuid, key, value, result.simpleCallback())
            ?: result.success(roomContextNotFound)
    }

    override fun deleteMemberProperty(
        roomUuid: String,
        userUuid: String,
        key: String,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.deleteMemberProperty(userUuid, key, result.simpleCallback())
            ?: result.success(roomContextNotFound)
    }

    override fun lockRoom(roomUuid: String, result: Pigeon.Result<Pigeon.SimpleResponse>) {
        roomService.getRoomContext(roomUuid)?.lockRoom(result.simpleCallback())
            ?: result.success(roomContextNotFound)
    }

    override fun unlockRoom(roomUuid: String, result: Pigeon.Result<Pigeon.SimpleResponse>) {
        roomService.getRoomContext(roomUuid)?.unlockRoom(result.simpleCallback())
            ?: result.success(roomContextNotFound)
    }
}

class PreviewRoomListenerPlatform(
    private val eventSink: Pigeon.PreviewRoomEventSink?
) : NEPreviewRoomListenerAdapter() {
    private val reply = Pigeon.PreviewRoomEventSink.Reply<Void> { }
    override fun onRtcVirtualBackgroundSourceEnabled(enabled: Boolean, reason: Int) {
        eventSink?.onRtcVirtualBackgroundSourceEnabled(
            Pigeon.RtcVirtualBackgroundSourceEvent.Builder()
                .setRoomUuid("")
                .setEnabled(enabled)
                .setReason(reason.toLong())
                .build(),
            reply
        )
    }
}

class RoomListenerPlatform(
    private val roomUuid: String,
    private val eventSink: Pigeon.RoomEventSink?
) : NERoomListenerAdapter() {

    private val reply = Pigeon.RoomEventSink.Reply<Void> { }

    override fun onMemberRoleChanged(
        member: NERoomMember,
        oldRole: NERoomRole,
        newRole: NERoomRole
    ) {
        eventSink?.onMemberRoleChanged(
            Pigeon.MemberRoleChangeEvent
                .Builder()
                .setRoomUuid(roomUuid)
                .setRole(newRole.mapToRoomRole())
                .setUserUuid(member.uuid)
                .build(),
            reply
        )
    }

    override fun onMemberNameChanged(member: NERoomMember, name: String) {
        eventSink?.onMemberNameChanged(
            Pigeon.CommonEvent
                .Builder()
                .setRoomUuid(roomUuid)
                .setUserUuid(member.uuid)
                .setSArg1(name)
                .build(),
            reply
        )
    }

    override fun onMemberJoinRoom(members: List<NERoomMember>) {
        eventSink?.onMemberJoinRoom(
            Pigeon.MembersEvent.Builder()
                .setRoomUuid(roomUuid)
                .setMembers(members.map { it.mapToRoomMember() })
                .build(),
            reply
        )
    }

    override fun onMemberLeaveRoom(members: List<NERoomMember>) {
        eventSink?.onMemberLeaveRoom(
            Pigeon.MembersEvent.Builder()
                .setRoomUuid(roomUuid)
                .setUuids(members.map { it.uuid })
                .build(),
            reply
        )
    }

    override fun onRoomEnded(reason: NERoomEndReason) {
        eventSink?.onRoomEnd(roomUuid, stringifyRoomEndReason(reason), reply)
    }

    override fun onMemberJoinRtcChannel(members: List<NERoomMember>) {
        eventSink?.onMemberJoinRtcChannel(
            Pigeon.MembersEvent.Builder()
                .setRoomUuid(roomUuid)
                .setUuids(members.map { it.uuid }).build(),
            reply
        )
    }

    override fun onMemberLeaveRtcChannel(members: List<NERoomMember>) {
        eventSink?.onMemberLeaveRtcChannel(
            Pigeon.MembersEvent.Builder()
                .setRoomUuid(roomUuid)
                .setUuids(members.map { it.uuid }).build(),
            reply
        )
    }

    override fun onRtcChannelError(code: Int) {
        eventSink?.onRtcChannelError(
            Pigeon.CommonEvent.Builder()
                .setRoomUuid(roomUuid)
                .setArg1(code.toLong()).build(),
            reply
        )
    }

    override fun onRtcRecvSEIMsg(uuid: String, seiMsg: String) {
    }

    override fun onMemberJoinChatroom(members: List<NERoomMember>) {
        eventSink?.onMemberJoinChatroom(
            Pigeon.MembersEvent.Builder()
                .setRoomUuid(roomUuid)
                .setUuids(members.map { it.uuid }).build(),
            reply
        )
    }

    override fun onMemberLeaveChatroom(members: List<NERoomMember>) {
        eventSink?.onMemberLeaveChatroom(
            Pigeon.MembersEvent.Builder()
                .setRoomUuid(roomUuid)
                .setUuids(members.map { it.uuid }).build(),
            reply
        )
    }

    override fun onMemberAudioMuteChanged(member: NERoomMember, mute: Boolean, operateBy: NERoomMember?) {
        eventSink?.onMemberAudioMuteChanged(
            Pigeon.MemberToggleEvent.Builder()
                .setRoomUuid(roomUuid)
                .setOperateUuid(operateBy?.uuid)
                .setUserUuid(member.uuid)
                .setIsOn(!mute)
                .build(),
            reply
        )
    }

    override fun onMemberVideoMuteChanged(member: NERoomMember, mute: Boolean, operateBy: NERoomMember?) {
        eventSink?.onMemberVideoMuteChanged(
            Pigeon.MemberToggleEvent.Builder()
                .setRoomUuid(roomUuid)
                .setUserUuid(member.uuid)
                .setOperateUuid(operateBy?.uuid)
                .setIsOn(!mute)
                .build(),
            reply
        )
    }

    override fun onMemberScreenShareStateChanged(member: NERoomMember, isSharing: Boolean, operateBy: NERoomMember?) {
        eventSink?.onMemberScreenShareStateChanged(
            Pigeon.MemberToggleEvent.Builder()
                .setRoomUuid(roomUuid)
                .setUserUuid(member.uuid)
                .setOperateUuid(operateBy?.uuid)
                .setIsOn(isSharing)
                .build(),
            reply
        )
    }

    override fun onReceiveChatroomMessages(messages: List<NERoomChatMessage>) {
        eventSink?.onReceiveChatroomMessages(
            Pigeon.RoomMessageEvent.Builder()
                .setRoomUuid(roomUuid)
                .setMessages(messages.map { it.mapToRoomMessage() })
                .build(),
            reply
        )
    }

    override fun onAudioMixingStateChanged(reason: Int) {
        eventSink?.onAudioMixingStateChanged(
            Pigeon.CommonEvent.Builder()
                .setRoomUuid(roomUuid)
                .setArg1(reason.toLong())
                .build(),
            reply
        )
    }

    override fun onChatroomMessageAttachmentProgress(
        messageUuid: String,
        transferred: Long,
        total: Long
    ) {
        eventSink?.onChatroomMessageAttachmentProgress(
            Pigeon.CommonEvent.Builder()
                .setRoomUuid(roomUuid)
                .setSArg1(messageUuid)
                .setArg1(transferred)
                .setArg2(total)
                .build(),
            reply
        )
    }

    override fun onMemberWhiteboardStateChanged(member: NERoomMember, isSharing: Boolean, operateBy: NERoomMember?) {
        eventSink?.onMemberWhiteboardShareStateChanged(
            Pigeon.MemberToggleEvent.Builder()
                .setRoomUuid(roomUuid)
                .setUserUuid(member.uuid)
                .setOperateUuid(operateBy?.uuid)
                .setIsOn(isSharing)
                .build(),
            reply
        )
    }

    override fun onWhiteboardError(code: Int, message: String) {
        eventSink?.onWhiteboardError(
            Pigeon.CommonEvent.Builder()
                .setRoomUuid(roomUuid)
                .setArg1(code.toLong())
                .setSArg1(message)
                .build(),
            reply
        )
    }

    override fun onRoomLiveStateChanged(state: NERoomLiveState) {
        eventSink?.onRoomLiveStateChanged(
            Pigeon.CommonEvent.Builder()
                .setRoomUuid(roomUuid)
                .setArg1(state.value.toLong())
                .build(),
            reply
        )
    }

    override fun onWhiteboardShowFileChooser(
        types: Array<String>,
        callback: NEValueCallback<Array<Uri>?>
    ) {
        eventSink?.onWhiteboardShowFileChooser(
            Pigeon.WhiteboardShowFileChooserTypeEvent.Builder()
                .setRoomUuid(roomUuid)
                .setTypes(types.toMutableList())
                .build()
        ) {
            callback.onReceiveValue(
                value = it?.let {
                    arrayOf(Uri.fromFile(File(it)))
                }
            )
        }
    }

    override fun onRoomPropertiesChanged(properties: Map<String, String>) {
        eventSink?.onRoomPropertiesChanged(
            Pigeon.PropertyEvent.Builder()
                .setRoomUuid(roomUuid)
                .setProperties(properties)
                .build(),
            reply
        )
    }

    override fun onRoomPropertiesDeleted(properties: Map<String, String>) {
        eventSink?.onRoomPropertiesDeleted(
            Pigeon.PropertyEvent.Builder()
                .setRoomUuid(roomUuid)
                .setProperties(properties)
                .build(),
            reply
        )
    }

    override fun onMemberPropertiesChanged(member: NERoomMember, properties: Map<String, String>) {
        eventSink?.onMemberPropertiesChanged(
            Pigeon.PropertyEvent.Builder()
                .setRoomUuid(roomUuid)
                .setUserUuid(member.uuid)
                .setProperties(properties)
                .build(),
            reply
        )
    }

    override fun onMemberPropertiesDeleted(member: NERoomMember, properties: Map<String, String>) {
        eventSink?.onMemberPropertiesDeleted(
            Pigeon.PropertyEvent.Builder()
                .setRoomUuid(roomUuid)
                .setUserUuid(member.uuid)
                .setProperties(properties)
                .build(),
            reply
        )
    }

    override fun onRoomLockStateChanged(isLocked: Boolean) {
        eventSink?.onRoomLockStateChanged(
            Pigeon.RoomToggleEvent.Builder()
                .setRoomUuid(roomUuid)
                .setIsOn(isLocked)
                .build(),
            reply
        )
    }

    override fun onRtcRemoteAudioVolumeIndication(
        volumes: List<NEMemberVolumeInfo>,
        totalVolume: Int
    ) {
        eventSink?.onRtcAudioVolumeIndication(
            Pigeon.RtcVolumeEvent.Builder()
                .setRoomUuid(roomUuid)
                .setTotalVolume(totalVolume.toLong())
                .setVolumes(
                    volumes.map {
                        Pigeon.MemberVolumeInfo.Builder()
                            .setUserUuid(it.userUuid)
                            .setVolume(it.volume.toLong())
                            .build()
                    }.toList()
                )
                .build(),
            reply
        )
    }

    override fun onRtcLocalAudioVolumeIndication(volume: Int, vadFlag: Boolean) {
        eventSink?.onLocalAudioVolumeIndicationwithVad(
            Pigeon.LocalVolumeEvent.Builder()
                .setRoomUuid(roomUuid)
                .setVolume(volume.toLong())
                .setEnableVad(vadFlag)
                .build(),
            reply
        )
    }

    override fun onRtcAudioOutputDeviceChanged(device: NEAudioOutputDevice) {
        eventSink?.onRtcAudioOutputDeviceChanged(
            Pigeon.CommonEvent.Builder()
                .setRoomUuid(roomUuid)
                .setArg1(device.ordinal.toLong())
                .build(),
            reply
        )
    }

    override fun onRtcVirtualBackgroundSourceEnabled(enabled: Boolean, reason: Int) {
        eventSink?.onRtcVirtualBackgroundSourceEnabled(
            Pigeon.RtcVirtualBackgroundSourceEvent.Builder()
                .setRoomUuid(roomUuid)
                .setEnabled(enabled)
                .setReason(reason.toLong())
                .build(),
            reply
        )
    }
}

class SeatListenerPlatform(
    private val eventSink: Pigeon.SeatEventSink?
) : NESeatEventListener() {
    private val reply = Pigeon.SeatEventSink.Reply<Void> { }

    override fun onSeatInvitationAccepted(seatIndex: Int, user: String, isAutoAgree: Boolean) {
        eventSink?.onSeatInvitationAccepted(seatIndex.toLong(), user, isAutoAgree, reply)
    }

    override fun onSeatInvitationCancelled(seatIndex: Int, user: String, operateBy: String) {
        eventSink?.onSeatInvitationCancelled(seatIndex.toLong(), user, operateBy, reply)
    }

    override fun onSeatInvitationReceived(seatIndex: Int, user: String, operateBy: String) {
        eventSink?.onSeatInvitationReceived(seatIndex.toLong(), user, operateBy, reply)
    }

    override fun onSeatKicked(seatIndex: Int, user: String, operateBy: String) {
        eventSink?.onSeatKicked(seatIndex.toLong(), user, operateBy, reply)
    }

    override fun onSeatLeave(seatIndex: Int, user: String) {
        eventSink?.onSeatLeave(seatIndex.toLong(), user, reply)
    }

    override fun onSeatInvitationRejected(seatIndex: Int, user: String) {
        eventSink?.onSeatInvitationRejected(seatIndex.toLong(), user, reply)
    }

    override fun onSeatManagerAdded(managers: List<String>) {
        eventSink?.onSeatManagerAdded(managers, reply)
    }

    override fun onSeatManagerRemoved(managers: List<String>) {
        eventSink?.onSeatManagerRemoved(managers, reply)
    }

    override fun onSeatRequestApproved(
        seatIndex: Int,
        user: String,
        operateBy: String,
        isAutoAgree: Boolean
    ) {
        eventSink?.onSeatRequestApproved(seatIndex.toLong(), user, operateBy, isAutoAgree, reply)
    }

    override fun onSeatRequestCancelled(seatIndex: Int, user: String) {
        eventSink?.onSeatRequestCancelled(seatIndex.toLong(), user, reply)
    }

    override fun onSeatRequestRejected(seatIndex: Int, user: String, operateBy: String) {
        eventSink?.onSeatRequestRejected(seatIndex.toLong(), user, operateBy, reply)
    }

    override fun onSeatRequestSubmitted(seatIndex: Int, user: String) {
        eventSink?.onSeatRequestSubmitted(seatIndex.toLong(), user, reply)
    }

    override fun onSeatListChanged(seatItems: List<NESeatItem>) {
        val list = ArrayList<Pigeon.SeatItem>()
        for (seatItem in seatItems) {
            list.add(
                Pigeon.SeatItem.Builder()
                    .setUserName(seatItem.userName)
                    .setUser(seatItem.user)
                    .setUpdated(seatItem.updated)
                    .setStatus(seatItem.status.toLong())
                    .setOnSeatType(seatItem.onSeatType.toLong())
                    .setIndex(seatItem.index.toLong())
                    .setIcon(seatItem.icon)
                    .build()
            )
        }
        eventSink?.onSeatListChanged(list, reply)
    }
}

class RoomLiveControllerPlatform : Pigeon.RoomLiveControllerApi, IPlatform {

    private val roomService by lazy {
        NERoomKit.instance.getService(NERoomService::class.java)
    }

    override fun onAttachedToEngine(binding: FlutterPlugin.FlutterPluginBinding) {
        Pigeon.RoomLiveControllerApi.setup(binding.binaryMessenger, this)
    }

    override fun onDetachedFromEngine(binding: FlutterPlugin.FlutterPluginBinding) {
        Pigeon.RoomLiveControllerApi.setup(binding.binaryMessenger, null)
    }

    override fun startLive(
        roomUuid: String,
        request: Pigeon.RoomLiveRequest,
        result: Pigeon.Result<Pigeon.RoomLiveResponse>
    ) {
        val liveRequest = NERoomLiveRequest(
            title = request.title,
            liveLayout = mapToLiveLayout(request.liveLayout.toInt()),
            password = request.password,
            userUuidList = request.userUuidList,
            extensionConfig = request.extensionConfig
        )
        roomService.getRoomContext(roomUuid)?.liveController?.startLive(
            liveRequest,
            callback = FlutterResultCallback(result) { code: Int, message: String?, info: NERoomLiveInfo? ->
                roomLiveResponse(code, message, info)
            }
        )
    }

    override fun stopLive(roomUuid: String, result: Pigeon.Result<Pigeon.SimpleResponse>) {
        roomService.getRoomContext(roomUuid)?.liveController?.stopLive(result.simpleCallback())
            ?: result.success(roomContextNotFound)
    }

    override fun updateLive(
        roomUuid: String,
        request: Pigeon.RoomLiveRequest,
        result: Pigeon.Result<Pigeon.RoomLiveResponse>
    ) {
        val liveRequest = NERoomLiveRequest(
            title = request.title,
            liveLayout = mapToLiveLayout(request.liveLayout.toInt()),
            password = request.password,
            userUuidList = request.userUuidList,
            extensionConfig = request.extensionConfig
        )
        roomService.getRoomContext(roomUuid)?.liveController?.updateLive(
            liveRequest,
            callback = FlutterResultCallback(result) { code: Int, message: String?, info: NERoomLiveInfo? ->
                roomLiveResponse(code, message, info)
            }
        )
    }

    override fun getLiveInfo(roomUuid: String, result: Pigeon.Result<Pigeon.RoomLiveResponse>) {
        val roomLiveResponse = roomService.getRoomContext(roomUuid)?.liveController?.getLiveInfo()?.let {
            roomLiveResponse(0, "", it)
        }
        result.success(
            roomLiveResponse ?: Pigeon.RoomLiveResponse.Builder()
                .setCode(-1)
                .setMsg("暂不支持直播").build()
        )
    }

    override fun addLiveStreamTask(
        roomUuid: String,
        taskInfo: Pigeon.RoomLiveStreamTaskInfo,
        result: Pigeon.Result<Pigeon.SimpleResponse>?
    ) {
        val liveStreamTaskInfo = pigeonTaskInfo2NERoomLiveStreamTaskInfo(taskInfo)
        roomService.getRoomContext(roomUuid)?.liveController?.addLiveStreamTask(liveStreamTaskInfo, result?.simpleCallback())
    }

    private fun pigeonTaskInfo2NERoomLiveStreamTaskInfo(taskInfo: Pigeon.RoomLiveStreamTaskInfo): NERoomLiveStreamTaskInfo {
        val roomLiveStreamTaskInfo = NERoomLiveStreamTaskInfo()
        taskInfo.config?.let {
            val liveConfig = NERoomLiveConfig()
            liveConfig.audioCodecProfile =
                when (taskInfo.config!!.audioCodecProfile) {
                    1L -> NERoomLiveStreamAudioCodecProfile.ProfileHEAAC
                    else -> NERoomLiveStreamAudioCodecProfile.ProfileLCAAC
                }
            liveConfig.sampleRate =
                when (taskInfo.config!!.channels) {
                    1L -> NERoomLiveStreamAudioSampleRate.Rate44100
                    2L -> NERoomLiveStreamAudioSampleRate.Rate44100
                    else -> NERoomLiveStreamAudioSampleRate.Rate32000
                }
            liveConfig.audioBitrate = taskInfo.config!!.audioBitrate.toInt()
            liveConfig.channels = taskInfo.config!!.channels.toInt()
            liveConfig.singleVideoPassThrough = taskInfo.config!!.singleVideoPassthrough
            roomLiveStreamTaskInfo.config = liveConfig
        }
        taskInfo.layout?.let {
            val liveStreamLayout = NERoomLiveStreamLayout()

            taskInfo.layout!!.bgImage?.let {
                if (!TextUtils.isEmpty(taskInfo.layout!!.bgImage!!.url)) {
                    val liveStreamImageInfo = NERoomLiveStreamImageInfo()
                    liveStreamImageInfo.height = taskInfo.layout!!.bgImage!!.height.toInt()
                    liveStreamImageInfo.width = taskInfo.layout!!.bgImage!!.width.toInt()
                    liveStreamImageInfo.url = taskInfo.layout!!.bgImage!!.url
                    liveStreamImageInfo.x = taskInfo.layout!!.bgImage!!.x.toInt()
                    liveStreamImageInfo.y = taskInfo.layout!!.bgImage!!.y.toInt()
                    liveStreamLayout.backgroundImg = liveStreamImageInfo
                }
            }

            taskInfo.layout!!.users?.let {
                val liveStreamUserTranscodingList = ArrayList<NERoomLiveStreamUserTranscoding>()
                taskInfo.layout!!.users!!.forEach {
                    val liveStreamUserTranscoding = NERoomLiveStreamUserTranscoding()
                    liveStreamUserTranscoding.adaption =
                        when (it.adaption) {
                            1L -> NERoomLiveStreamVideoScaleMode.ScaleCropFill
                            else -> NERoomLiveStreamVideoScaleMode.ScaleFit
                        }
                    liveStreamUserTranscoding.audioPush = it.audioPush
                    liveStreamUserTranscoding.height = it.height.toInt()
                    liveStreamUserTranscoding.width = it.width.toInt()
                    liveStreamUserTranscoding.uuid = it.userUuid
                    liveStreamUserTranscoding.videoPush = it.videoPush
                    liveStreamUserTranscoding.x = it.x.toInt()
                    liveStreamUserTranscoding.y = it.y.toInt()
                    liveStreamUserTranscoding.zOrder = it.zOrder.toInt()

                    liveStreamUserTranscodingList.add(liveStreamUserTranscoding)
                }
                liveStreamLayout.userTranscodingList = liveStreamUserTranscodingList
            }
            liveStreamLayout.backgroundColor = taskInfo.layout!!.backgroundColor.toInt()
            liveStreamLayout.height = taskInfo.layout!!.height.toInt()
            liveStreamLayout.width = taskInfo.layout!!.width.toInt()
            roomLiveStreamTaskInfo.layout = liveStreamLayout
        }
        roomLiveStreamTaskInfo.mode =
            when (taskInfo.lsMode) {
                1L -> NERoomLiveStreamMode.ModeAudio
                else -> NERoomLiveStreamMode.ModeVideo
            }
        roomLiveStreamTaskInfo.extraInfo = taskInfo.extraInfo
        roomLiveStreamTaskInfo.serverRecordEnabled = taskInfo.serverRecordEnabled
        roomLiveStreamTaskInfo.taskId = taskInfo.taskID
        roomLiveStreamTaskInfo.streamUrl = taskInfo.streamURL

        return roomLiveStreamTaskInfo
    }

    override fun updateLiveStreamTask(
        roomUuid: String,
        taskInfo: Pigeon.RoomLiveStreamTaskInfo,
        result: Pigeon.Result<Pigeon.SimpleResponse>?
    ) {
        val liveStreamTaskInfo = pigeonTaskInfo2NERoomLiveStreamTaskInfo(taskInfo)
        roomService.getRoomContext(roomUuid)?.liveController?.updateLiveStreamTask(liveStreamTaskInfo, result?.simpleCallback())
    }

    override fun removeLiveStreamTask(
        roomUuid: String,
        taskID: String,
        result: Pigeon.Result<Pigeon.SimpleResponse>?
    ) {
        roomService.getRoomContext(roomUuid)?.liveController?.removeLiveStreamTask(taskID, result?.simpleCallback())
    }

    private fun roomLiveResponse(
        code: Int,
        message: String?,
        info: NERoomLiveInfo?
    ) = Pigeon.RoomLiveResponse.Builder()
        .setCode(code.toLong())
        .setMsg(message)
        .setLiveInfo(
            info?.let { data ->
                Pigeon.RoomLiveInfo.Builder()
                    .setChatRoomId(data.chatRoomId)
                    .setRtmpPullUrl(data.pullRtmpUrl)
                    .setHlsPullUrl(data.pullHlsUrl)
                    .setHttpPullUrl(data.pullHttpUrl)
                    .setLiveLayout(data.liveLayout.value.toLong())
                    .setPassword(data.password)
                    .setUserUuidList(data.userUuidList)
                    .setExtensionConfig(data.extensionConfig)
                    .setPushUrl(data.pushUrl)
                    .setTaskId(data.taskId)
                    .setState(data.state.value.toLong())
                    .setTitle(data.title)
                    .build()
            }
        ).build()
}

class RoomSeatControllerPlatform : Pigeon.RoomSeatControllerApi, IPlatform {

    private val roomService by lazy {
        NERoomKit.instance.getService(NERoomService::class.java)
    }

    override fun onAttachedToEngine(binding: FlutterPlugin.FlutterPluginBinding) {
        Pigeon.RoomSeatControllerApi.setup(binding.binaryMessenger, this)
    }

    override fun onDetachedFromEngine(binding: FlutterPlugin.FlutterPluginBinding) {
        Pigeon.RoomSeatControllerApi.setup(binding.binaryMessenger, null)
    }

    override fun submitSeatRequest(
        roomUuid: String,
        seatIndex: Long,
        exclusive: Boolean,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.seatController?.submitSeatRequest(seatIndex.toInt(), exclusive, result.simpleCallback())
            ?: result.success(roomContextNotFound)
    }

    override fun cancelSeatRequest(
        roomUuid: String,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.seatController?.cancelSeatRequest(result.simpleCallback())
            ?: result.success(roomContextNotFound)
    }

    override fun leaveSeat(roomUuid: String, result: Pigeon.Result<Pigeon.SimpleResponse>) {
        roomService.getRoomContext(roomUuid)?.seatController?.leaveSeat(result.simpleCallback())
            ?: result.success(roomContextNotFound)
    }

    override fun approveSeatRequest(
        roomUuid: String,
        user: String,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.seatController?.approveSeatRequest(user, result.simpleCallback())
            ?: result.success(roomContextNotFound)
    }

    override fun rejectSeatRequest(
        roomUuid: String,
        user: String,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.seatController?.rejectSeatRequest(user, result.simpleCallback())
            ?: result.success(roomContextNotFound)
    }

    override fun kickSeat(
        roomUuid: String,
        user: String,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.seatController?.kickSeat(user, result.simpleCallback())
            ?: result.success(roomContextNotFound)
    }

    override fun sendSeatInvitation(
        roomUuid: String,
        seatIndex: Long,
        user: String,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.seatController?.sendSeatInvitation(seatIndex.toInt(), user, result.simpleCallback())
            ?: result.success(roomContextNotFound)
    }

    override fun cancelSeatInvitation(
        roomUuid: String,
        user: String,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.seatController?.cancelSeatInvitation(user, result.simpleCallback())
            ?: result.success(roomContextNotFound)
    }

    override fun acceptSeatInvitation(
        roomUuid: String,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.seatController?.acceptSeatInvitation(result.simpleCallback())
            ?: result.success(roomContextNotFound)
    }

    override fun rejectSeatInvitation(
        roomUuid: String,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.seatController?.rejectSeatInvitation(result.simpleCallback())
            ?: result.success(roomContextNotFound)
    }

    override fun closeSeats(
        roomUuid: String,
        seatIndices: MutableList<Long>,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        val list = ArrayList<Int>()
        for (seatIndex in seatIndices) {
            list.add(seatIndex.toInt())
        }
        roomService.getRoomContext(roomUuid)?.seatController?.closeSeats(list, result.simpleCallback())
            ?: result.success(roomContextNotFound)
    }

    override fun openSeats(
        roomUuid: String,
        seatIndices: MutableList<Long>,
        result: Pigeon.Result<Pigeon.SimpleResponse>
    ) {
        val list = ArrayList<Int>()
        for (seatIndex in seatIndices) {
            list.add(seatIndex.toInt())
        }
        roomService.getRoomContext(roomUuid)?.seatController?.openSeats(list, result.simpleCallback())
            ?: result.success(roomContextNotFound)
    }

    override fun getSeatInfo(roomUuid: String, result: Pigeon.Result<Pigeon.SeatInfoResponse>) {
        roomService.getRoomContext(roomUuid)?.seatController?.getSeatInfo(object : NECallback<NESeatInfo> {
            override fun onResult(code: Int, message: String?, data: NESeatInfo?) {
                val seatItems = ArrayList<Pigeon.SeatItem>()
                data?.seatItems?.forEach {
                    seatItems.add(
                        Pigeon.SeatItem.Builder()
                            .setIcon(it.icon)
                            .setIndex(it.index.toLong())
                            .setOnSeatType(it.onSeatType.toLong())
                            .setStatus(it.status.toLong())
                            .setUpdated(it.updated)
                            .setUser(it.user)
                            .setUserName(it.userName).build()
                    )
                }

                val seatInfo = Pigeon.SeatInfo.Builder()
                    .setCreator(data?.creator)
                    .setManagers(data?.managers)
                    .setSeatItems(seatItems)
                    .build()

                result.success(
                    Pigeon.SeatInfoResponse.Builder()
                        .setCode(code.toLong())
                        .setMsg(message)
                        .setSeatInfo(seatInfo).build()
                )
            }
        })
    }

    override fun getSeatRequestList(
        roomUuid: String,
        result: Pigeon.Result<Pigeon.SeatRequestListResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.seatController?.getSeatRequestList(object : NECallback<List<NESeatRequestItem>> {
            override fun onResult(code: Int, message: String?, data: List<NESeatRequestItem>?) {
                val list = ArrayList<Pigeon.SeatRequestItem>()
                data?.forEach {
                    list.add(
                        Pigeon.SeatRequestItem.Builder()
                            .setIcon(it.icon)
                            .setIndex(it.index.toLong())
                            .setUser(it.user)
                            .setUserName(it.userName)
                            .build()
                    )
                }
                result.success(
                    Pigeon.SeatRequestListResponse.Builder()
                        .setCode(code.toLong())
                        .setMsg(message)
                        .setSeatRequestList(list)
                        .build()
                )
            }
        })
    }

    override fun getSeatInvitationList(
        roomUuid: String,
        result: Pigeon.Result<Pigeon.SeatInvitationListResponse>
    ) {
        roomService.getRoomContext(roomUuid)?.seatController?.getSeatInvitationList(object : NECallback<List<NESeatInvitationItem>> {
            override fun onResult(code: Int, message: String?, data: List<NESeatInvitationItem>?) {
                val list = ArrayList<Pigeon.SeatInvitationItem>()
                data?.forEach {
                    list.add(
                        Pigeon.SeatInvitationItem.Builder()
                            .setIndex(it.index.toLong())
                            .setUser(it.user)
                            .build()
                    )
                }
                result.success(
                    Pigeon.SeatInvitationListResponse.Builder()
                        .setCode(code.toLong())
                        .setMsg(message)
                        .setSeatInvitationList(list)
                        .build()
                )
            }
        })
    }
}

class RoomRtcStatsListenerPlatform(
    private val roomUuid: String,
    private val eventSink: Pigeon.RoomStatsSink?
) : NERoomRtcStatsListenerAdapter() {

    private val reply = Pigeon.RoomStatsSink.Reply<Void> { }

    override fun onNetworkQuality(statsArray: Array<NERoomRtcNetworkQualityInfo>) {
        eventSink?.onNetworkQuality(
            roomUuid,
            statsArray.map {
                Pigeon.RoomNetworkQuality
                    .Builder()
                    .setUserId(it.userUuid)
                    .setUpStatus(it.upStatus.toLong())
                    .setDownStatus(it.downStatus.toLong())
                    .build()
            }.toList(),
            reply
        )
    }

    override fun onRtcStats(stats: NERoomRtcStats) {
        eventSink?.onRtcStats(
            roomUuid,
            Pigeon.RoomStatsEvent
                .Builder()
                .setCpuAppUsage(stats.cpuAppUsage.toLong())
                .setCpuTotalUsage(stats.cpuTotalUsage.toLong())
                .setDownRtt(stats.downRtt)
                .setUpRtt(stats.upRtt)
                .setRxBytes(stats.rxBytes)
                .setTxBytes(stats.txBytes)
                .setRxAudioBytes(stats.rxAudioBytes)
                .setTxAudioBytes(stats.txAudioBytes)
                .setRxAudioJitter(stats.rxAudioJitter.toLong())
                .setTxAudioJitter(stats.txAudioJitter.toLong())
                .setRxAudioKBitRate(stats.rxAudioKBitRate.toLong())
                .setTxAudioKBitRate(stats.txAudioKBitRate.toLong())
                .setRxVideoBytes(stats.rxVideoBytes)
                .setTxVideoBytes(stats.txVideoBytes)
                .setRxVideoJitter(stats.rxVideoJitter.toLong())
                .setTxVideoJitter(stats.txVideoJitter.toLong())
                .setRxVideoKBitRate(stats.rxVideoKBitRate.toLong())
                .setTxVideoKBitRate(stats.txVideoKBitRate.toLong())
                .setMemoryAppUsageInKBytes(stats.memoryAppUsageInKBytes)
                .setMemoryAppUsageRatio(stats.memoryAppUsageRatio.toLong())
                .setMemoryTotalUsageRatio(stats.memoryTotalUsageRatio.toLong())
                .setTotalDuration(stats.totalDuration)
                .setRxAudioPacketLossRate(stats.rxAudioPacketLossRate.toLong())
                .setTxAudioPacketLossRate(stats.txAudioPacketLossRate.toLong())
                .setRxAudioPacketLossSum(stats.rxAudioPacketLossSum)
                .setTxAudioPacketLossSum(stats.txAudioPacketLossSum.toLong())
                .setRxVideoPacketLossRate(stats.rxVideoPacketLossRate.toLong())
                .setTxVideoPacketLossRate(stats.txVideoPacketLossRate.toLong())
                .setRxVideoPacketLossSum(stats.rxVideoPacketLossSum.toLong())
                .setTxVideoPacketLossSum(stats.txVideoPacketLossSum.toLong())
                .build(),
            reply
        )
    }
}
