/*
 * Copyright (c) 2022 NetEase, Inc. All rights reserved.
 * Use of this source code is governed by a MIT license that can be
 * found in the LICENSE file.
 */

package com.netease.yunxin.flutter.plugins.roomkit.rtc

import android.graphics.SurfaceTexture
import com.netease.lava.api.IVideoRender
import com.netease.lava.api.IVideoRender.VideoBufferType
import com.netease.lava.webrtc.EglBase
import com.netease.lava.webrtc.VideoFrame
import com.netease.yunxin.lite.video.VideoViewActionListener
import io.flutter.plugin.common.BinaryMessenger
import io.flutter.plugin.common.EventChannel
import io.flutter.plugin.common.EventChannel.EventSink
import io.flutter.view.TextureRegistry.SurfaceTextureEntry
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.GlobalScope
import kotlinx.coroutines.launch

class FlutterVideoRenderer(
    messenger: BinaryMessenger,
    private val entry: SurfaceTextureEntry,
    sharedEglContext: EglBase.Context?
) : IVideoRender, EventChannel.StreamHandler {
    private val id: Long
    private var eventSink: EventSink? = null
    private var surfaceTextureRenderer: SurfaceTextureRenderer?
    private val texture: SurfaceTexture
    private var sharedEglContext: EglBase.Context? = null
    private var isInitializeCalled = false
    private var isInitialized = false
    private val eventChannel: EventChannel
    private var isMirror = false
    private var userRequestedMirror: Boolean? = null

    @Volatile private var disposed = false

    interface RendererEvents {
        fun onFirstFrameRendered()
        fun onFrameResolutionChanged(width: Int, height: Int, rotation: Int)
    }

    private val rendererEvents: RendererEvents = object : RendererEvents {
        override fun onFirstFrameRendered() {
            GlobalScope.launch(Dispatchers.Main) {
                eventSink?.success(
                    mapOf(
                        "id" to id(),
                        "event" to "onFirstFrameRendered"
                    )
                )
            }
        }

        override fun onFrameResolutionChanged(
            width: Int,
            height: Int,
            rotation: Int
        ) {
            GlobalScope.launch(Dispatchers.Main) {
                eventSink?.success(
                    mapOf(
                        "id" to id(),
                        "event" to "onFrameResolutionChanged",
                        "width" to width,
                        "height" to height,
                        "rotation" to rotation
                    )
                )
            }
        }
    }

    internal fun prepareToAttach() = surfaceTextureRenderer?.prepareToAttach()

    fun dispose() {
        disposed = true
        eventChannel.setStreamHandler(null)
        if (surfaceTextureRenderer != null) {
            surfaceTextureRenderer!!.release()
            surfaceTextureRenderer = null
        }
        sharedEglContext = null
        entry.release()
    }

    override fun setScalingType(scalingType: IVideoRender.ScalingType) {}

    override fun setVideoBufferType(videoBufferType: VideoBufferType) {
    }
    override fun getVideoBufferType(): VideoBufferType {
        return VideoBufferType.VIDEO_BUFFER_TYPE_RAW_DATA
    }

    // SDK 会调用该方法，如果不重载，运行时可能找不到该方法导致画布设置失败出现黑屏
    override fun setViewActionListener(viewActionListener: VideoViewActionListener?, type: Int) {
        // keep this method
    }

    override fun setExternalRender(external: Boolean) {
    }
    override fun isExternalRender(): Boolean {
        return true
    }

    internal fun setRequestedMirror(mirror: Boolean) {
        userRequestedMirror = mirror
        this.isMirror = mirror
        surfaceTextureRenderer?.setMirror(mirror)
    }

    // RTC 在 setupLocalVideoCanvas 时会调用该方法，把之前设置的值给覆盖掉
    // Flutter 在设置时不走这个方法，走 [setRequestedMirror]
    override fun setMirror(mirror: Boolean) {
        if (userRequestedMirror != null) {
            return
        }
        setRequestedMirror(mirror)
    }

    override fun isMirror() = isMirror
    override fun clearImage() {
        surfaceTextureRenderer?.clearImage()
    }

    override fun onFrame(videoFrame: VideoFrame) {
        ensureRenderInitialized()
        if (isInitialized) {
            surfaceTextureRenderer?.onFrame(videoFrame)
        }
    }

    private fun ensureRenderInitialized() {
        if (!isInitializeCalled) {
            isInitializeCalled = true
            GlobalScope.launch(Dispatchers.Main.immediate) {
                if (disposed) return@launch
                surfaceTextureRenderer?.init(sharedEglContext, rendererEvents)
                surfaceTextureRenderer?.surfaceCreated(texture)
                isInitialized = true
            }
        }
    }

    fun id(): Long = id

    init {
        id = entry.id()
        texture = entry.surfaceTexture()
        surfaceTextureRenderer = SurfaceTextureRenderer("SurfaceTextureRenderer/$id")
        eventChannel = EventChannel(messenger, "NERoomKitFlutterVideoRenderer/Texture$id")
            .also { it.setStreamHandler(this) }
        if (sharedEglContext != null) this.sharedEglContext = sharedEglContext
    }

    override fun onListen(arguments: Any?, events: EventSink) {
        eventSink = events
    }

    override fun onCancel(arguments: Any?) {
        eventSink = null
    }
}
