/*
 * Copyright (c) 2022 NetEase, Inc. All rights reserved.
 * Use of this source code is governed by a MIT license that can be
 * found in the LICENSE file.
 */

package com.netease.yunxin.flutter.plugins.roomkit.rtc

import android.graphics.SurfaceTexture
import com.netease.lava.webrtc.EglBase
import com.netease.lava.webrtc.EglRenderer
import com.netease.lava.webrtc.GlRectDrawer
import com.netease.lava.webrtc.VideoFrame
import java.util.concurrent.CountDownLatch

class SurfaceTextureRenderer(name: String?) : EglRenderer(name) {
    private val layoutLock = Any()
    private var isFirstFrameRendered = false
    private var rotatedFrameWidth = 0
    private var rotatedFrameHeight = 0
    private var frameRotation = 0
    private var rendererEvents: FlutterVideoRenderer.RendererEvents? = null
    private var texture: SurfaceTexture? = null
    fun init(
        sharedContext: EglBase.Context?,
        rendererEvents: FlutterVideoRenderer.RendererEvents?
    ) {
        com.netease.lava.webrtc.ThreadUtils.checkIsOnMainThread()
        this.rendererEvents = rendererEvents
        prepareToAttach()
        super.init(sharedContext, EglBase.CONFIG_PLAIN, GlRectDrawer())
    }

    internal fun prepareToAttach() = synchronized(layoutLock) {
        isFirstFrameRendered = false
        rotatedFrameWidth = 0
        rotatedFrameHeight = 0
        frameRotation = 0
    }

    override fun onFrame(frame: VideoFrame) {
        updateFrameDimensionsAndReportEvents(frame)
        super.onFrame(frame)
    }

    override fun release() {
        this.rendererEvents = null
        super.release()
    }

    fun surfaceCreated(texture: SurfaceTexture?) {
        com.netease.lava.webrtc.ThreadUtils.checkIsOnMainThread()
        this.texture = texture
        createEglSurface(texture)
    }

    fun surfaceDestroyed() {
        com.netease.lava.webrtc.ThreadUtils.checkIsOnMainThread()
        val completionLatch = CountDownLatch(1)
        releaseEglSurface(Runnable { completionLatch.countDown() })
        com.netease.lava.webrtc.ThreadUtils.awaitUninterruptibly(completionLatch)
    }

    private fun updateFrameDimensionsAndReportEvents(frame: VideoFrame) {
        synchronized(layoutLock) {
            if (!isFirstFrameRendered) {
                isFirstFrameRendered = true
                rendererEvents?.onFirstFrameRendered()
            }
            if (rotatedFrameWidth != frame.rotatedWidth || rotatedFrameHeight != frame.rotatedHeight || frameRotation != frame.rotation
            ) {
                rendererEvents?.onFrameResolutionChanged(
                    frame.buffer.width,
                    frame.buffer.height,
                    frame.rotation
                )
                rotatedFrameWidth = frame.rotatedWidth
                rotatedFrameHeight = frame.rotatedHeight
                texture!!.setDefaultBufferSize(rotatedFrameWidth, rotatedFrameHeight)
                frameRotation = frame.rotation
            }
        }
    }
}
