// Copyright (c) 2022 NetEase, Inc. All rights reserved.
// Use of this source code is governed by a MIT license that can be
// found in the LICENSE file.

import Foundation
import NERoomKit

class FLTRoomChatController: NSObject, FLTRoomChatControllerApi {
  func fetchChatRoomMembersRoomUuid(_ roomUuid: String?, queryType: NSNumber?, limit: NSNumber?,
                                    completion: @escaping (FLTChatRoomMembersResponse?,
                                                           FlutterError?) -> Void) {
    guard let roomUuid = roomUuid,
          let roomContext = NERoomKit.shared().roomService.getRoomContext(roomUuid: roomUuid),
          let type = queryType?.intValue,
          let limit = limit?.intValue else {
      let simpleResponse = FLTChatRoomMembersResponse.make(
        withCode: NSNumber(value: -1),
        msg: paramsError,
        chatRoomMembers: []
      )
      completion(simpleResponse, nil)
      return
    }
    var qType = NEChatroomMemberQueryType.normal
    switch type {
    case 1:
      qType = .onlineNormal
    case 2:
      qType = .guestDesc
    case 3:
      qType = .guestAsc
    default:
      qType = .normal
    }
    roomContext.chatController
      .fetchChatroomMembers(type: qType, limit: limit) { code, msg, members in
        let simpleResponse = FLTChatRoomMembersResponse.make(
          withCode: NSNumber(value: code),
          msg: msg,
          chatRoomMembers: members?.map { m in
            FLTChatRoomMember.make(withNick: m.nick, avatar: m.avatar)
          } ?? []
        )
        completion(simpleResponse, nil)
      }
  }

  func joinChatroomRoomUuid(_ roomUuid: String?,
                            completion: @escaping (FLTSimpleResponse?, FlutterError?) -> Void) {
    guard let roomUuid = roomUuid,
          let roomContext = NERoomKit.shared().roomService.getRoomContext(roomUuid: roomUuid)
    else {
      let simpleResponse = FLTSimpleResponse.make(
        withCode: NSNumber(value: -1),
        msg: paramsError,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(simpleResponse, nil)
      return
    }
    roomContext.chatController.joinChatroom { code, str, _ in
      let response = FLTSimpleResponse.make(
        withCode: NSNumber(value: code),
        msg: str,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(response, nil)
    }
  }

  func leaveChatroomRoomUuid(_ roomUuid: String?,
                             completion: @escaping (FLTSimpleResponse?, FlutterError?) -> Void) {
    guard let roomUuid = roomUuid,
          let roomContext = NERoomKit.shared().roomService.getRoomContext(roomUuid: roomUuid)
    else {
      let simpleResponse = FLTSimpleResponse.make(
        withCode: NSNumber(value: -1),
        msg: paramsError,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(simpleResponse, nil)
      return
    }
    roomContext.chatController.leaveChatroom { code, str, _ in
      let response = FLTSimpleResponse.make(
        withCode: NSNumber(value: code),
        msg: str,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(response, nil)
    }
  }

  func sendBroadcastTextMessageRoomUuid(_ roomUuid: String?,
                                        message: String?,
                                        completion: @escaping (FLTSimpleResponse?, FlutterError?)
                                          -> Void) {
    guard let roomUuid = roomUuid,
          let roomContext = NERoomKit.shared().roomService.getRoomContext(roomUuid: roomUuid),
          let message = message else {
      let simpleResponse = FLTSimpleResponse.make(
        withCode: NSNumber(value: -1),
        msg: paramsError,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(simpleResponse, nil)
      return
    }
    roomContext.chatController.sendBroadcastTextMessage(message: message) { code, str, _ in
      let response = FLTSimpleResponse.make(
        withCode: NSNumber(value: code),
        msg: str,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(response, nil)
    }
  }

  func sendDirectTextMessageRoomUuid(_ roomUuid: String?,
                                     userUuid: String?,
                                     message: String?,
                                     completion: @escaping (FLTSimpleResponse?, FlutterError?)
                                       -> Void) {
    guard let roomUuid = roomUuid,
          let roomContext = NERoomKit.shared().roomService.getRoomContext(roomUuid: roomUuid),
          let userUuid = userUuid,
          let message = message else {
      let simpleResponse = FLTSimpleResponse.make(
        withCode: NSNumber(value: -1),
        msg: paramsError,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(simpleResponse, nil)
      return
    }
    roomContext.chatController.sendDirectTextMessage(userUuid: userUuid,
                                                     message: message) { code, str, _ in
      let response = FLTSimpleResponse.make(
        withCode: NSNumber(value: code),
        msg: str,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(response, nil)
    }
  }

  func sendGroupTextMessageRoomUuid(_ roomUuid: String?,
                                    userUuids: [String]?,
                                    message: String?,
                                    completion: @escaping (FLTSimpleResponse?, FlutterError?)
                                      -> Void) {
    guard let roomUuid = roomUuid,
          let roomContext = NERoomKit.shared().roomService.getRoomContext(roomUuid: roomUuid),
          let userUuids = userUuids,
          let message = message else {
      let simpleResponse = FLTSimpleResponse.make(
        withCode: NSNumber(value: -1),
        msg: paramsError,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(simpleResponse, nil)
      return
    }

    roomContext.chatController.sendGroupTextMessage(userUuids: userUuids,
                                                    message: message) { code, str, _ in
      let response = FLTSimpleResponse.make(
        withCode: NSNumber(value: code),
        msg: str,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(response, nil)
    }
  }

  func sendFileMessageRoomUuid(_ roomUuid: String?, messageUuid: String?, filePath: String?,
                               userUuids: [String]?,
                               completion: @escaping (FLTSimpleResponse?, FlutterError?)
                                 -> Void) {
    guard let roomUuid = roomUuid,
          let roomContext = NERoomKit.shared().roomService.getRoomContext(roomUuid: roomUuid),
          let messageUuid = messageUuid,
          let filePath = filePath else {
      let simpleResponse = FLTSimpleResponse.make(
        withCode: NSNumber(value: -1),
        msg: paramsError,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(simpleResponse, nil)
      return
    }
    roomContext.chatController.sendFileMessage(messageUuid,
                                               filePath: filePath,
                                               userUuids: userUuids) { code, str, _ in
      let response = FLTSimpleResponse.make(
        withCode: NSNumber(value: code),
        msg: str,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(response, nil)
    }
  }

  func sendImageMessageRoomUuid(_ roomUuid: String?, messageUuid: String?, imagePath: String?,
                                userUuids: [String]?,
                                completion: @escaping (FLTSimpleResponse?, FlutterError?)
                                  -> Void) {
    guard let roomUuid = roomUuid,
          let roomContext = NERoomKit.shared().roomService.getRoomContext(roomUuid: roomUuid),
          let messageUuid = messageUuid,
          let imagePath = imagePath else {
      let simpleResponse = FLTSimpleResponse.make(
        withCode: NSNumber(value: -1),
        msg: paramsError,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(simpleResponse, nil)
      return
    }
    roomContext.chatController.sendImageMessage(messageUuid,
                                                imagePath: imagePath,
                                                userUuids: userUuids) { code, str, _ in
      let response = FLTSimpleResponse.make(
        withCode: NSNumber(value: code),
        msg: str,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(response, nil)
    }
  }

  func downloadAttachmentRoomUuid(_ roomUuid: String?, messageUuid: String?,
                                  completion: @escaping (FLTSimpleResponse?, FlutterError?)
                                    -> Void) {
    guard let roomUuid = roomUuid,
          let roomContext = NERoomKit.shared().roomService.getRoomContext(roomUuid: roomUuid),
          let messageUuid = messageUuid else {
      let simpleResponse = FLTSimpleResponse.make(
        withCode: NSNumber(value: -1),
        msg: paramsError,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(simpleResponse, nil)
      return
    }

    roomContext.chatController.downloadAttachment(messageUuid) { code, str, _ in
      let response = FLTSimpleResponse.make(
        withCode: NSNumber(value: code),
        msg: str,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(response, nil)
    }
  }

  func cancelDownloadAttachmentRoomUuid(_ roomUuid: String?, messageUuid: String?,
                                        completion: @escaping (FLTSimpleResponse?, FlutterError?)
                                          -> Void) {
    guard let roomUuid = roomUuid,
          let roomContext = NERoomKit.shared().roomService.getRoomContext(roomUuid: roomUuid),
          let messageUuid = messageUuid else {
      let simpleResponse = FLTSimpleResponse.make(
        withCode: NSNumber(value: -1),
        msg: paramsError,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(simpleResponse, nil)
      return
    }
    roomContext.chatController.cancelDownloadAttachment(messageUuid) { code, str, _ in
      let response = FLTSimpleResponse.make(
        withCode: NSNumber(value: code),
        msg: str,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(response, nil)
    }
  }
}

extension FLTRoomChatController: FLTAssociativeWrapper {
  func onAttachedToEngine(_ registrar: FlutterPluginRegistrar) {
    FLTRoomChatControllerApiSetup(registrar.messenger(), self)
  }

  func onDetachedFromEngine(_ registrar: FlutterPluginRegistrar) {
    FLTRoomChatControllerApiSetup(registrar.messenger(), nil)
  }
}
