// Copyright (c) 2022 NetEase, Inc. All rights reserved.
// Use of this source code is governed by a MIT license that can be
// found in the LICENSE file.

import Foundation
import NERoomKit

extension NERoomLiveInfo {
  func conversionLiveInfo() -> FLTRoomLiveInfo {
    FLTRoomLiveInfo.make(withTaskId: taskId,
                         pushUrl: pushUrl ?? "",
                         chatRoomId: chatRoomId ?? "",
                         httpPullUrl: pullHttpUrl ?? "",
                         rtmpPullUrl: pullRtmpUrl ?? "",
                         hlsPullUrl: pullHlsUrl ?? "",
                         title: title,
                         liveLayout: NSNumber(value: liveLayout.rawValue),
                         password: password,
                         userUuidList: userUuidList,
                         state: NSNumber(value: state.rawValue),
                         extensionConfig: extensionConfig)
  }
}

class FLTRoomLiveController: NSObject, FLTRoomLiveControllerApi {
  func startLiveRoomUuid(_ roomUuid: String?,
                         request: FLTRoomLiveRequest?,
                         completion: @escaping (FLTRoomLiveResponse?, FlutterError?) -> Void) {
    guard let roomUuid = roomUuid,
          let roomContext = NERoomKit.shared().roomService.getRoomContext(roomUuid: roomUuid),
          let request = request else {
      let liveResponse = FLTRoomLiveResponse.make(
        withCode: NSNumber(value: -1),
        msg: paramsError,
        liveInfo: nil
      )
      completion(liveResponse, nil)
      return
    }
    let liveRequest = NERoomLiveRequest()
    liveRequest.userUuidList = request.userUuidList
    liveRequest.title = request.title
    liveRequest.liveLayout = NERoomLiveLayout(rawValue: request.liveLayout.intValue) ?? .gallery
    liveRequest.password = request.password
    liveRequest.extensionConfig = request.extensionConfig
    roomContext.liveController.startLive(request: liveRequest) { code, str, liveInfo in
      let liveResponse = FLTRoomLiveResponse.make(
        withCode: NSNumber(value: code),
        msg: str,
        liveInfo: liveInfo?.conversionLiveInfo()
      )
      completion(liveResponse, nil)
    }
  }

  func stopLiveRoomUuid(_ roomUuid: String?,
                        completion: @escaping (FLTSimpleResponse?, FlutterError?) -> Void) {
    guard let roomUuid = roomUuid,
          let roomContext = NERoomKit.shared().roomService.getRoomContext(roomUuid: roomUuid)
    else {
      let simpleResponse = FLTSimpleResponse.make(
        withCode: NSNumber(value: -1),
        msg: paramsError,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(simpleResponse, nil)
      return
    }
    roomContext.liveController.stopLive { code, str, _ in
      let simpleResponse = FLTSimpleResponse.make(
        withCode: NSNumber(value: code),
        msg: str,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(simpleResponse, nil)
    }
  }

  func updateLiveRoomUuid(_ roomUuid: String?,
                          request: FLTRoomLiveRequest?,
                          completion: @escaping (FLTRoomLiveResponse?, FlutterError?) -> Void) {
    guard let roomUuid = roomUuid,
          let roomContext = NERoomKit.shared().roomService.getRoomContext(roomUuid: roomUuid),
          let request = request else {
      let liveResponse = FLTRoomLiveResponse.make(
        withCode: NSNumber(value: -1),
        msg: paramsError,
        liveInfo: nil
      )
      completion(liveResponse, nil)
      return
    }
    let liveRequest = NERoomLiveRequest()
    liveRequest.userUuidList = request.userUuidList
    liveRequest.title = request.title
    liveRequest.liveLayout = NERoomLiveLayout(rawValue: request.liveLayout.intValue) ?? .gallery
    liveRequest.password = request.password
    liveRequest.extensionConfig = request.extensionConfig
    roomContext.liveController.updateLive(request: liveRequest) { code, str, liveInfo in
      let liveResponse = FLTRoomLiveResponse.make(
        withCode: NSNumber(value: code),
        msg: str,
        liveInfo: liveInfo?.conversionLiveInfo()
      )
      completion(liveResponse, nil)
    }
  }

  func getLiveInfoRoomUuid(_ roomUuid: String?,
                           completion: @escaping (FLTRoomLiveResponse?, FlutterError?) -> Void) {
    guard let roomUuid = roomUuid,
          let roomContext = NERoomKit.shared().roomService.getRoomContext(roomUuid: roomUuid)
    else {
      let liveResponse = FLTRoomLiveResponse.make(
        withCode: NSNumber(value: -1),
        msg: paramsError,
        liveInfo: nil
      )
      completion(liveResponse, nil)
      return
    }
    let liveInfo = roomContext.liveController.getLiveInfo()?.conversionLiveInfo()
    var num: Int = -1
    if let _ = liveInfo { num = 0 }
    let liveResponse = FLTRoomLiveResponse.make(
      withCode: NSNumber(value: num),
      msg: nil,
      liveInfo: liveInfo
    )
    completion(liveResponse, nil)
  }

  func addLiveStreamTaskRoomUuid(_ roomUuid: String?, taskInfo: FLTRoomLiveStreamTaskInfo?,
                                 completion: @escaping (FLTSimpleResponse?, FlutterError?)
                                   -> Void) {
    guard let roomUuid = roomUuid,
          let roomContext = NERoomKit.shared().roomService.getRoomContext(roomUuid: roomUuid),
          let taskInfo = taskInfo else {
      let simpleResponse = FLTSimpleResponse.make(
        withCode: NSNumber(value: -1),
        msg: paramsError,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(simpleResponse, nil)
      return
    }
    roomContext.liveController
      .addLiveStreamTask(task: _convertTaskInfo(info: taskInfo)) { code, str, _ in
        let response = FLTSimpleResponse.make(
          withCode: NSNumber(value: code),
          msg: nil,
          data1: nil,
          data2: nil,
          data3: nil
        )
        completion(response, nil)
      }
  }

  func updateLiveStreamTaskRoomUuid(_ roomUuid: String?, taskInfo: FLTRoomLiveStreamTaskInfo?,
                                    completion: @escaping (FLTSimpleResponse?, FlutterError?)
                                      -> Void) {
    guard let roomUuid = roomUuid,
          let roomContext = NERoomKit.shared().roomService.getRoomContext(roomUuid: roomUuid),
          let taskInfo = taskInfo else {
      let simpleResponse = FLTSimpleResponse.make(
        withCode: NSNumber(value: -1),
        msg: paramsError,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(simpleResponse, nil)
      return
    }
    roomContext.liveController
      .updateLiveStreamTask(task: _convertTaskInfo(info: taskInfo)) { code, str, _ in
        let response = FLTSimpleResponse.make(
          withCode: NSNumber(value: code),
          msg: nil,
          data1: nil,
          data2: nil,
          data3: nil
        )
        completion(response, nil)
      }
  }

  func removeLiveStreamTaskRoomUuid(_ roomUuid: String?, taskID: String?,
                                    completion: @escaping (FLTSimpleResponse?, FlutterError?)
                                      -> Void) {
    guard let roomUuid = roomUuid,
          let roomContext = NERoomKit.shared().roomService.getRoomContext(roomUuid: roomUuid),
          let taskID = taskID else {
      let simpleResponse = FLTSimpleResponse.make(
        withCode: NSNumber(value: -1),
        msg: paramsError,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(simpleResponse, nil)
      return
    }
    roomContext.liveController.removeLiveStreamTask(taskId: taskID) { code, str, _ in
      let response = FLTSimpleResponse.make(
        withCode: NSNumber(value: code),
        msg: nil,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(response, nil)
    }
  }
}

extension FLTRoomLiveController: FLTAssociativeWrapper {
  func onAttachedToEngine(_ registrar: FlutterPluginRegistrar) {
    FLTRoomLiveControllerApiSetup(registrar.messenger(), self)
  }

  func onDetachedFromEngine(_ registrar: FlutterPluginRegistrar) {
    FLTRoomLiveControllerApiSetup(registrar.messenger(), nil)
  }
}

extension FLTRoomLiveController {
  func _convertTaskInfo(info: FLTRoomLiveStreamTaskInfo) -> NERoomLiveStreamTaskInfo {
    let taskInfo = NERoomLiveStreamTaskInfo()
    taskInfo.taskId = info.taskID
    taskInfo.streamUrl = info.streamURL
    taskInfo.serverRecordEnabled = info.serverRecordEnabled.boolValue
    taskInfo.mode = NERoomLiveStreamMode(rawValue: UInt(info.lsMode.intValue)) ?? .video
    taskInfo.layout = _convertLayout(layout: info.layout)
    taskInfo.config = _convertConfig(config: info.config)
    taskInfo.extraInfo = info.extraInfo
    return taskInfo
  }

  func _convertConfig(config: FLTRoomLiveConfig?) -> NERoomLiveConfig? {
    if let config = config {
      let liveConfig = NERoomLiveConfig()
      liveConfig.singleVideoPassthrough = config.singleVideoPassthrough.boolValue
      liveConfig.audioBitrate = config.audioBitrate.intValue
      liveConfig
        .sampleRate =
        NERoomLiveStreamAudioSampleRate(rawValue: UInt(config.sampleRate.intValue)) ?? .K48
      liveConfig.channels = config.channels.intValue
      liveConfig
        .audioCodecProfile =
        NERoomLiveStreamAudioCodecProfile(
          rawValue: UInt(config.audioCodecProfile.intValue)
        ) ??
        .LCAAC
      return liveConfig
    }
    return nil
  }

  func _convertLayout(layout: FLTRoomLiveStreamLayout?) -> NERoomLiveStreamLayout? {
    if let layout = layout {
      let streamLayout = NERoomLiveStreamLayout()
      streamLayout.width = layout.width.intValue
      streamLayout.height = layout.height.intValue
      streamLayout.backgroundColor = layout.backgroundColor.uintValue
      streamLayout.userTranscodingList = layout.users?.compactMap { transcoding in
        _convertTranscoding(transcoding: transcoding)
      }
      streamLayout.backgroundImg = _convertImage(image: layout.bgImage)
      return streamLayout
    }
    return nil
  }

  func _convertTranscoding(transcoding: FLTRoomLiveStreamUserTranscoding?)
    -> NERoomLiveStreamUserTranscoding? {
    if let transcoding = transcoding {
      let userTranscoding = NERoomLiveStreamUserTranscoding()
      userTranscoding.userUuid = transcoding.userUuid
      userTranscoding.videoPush = transcoding.videoPush.boolValue
      userTranscoding.x = transcoding.x.intValue
      userTranscoding.y = transcoding.y.intValue
      userTranscoding.width = transcoding.width.intValue
      userTranscoding.height = transcoding.height.intValue
      userTranscoding.audioPush = transcoding.audioPush.boolValue
      userTranscoding.zOrder = transcoding.zOrder.intValue
      userTranscoding
        .adaption =
        NERoomLiveStreamVideoScaleMode(rawValue: UInt(transcoding.adaption.intValue)) ??
        .cropFill
      return userTranscoding
    }
    return nil
  }

  func _convertImage(image: FLTRoomLiveStreamImageInfo?) -> NERoomLiveStreamImageInfo? {
    if let image = image {
      let imageInfo = NERoomLiveStreamImageInfo()
      imageInfo.url = image.url
      imageInfo.x = image.x.intValue
      imageInfo.y = image.y.intValue
      imageInfo.width = image.width.intValue
      imageInfo.height = image.height.intValue
      return imageInfo
    }
    return nil
  }
}
