// Copyright (c) 2022 NetEase, Inc. All rights reserved.
// Use of this source code is governed by a MIT license that can be
// found in the LICENSE file.

import Foundation
import NERoomKit
import NERtcSDK

class FLTRoomRtcRender: NSObject, FLTRoomRtcVideoRender {
  var textureRegistry: FlutterTextureRegistry?
  var messenger: FlutterBinaryMessenger?

  var renderers = [Int64: FLTRenderEntry]()

  func createVideoRendererRoomUuid(_ roomUuid: String?,
                                   completion: @escaping (FLTSimpleResponse?, FlutterError?)
                                     -> Void) {
    guard let _ = textureRegistry,
          let _ = messenger,
          let roomUuid = roomUuid else {
      let simpleResponse = FLTSimpleResponse.make(
        withCode: NSNumber(value: -1),
        msg: paramsError,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(simpleResponse, nil)
      return
    }

    let videoRenderer = FLTRoomVideoRenderer(
      textureRegistry: textureRegistry!,
      messenger: messenger!
    )
    renderers[videoRenderer.textureId] = FLTRenderEntry(roomUuid, videoRender: videoRenderer)

    let simpleResponse = FLTSimpleResponse.make(
      withCode: NSNumber(value: videoRenderer.textureId),
      msg: nil,
      data1: nil,
      data2: nil,
      data3: nil
    )
    completion(simpleResponse, nil)
  }

  func setupLocalVideoRendererTextureId(_ textureId: NSNumber?,
                                        completion: @escaping (FLTSimpleResponse?, FlutterError?)
                                          -> Void) {
    guard let textureId = textureId,
          let renderEntry = renderers[textureId.int64Value] else {
      let simpleResponse = FLTSimpleResponse.make(
        withCode: NSNumber(value: -1),
        msg: paramsError,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(simpleResponse, nil)
      return
    }

    renderEntry.videoRender.setFirstRender(false)
    let canvas = NERoomVideoView()
    canvas.useExternalRender = true
    canvas.externalVideoRender = renderEntry.videoRender

    if renderEntry.roomUuid.isEmpty {
      let previewContext = NERoomKit.shared().roomService.getPreviewRoomContext()
      guard let context = previewContext else {
        let response = FLTSimpleResponse.make(
          withCode: NSNumber(value: -1),
          msg: nil,
          data1: nil,
          data2: nil,
          data3: nil
        )
        completion(response, nil)
        return
      }
      let code = context.previewController.setupLocalVideoCanvas(videoView: canvas)
      let response = FLTSimpleResponse.make(
        withCode: NSNumber(value: code),
        msg: nil,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(response, nil)
    } else {
      let roomContext = NERoomKit.shared().roomService
        .getRoomContext(roomUuid: renderEntry.roomUuid)
      guard let context = roomContext else {
        let response = FLTSimpleResponse.make(
          withCode: NSNumber(value: -1),
          msg: nil,
          data1: nil,
          data2: nil,
          data3: nil
        )
        completion(response, nil)
        return
      }
      let code = context.rtcController.setupLocalVideoCanvas(videoView: canvas)
      let response = FLTSimpleResponse.make(
        withCode: NSNumber(value: code),
        msg: nil,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(response, nil)
    }
  }

  func setMirrorTextureId(_ textureId: NSNumber?, mirror: NSNumber?,
                          completion: @escaping (FLTSimpleResponse?, FlutterError?) -> Void) {
    guard let textureId = textureId,
          let mirror = mirror,
          let renderEntry = renderers[textureId.int64Value] else {
      let simpleResponse = FLTSimpleResponse.make(
        withCode: NSNumber(value: -1),
        msg: paramsError,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(simpleResponse, nil)
      return
    }
    renderEntry.videoRender.setMirror(mirror.boolValue)
    let simpleResponse = FLTSimpleResponse.make(
      withCode: NSNumber(value: 0),
      msg: nil,
      data1: nil,
      data2: nil,
      data3: nil
    )
    completion(simpleResponse, nil)
  }

  func setupRemoteSubStreamVideoRendererUserUuid(_ userUuid: String?,
                                                 textureId: NSNumber?,
                                                 completion: @escaping (FLTSimpleResponse?,
                                                                        FlutterError?) -> Void) {
    guard let textureId = textureId,
          let userUuid = userUuid,
          let renderEntry = renderers[textureId.int64Value] else {
      let simpleResponse = FLTSimpleResponse.make(
        withCode: NSNumber(value: -1),
        msg: paramsError,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(simpleResponse, nil)
      return
    }
    renderEntry.videoRender.setFirstRender(false)

    let canvas = NERoomVideoView()
    canvas.useExternalRender = true
    canvas.externalVideoRender = renderEntry.videoRender
    guard let context = NERoomKit.shared().roomService
      .getRoomContext(roomUuid: renderEntry.roomUuid) else {
      let response = FLTSimpleResponse.make(
        withCode: NSNumber(value: -1),
        msg: paramsError,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(response, nil)
      return
    }
    let code = context.rtcController.setupRemoteVideoSubStreamCanvas(
      videoView: canvas,
      userUuid: userUuid
    )
    let response = FLTSimpleResponse.make(
      withCode: NSNumber(value: code),
      msg: nil,
      data1: nil,
      data2: nil,
      data3: nil
    )
    completion(response, nil)
  }

  func setupRemoteVideoRendererUserUuid(_ userUuid: String?,
                                        textureId: NSNumber?,
                                        completion: @escaping (FLTSimpleResponse?, FlutterError?)
                                          -> Void) {
    guard let textureId = textureId,
          let userUuid = userUuid,
          let renderEntry = renderers[textureId.int64Value],
          let roomContext = NERoomKit.shared().roomService
          .getRoomContext(roomUuid: renderEntry.roomUuid) else {
      let simpleResponse = FLTSimpleResponse.make(
        withCode: NSNumber(value: -1),
        msg: paramsError,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(simpleResponse, nil)
      return
    }
    renderEntry.videoRender.setFirstRender(false)
    let canvas = NERoomVideoView()
    canvas.useExternalRender = true
    canvas.externalVideoRender = renderEntry.videoRender
    let code = roomContext.rtcController.setupRemoteVideoCanvas(
      videoView: canvas,
      userUuid: userUuid
    )
    let response = FLTSimpleResponse.make(
      withCode: NSNumber(value: code),
      msg: nil,
      data1: nil,
      data2: nil,
      data3: nil
    )
    completion(response, nil)
  }

  func detachLocalVideoRendererSubStream(_ subStream: NSNumber?,
                                         textureId: NSNumber?,
                                         completion: @escaping (FLTSimpleResponse?,
                                                                FlutterError?) -> Void) {
    guard let textureId = textureId,
          let subStream = subStream,
          let renderEntry = renderers[textureId.int64Value] else {
      let simpleResponse = FLTSimpleResponse.make(
        withCode: NSNumber(value: -1),
        msg: paramsError,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(simpleResponse, nil)
      return
    }
    if subStream.boolValue {
      let response = FLTSimpleResponse.make(
        withCode: NSNumber(value: -1),
        msg: "Not support",
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(response, nil)
      return
    }
    if renderEntry.roomUuid.isEmpty { // preview
      guard let context = NERoomKit.shared().roomService.getPreviewRoomContext() else {
        let response = FLTSimpleResponse.make(
          withCode: NSNumber(value: -1),
          msg: paramsError,
          data1: nil,
          data2: nil,
          data3: nil
        )
        completion(response, nil)
        return
      }
      context.previewController.stopPreview()
      let code = context.previewController.setupLocalVideoCanvas(videoView: nil)
      let response = FLTSimpleResponse.make(
        withCode: NSNumber(value: code),
        msg: nil,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(response, nil)
    } else {
      guard let rtcController = NERoomKit.shared().roomService
        .getRoomContext(roomUuid: renderEntry.roomUuid)?.rtcController else {
        let response = FLTSimpleResponse.make(
          withCode: NSNumber(value: -1),
          msg: paramsError,
          data1: nil,
          data2: nil,
          data3: nil
        )
        completion(response, nil)
        return
      }
      let code = rtcController.setupLocalVideoCanvas(videoView: nil)
      let response = FLTSimpleResponse.make(
        withCode: NSNumber(value: code),
        msg: nil,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(response, nil)
    }
  }

  func detachRemoteVideoRendererUserUuid(_ userUuid: String?,
                                         subStream: NSNumber?,
                                         textureId: NSNumber?,
                                         completion: @escaping (FLTSimpleResponse?,
                                                                FlutterError?) -> Void) {
    guard let textureId = textureId,
          let userUuid = userUuid,
          let subStream = subStream,
          let renderEntry = renderers[textureId.int64Value] else {
      let simpleResponse = FLTSimpleResponse.make(
        withCode: NSNumber(value: -1),
        msg: paramsError,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(simpleResponse, nil)
      return
    }
    guard let roomContext = NERoomKit.shared().roomService
      .getRoomContext(roomUuid: renderEntry.roomUuid) else {
      let response = FLTSimpleResponse.make(
        withCode: NSNumber(value: -1),
        msg: paramsError,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(response, nil)
      return
    }

    if subStream.boolValue {
      let code = roomContext.rtcController.setupRemoteVideoSubStreamCanvas(
        videoView: nil,
        userUuid: userUuid
      )
      let response = FLTSimpleResponse.make(
        withCode: NSNumber(value: code),
        msg: nil,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(response, nil)
    } else {
      let code = roomContext.rtcController.setupRemoteVideoCanvas(videoView: nil,
                                                                  userUuid: userUuid)
      let response = FLTSimpleResponse.make(
        withCode: NSNumber(value: code),
        msg: nil,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(response, nil)
    }
  }

  func disposeVideoRendererTextureId(_ textureId: NSNumber?,
                                     completion: @escaping (FLTSimpleResponse?, FlutterError?)
                                       -> Void) {
    guard let textureId = textureId,
          let renderEntry = renderers[textureId.int64Value] else {
      let simpleResponse = FLTSimpleResponse.make(
        withCode: NSNumber(value: -1),
        msg: paramsError,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(simpleResponse, nil)
      return
    }
    renderEntry.videoRender.dispose()
    renderers.removeValue(forKey: textureId.int64Value)
    let response = FLTSimpleResponse.make(
      withCode: NSNumber(value: 0),
      msg: nil,
      data1: nil,
      data2: nil,
      data3: nil
    )
    completion(response, nil)
  }
}

extension FLTRoomRtcRender: FLTAssociativeWrapper {
  func onAttachedToEngine(_ registrar: FlutterPluginRegistrar) {
    FLTRoomRtcVideoRenderSetup(registrar.messenger(), self)
    textureRegistry = registrar.textures()
    messenger = registrar.messenger()
  }

  func onDetachedFromEngine(_ registrar: FlutterPluginRegistrar) {
    FLTRoomRtcVideoRenderSetup(registrar.messenger(), nil)
  }
}

@objcMembers
class FLTRenderEntry: NSObject {
  var roomUuid: String
  var videoRender: FLTRoomVideoRenderer
  init(_ roomUuid: String,
       videoRender: FLTRoomVideoRenderer) {
    self.roomUuid = roomUuid
    self.videoRender = videoRender
  }
}
