// Copyright (c) 2022 NetEase, Inc. All rights reserved.
// Use of this source code is governed by a MIT license that can be
// found in the LICENSE file.

#import "FLTRoomVideoRenderer.h"
// #include <NEWEYUV/Libyuv.h>
#include <NEDyldYuv/Libyuv.h>
@interface FLTRoomVideoRenderer () <FlutterStreamHandler>
@property(nonatomic, weak) id<FlutterTextureRegistry> registry;
@property(nonatomic, strong) FlutterEventSink eventSink;
@end

@implementation FLTRoomVideoRenderer {
  CVPixelBufferRef _pixelBufferRef;
  CGSize _frameSize;
  CGSize _renderSize;
  NERoomVideoRotationType _rotation;
  FlutterEventChannel *_eventChannel;
  bool _isFirstFrameRendered;
  bool _mirror;
}

@synthesize textureId = _textureId;

- (instancetype)initWithTextureRegistry:(id<FlutterTextureRegistry>)registry
                              messenger:(NSObject<FlutterBinaryMessenger> *)messenger {
  self = [super init];
  if (self) {
    _isFirstFrameRendered = false;
    _registry = registry;
    _pixelBufferRef = nil;
    _eventSink = nil;
    _rotation = NERoomVideoRotationTypeRotation_0;
    _frameSize = CGSizeZero;
    _renderSize = CGSizeZero;
    _mirror = false;
    _textureId = [registry registerTexture:self];
    _eventChannel = [FlutterEventChannel
        eventChannelWithName:[NSString
                                 stringWithFormat:@"NERoomKitFlutterVideoRenderer/Texture%lld",
                                                  _textureId]
             binaryMessenger:messenger];
    [_eventChannel setStreamHandler:self];
  }
  return self;
}
- (void)dealloc {
  if (_pixelBufferRef) {
    CVBufferRelease(_pixelBufferRef);
    _pixelBufferRef = NULL;
  }
}

- (CVPixelBufferRef)copyPixelBuffer {
  if (_pixelBufferRef != nil) {
    CVBufferRetain(_pixelBufferRef);
    return _pixelBufferRef;
  }
  return nil;
}
- (void)dispose {
  [_registry unregisterTexture:_textureId];
}

- (void)clearImage {
  // TODO
}
- (void)setMirror:(BOOL)mirror {
  _mirror = mirror;
}

- (void)copyI420BufferToCVPixelBuffer:(CVPixelBufferRef)pixelBuffer
                            withFrame:(NERoomVideoFrame *)frame {
  CVPixelBufferRetain(pixelBuffer);
  CVPixelBufferLockBaseAddress(pixelBuffer, 0);

  const OSType pixelFormat = CVPixelBufferGetPixelFormatType(pixelBuffer);
  uint8_t *dst = CVPixelBufferGetBaseAddress(pixelBuffer);
  size_t bytesPerRow = CVPixelBufferGetBytesPerRow(pixelBuffer);

  uint8_t *src_y = (uint8_t *)frame.buffer;
  uint8_t *src_u = (uint8_t *)frame.uBuffer;
  uint8_t *src_v = (uint8_t *)frame.vBuffer;
  int width = frame.width;
  int height = frame.height;
  int dst_stride_argb = width * 4;
  int argb_size = width * height * 4;

  uint8_t *mirror_buffer = NULL;
  if (_mirror) {
    mirror_buffer = (uint8_t *)malloc(argb_size);
  }

  if ((int)frame.rotation != 0) {
    uint8_t *convert_buffer = (uint8_t *)malloc(argb_size);
    I420ToARGB(src_y, [frame.strides[0] intValue], src_u, [frame.strides[1] intValue], src_v,
               [frame.strides[2] intValue], convert_buffer, dst_stride_argb, width, height);

    uint8_t *dst_buffer = _mirror ? mirror_buffer : dst;
    int dst_buffer_stride = _mirror ? dst_stride_argb : bytesPerRow;
    ARGBRotate(convert_buffer, dst_stride_argb, dst_buffer, dst_buffer_stride, width, height,
               (int)frame.rotation);

    free(convert_buffer);
  } else {
    uint8_t *dst_buffer = _mirror ? mirror_buffer : dst;
    int dst_buffer_stride = _mirror ? dst_stride_argb : bytesPerRow;

    I420ToARGB(src_y, [frame.strides[0] intValue], src_u, [frame.strides[1] intValue], src_v,
               [frame.strides[2] intValue], dst_buffer, dst_buffer_stride, width, height);
  }

  if (_mirror) {
    ARGBMirror(mirror_buffer, dst_stride_argb, dst, bytesPerRow, width, height);

    free(mirror_buffer);
  }

  CVPixelBufferUnlockBaseAddress(pixelBuffer, 0);
  CVPixelBufferRelease(pixelBuffer);
}
- (void)setFirstRender:(BOOL)firstRender {
  _isFirstFrameRendered = firstRender;
  _frameSize = CGSizeZero;
  _renderSize = CGSizeZero;
}
#pragma mark-----------------------------  NERoomRtcVideoRenderSink  -----------------------------
- (void)onRenderWithFrame:(NERoomVideoFrame *)frame {
  // Current supported I420
  if (frame.format != NERoomVideoFormatTypeI420) return;

  CGSize current_size = (frame.rotation % 180 == 0) ? CGSizeMake(frame.width, frame.height)
                                                    : CGSizeMake(frame.height, frame.width);

  if (!CGSizeEqualToSize(_frameSize, current_size)) {
    [self setSize:current_size];
  }
  [self copyI420BufferToCVPixelBuffer:_pixelBufferRef withFrame:frame];

  __weak FLTRoomVideoRenderer *weakSelf = self;
  if (_renderSize.width != frame.width || _renderSize.height != frame.height ||
      _rotation != frame.rotation) {
    dispatch_async(dispatch_get_main_queue(), ^{
      FLTRoomVideoRenderer *strongSelf = weakSelf;
      if (strongSelf.eventSink) {
        strongSelf.eventSink(@{
          @"event" : @"onFrameResolutionChanged",
          @"id" : @(strongSelf.textureId),
          @"width" : @(frame.width),
          @"height" : @(frame.height),
          @"rotation" : @(frame.rotation)
        });
#ifdef DEBUG
        NSLog(@"CalledFlutter:onFrameResolutionChanged");
#endif
      }
    });
    _renderSize = CGSizeMake(frame.width, frame.height);
    _rotation = frame.rotation;
  }

  dispatch_async(dispatch_get_main_queue(), ^{
    FLTRoomVideoRenderer *strongSelf = weakSelf;
    [strongSelf.registry textureFrameAvailable:strongSelf.textureId];
    if (!strongSelf->_isFirstFrameRendered) {
      if (strongSelf.eventSink) {
        strongSelf.eventSink(@{@"event" : @"onFirstFrameRendered"});
        strongSelf->_isFirstFrameRendered = true;
#ifdef DEBUG
        NSLog(@"CalledFlutter:onFirstFrameRendered");
#endif
      }
    }
  });
}

- (void)setSize:(CGSize)size {
  if (_pixelBufferRef == nil ||
      (size.width != _frameSize.width || size.height != _frameSize.height)) {
    // CVPixelBufferCreate
    if (_pixelBufferRef) {
      CVBufferRelease(_pixelBufferRef);
      _pixelBufferRef = NULL;
    }
    NSDictionary *pixelAttributes = @{(id)kCVPixelBufferIOSurfacePropertiesKey : @{}};
    CVReturn retu =
        CVPixelBufferCreate(kCFAllocatorDefault, size.width, size.height, kCVPixelFormatType_32BGRA,
                            (__bridge CFDictionaryRef)(pixelAttributes), &_pixelBufferRef);

    //    CVBufferRetain(_pixelBufferRef);
    _frameSize = size;
  }
}

#pragma mark - FlutterStreamHandler

- (FlutterError *_Nullable)onCancelWithArguments:(id _Nullable)arguments {
  _eventSink = nil;
  return nil;
}

- (FlutterError *_Nullable)onListenWithArguments:(id _Nullable)arguments
                                       eventSink:(nonnull FlutterEventSink)sink {
  _eventSink = sink;
  return nil;
}

@end
