// Copyright (c) 2022 NetEase, Inc. All rights reserved.
// Use of this source code is governed by a MIT license that can be
// found in the LICENSE file.

import Foundation
import NERoomKit

class FLTRoomWhiteboardController: NSObject, FLTRoomWhiteboardControllerApi {
  func setCanvasBackgroundColorRoomUuid(_ roomUuid: String, color: String, completion: @escaping (FLTSimpleResponse?, FlutterError?) -> Void) {
    guard let roomContext = NERoomKit.shared().roomService.getRoomContext(roomUuid: roomUuid) else {
      let simpleResponse = FLTSimpleResponse.make(
        withCode: NSNumber(value: -1),
        msg: paramsError,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(simpleResponse, nil)
      return
    }
    let code = roomContext.whiteboardController.setCanvasBackgroundColor(color)
    let simpleResponse = FLTSimpleResponse.make(
      withCode: NSNumber(value: code),
      msg: nil,
      data1: nil,
      data2: nil,
      data3: nil
    )
    completion(simpleResponse, nil)
  }

  func lockCamera(withContentRoomUuid roomUuid: String, width: NSNumber, height: NSNumber, completion: @escaping (FLTSimpleResponse?, FlutterError?) -> Void) {
    guard let roomContext = NERoomKit.shared().roomService.getRoomContext(roomUuid: roomUuid) else {
      let simpleResponse = FLTSimpleResponse.make(
        withCode: NSNumber(value: -1),
        msg: paramsError,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(simpleResponse, nil)
      return
    }
    let code = roomContext.whiteboardController.lockCameraWithContent(width: width.intValue, height: height.intValue)
    let simpleResponse = FLTSimpleResponse.make(
      withCode: NSNumber(value: code),
      msg: nil,
      data1: nil,
      data2: nil,
      data3: nil
    )
    completion(simpleResponse, nil)
  }

  func setEnableDrawRoomUuid(_ roomUuid: String?,
                             enable: NSNumber?,
                             completion: @escaping (FLTSimpleResponse?, FlutterError?) -> Void) {
    guard let roomUuid = roomUuid,
          let enable = enable,
          let roomContext = NERoomKit.shared().roomService.getRoomContext(roomUuid: roomUuid)
    else {
      let simpleResponse = FLTSimpleResponse.make(
        withCode: NSNumber(value: -1),
        msg: paramsError,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(simpleResponse, nil)
      return
    }
    roomContext.whiteboardController.setEnableDraw(enable: enable.boolValue)
    let simpleResponse = FLTSimpleResponse.make(
      withCode: NSNumber(value: 0),
      msg: nil,
      data1: nil,
      data2: nil,
      data3: nil
    )
    completion(simpleResponse, nil)
  }

  func startWhiteboardShareRoomUuid(_ roomUuid: String?,
                                    completion: @escaping (FLTSimpleResponse?, FlutterError?)
                                      -> Void) {
    guard let roomUuid = roomUuid,
          let roomContext = NERoomKit.shared().roomService.getRoomContext(roomUuid: roomUuid)
    else {
      let simpleResponse = FLTSimpleResponse.make(
        withCode: NSNumber(value: -1),
        msg: paramsError,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(simpleResponse, nil)
      return
    }
    roomContext.whiteboardController.startWhiteboardShare { code, str, _ in
      let response = FLTSimpleResponse.make(
        withCode: NSNumber(value: code),
        msg: str,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(response, nil)
    }
  }

  func stopWhiteboardShareRoomUuid(_ roomUuid: String?,
                                   completion: @escaping (FLTSimpleResponse?, FlutterError?)
                                     -> Void) {
    guard let roomUuid = roomUuid,
          let roomContext = NERoomKit.shared().roomService.getRoomContext(roomUuid: roomUuid)
    else {
      let simpleResponse = FLTSimpleResponse.make(
        withCode: NSNumber(value: -1),
        msg: paramsError,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(simpleResponse, nil)
      return
    }
    roomContext.whiteboardController.stopWhiteboardShare { code, str, _ in
      let response = FLTSimpleResponse.make(
        withCode: NSNumber(value: code),
        msg: str,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(response, nil)
    }
  }

  func stopMemberWhiteboardShareRoomUuid(_ roomUuid: String?,
                                         userUuid: String?,
                                         completion: @escaping (FLTSimpleResponse?,
                                                                FlutterError?) -> Void) {
    guard let roomUuid = roomUuid,
          let userUuid = userUuid,
          let roomContext = NERoomKit.shared().roomService.getRoomContext(roomUuid: roomUuid)
    else {
      let simpleResponse = FLTSimpleResponse.make(
        withCode: NSNumber(value: -1),
        msg: paramsError,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(simpleResponse, nil)
      return
    }
    roomContext.whiteboardController
      .stopMemberWhiteboardShare(userUuid: userUuid) { code, str, _ in
        let simpleResponse = FLTSimpleResponse.make(
          withCode: NSNumber(value: code),
          msg: str,
          data1: nil,
          data2: nil,
          data3: nil
        )
        completion(simpleResponse, nil)
      }
  }
}

extension FLTRoomWhiteboardController: FLTAssociativeWrapper {
  func onAttachedToEngine(_ registrar: FlutterPluginRegistrar) {
    FLTRoomWhiteboardControllerApiSetup(registrar.messenger(), self)
    let factory = FLNativeViewFactory(flutterBinaryMessenger: registrar.messenger())

    registrar.register(
      factory,
      withId: "com.netease.yunxin.flutter.plugins.roomkit/whiteboardView"
    )
  }

  func onDetachedFromEngine(_ registrar: FlutterPluginRegistrar) {
    FLTRoomWhiteboardControllerApiSetup(registrar.messenger(), nil)
  }
}
