// Copyright (c) 2022 NetEase, Inc. All rights reserved.
// Use of this source code is governed by a MIT license that can be
// found in the LICENSE file.

import Foundation
import NERoomKit

class FLTRoomKit: NSObject, FLTRoomKitApi {
  func sdkVersionsWithError(_ error: AutoreleasingUnsafeMutablePointer<FlutterError?>)
    -> FLTSDKVersions? {
    let version = NERoomKit.shared().sdkVersions
    let sdkVersion = FLTSDKVersions.make(withImVersion: version.imVersion,
                                         rtcVersion: version.rtcVersion,
                                         whiteboardVersion: version.whiteboardVersion,
                                         roomKitVersion: version.roomKitVersion)
    return sdkVersion
  }

  func initializeOptions(_ options: FLTInitializeRequest?,
                         completion: @escaping (FLTSimpleResponse?, FlutterError?) -> Void) {
    let option = NERoomKitOptions(appKey: options?.appKey ?? "")
    option.extras = options?.extras ?? [String: String]()
    option.useAssetServerConfig = options?.useAssetServerConfig.boolValue ?? false
    option.serverUrl = options?.serverUrl
    if let jsonStr = options?.serverConfigJson {
      option.serverConfig = jsonStr.jsonStringMapModel(NEServerConfig.self)
    }
    NERoomKit.shared().initialize(options: option) { code, str, _ in
      let response = FLTSimpleResponse.make(
        withCode: NSNumber(value: code),
        msg: str,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(response, nil)
    }
  }

  func switchLanguageLanguageCode(_ languageCode: String,
                                  error: AutoreleasingUnsafeMutablePointer<FlutterError?>)
    -> NSNumber? {
    var type: NERoomLanguage = .automatic
    switch languageCode {
    case "zh":
      type = .chinese
    case "en":
      type = .english
    case "ja":
      type = .japanese
    default:
      type = .automatic
    }
    var code: Int = NERoomKit.shared().switchLanguage(language: type)
    return NSNumber(value: code)
  }
}

extension FLTRoomKit: FLTAssociativeWrapper {
  func onAttachedToEngine(_ registrar: FlutterPluginRegistrar) {
    FLTRoomKitApiSetup(registrar.messenger(), self)
  }

  func onDetachedFromEngine(_ registrar: FlutterPluginRegistrar) {
    FLTRoomKitApiSetup(registrar.messenger(), nil)
  }
}

extension Data {
  /// json data 转模型
  /// 可以直接拿网络请求成功的response直接转换
  func jsonDataMapModel<T: Decodable>(_ type: T.Type) -> T? {
    let decoder = JSONDecoder()
    do {
      return try decoder.decode(T.self, from: self)
    } catch {
      print("解析失败\(self)")
      return nil
    }
  }
}

extension String {
  /// json字符串转模型
  func jsonStringMapModel<T: Decodable>(_ type: T.Type) -> T? {
    if let jsonData = data(using: .utf8) {
      return jsonData.jsonDataMapModel(T.self)
    }
    print("解析失败\(self)")
    return nil
  }
}
