// Copyright (c) 2022 NetEase, Inc. All rights reserved.
// Use of this source code is governed by a MIT license that can be
// found in the LICENSE file.

import Foundation
import NERoomKit

class FLTAuthService: NSObject, FLTAuthServiceApi {
  var temp: FLTAuthListener?
  /// 账号登录
  /// - Parameters:
  ///   - account: 账号
  ///   - token: 令牌
  ///   - completion: 回调
  func loginAccount(_ account: String?,
                    token: String?,
                    completion: @escaping (FLTSimpleResponse?, FlutterError?) -> Void) {
    NERoomKit.shared().authService
      .login(account: account ?? "", token: token ?? "") { code, str, _ in
        let response = FLTSimpleResponse.make(
          withCode: NSNumber(value: code),
          msg: str,
          data1: nil,
          data2: nil,
          data3: nil
        )
        completion(response, nil)
      }
  }

  /// 登出
  /// - Parameter completion: 回调
  func logout(completion: @escaping (FLTSimpleResponse?, FlutterError?) -> Void) {
    NERoomKit.shared().authService.logout { code, str, _ in
      let response = FLTSimpleResponse.make(
        withCode: NSNumber(value: code),
        msg: str,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(response, nil)
    }
  }

  func isLoggedInWithError(_ error: AutoreleasingUnsafeMutablePointer<FlutterError?>) -> NSNumber? {
    NSNumber(value: NERoomKit.shared().authService.isLoggedIn)
  }
}

extension FLTAuthService: FLTAssociativeWrapper {
  func onAttachedToEngine(_ registrar: FlutterPluginRegistrar) {
    FLTAuthServiceApiSetup(registrar.messenger(), self)
    temp = FLTAuthListener(FLTAuthStatusEventSink(binaryMessenger: registrar.messenger()))
    NERoomKit.shared().authService.addAuthListener(temp!)
  }

  func onDetachedFromEngine(_ registrar: FlutterPluginRegistrar) {
    FLTAuthServiceApiSetup(registrar.messenger(), nil)
    if let _ = temp {
      NERoomKit.shared().authService.removeAuthListener(temp!)
    }
    temp = nil
  }
}

@objcMembers
class FLTAuthListener: NSObject, NEAuthListener {
  private var eventSink: FLTAuthStatusEventSink
  init(_ eventSink: FLTAuthStatusEventSink) {
    self.eventSink = eventSink
  }

  func onAuthEvent(evt: NEAuthEvent) {
    let statusEvent = FLTAuthStatusEvent.make(withStatus: evt.toString(),
                                              isLoggedIn: NSNumber(value: NERoomKit.shared()
                                                .authService.isLoggedIn))
    eventSink.onAuthStatusChangedEvent(statusEvent) { _ in }
  }
}

extension NEAuthEvent {
  func toString() -> String {
    switch self {
    case .kickOut:
      return "KickOut"
    case .tokenExpired:
      return "TokenExpired"
    case .incorrectToken:
      return "IncorrectToken"
    case .forbidden:
      return "Forbidden"
    case .accountTokenError:
      return "AccountTokenError"
    case .loggedIn:
      return "LoggedIn"
    case .loggedOut:
      return "LoggedOut"
    default:
      return ""
    }
  }
}
