// Copyright (c) 2022 NetEase, Inc. All rights reserved.
// Use of this source code is governed by a MIT license that can be
// found in the LICENSE file.

import Foundation
import NERoomKit

@objcMembers
class FLTMessageService: NSObject, FLTMessageServiceApi {
  func sendCustomMessageRoomUuid(_ roomUuid: String?,
                                 userUuid: String?,
                                 commandId: NSNumber?,
                                 data: String?,
                                 crossAppAuthorization: FLTCrossAppAuthorization?,
                                 completion: @escaping (FLTSimpleResponse?, FlutterError?)
                                   -> Void) {
    guard let roomUuid = roomUuid,
          let userUuid = userUuid,
          let data = data else {
      let simpleResponse = FLTSimpleResponse.make(
        withCode: NSNumber(value: -1),
        msg: paramsError,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(simpleResponse, nil)
      return
    }
    var crossAppAuth: NECrossAppAuthorization?
    if let crossAppAuthorization = crossAppAuthorization {
      crossAppAuth = NECrossAppAuthorization()
      crossAppAuth?.appKey = crossAppAuthorization.appKey
      crossAppAuth?.user = crossAppAuthorization.user
      crossAppAuth?.token = crossAppAuthorization.token
    }
    NERoomKit.shared().messageChannelService.sendCustomMessage(roomUuid: roomUuid,
                                                               userUuid: userUuid,
                                                               commandId: commandId?
                                                                 .intValue ?? 0,
                                                               data: data,
                                                               crossAppAuthorization: crossAppAuth) { code, str, _ in
      let simpleResponse = FLTSimpleResponse.make(
        withCode: NSNumber(value: code),
        msg: str,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(simpleResponse, nil)
    }
  }

  var messageListener: FLTMessageListener?
}

extension FLTMessageService: FLTAssociativeWrapper {
  func onAttachedToEngine(_ registrar: FlutterPluginRegistrar) {
    FLTMessageServiceApiSetup(registrar.messenger(), self)
    messageListener =
      FLTMessageListener(FLTMessageEventSink(binaryMessenger: registrar.messenger()))
    NERoomKit.shared().messageChannelService.addMessageChannelListener(listener: messageListener!)
  }

  func onDetachedFromEngine(_ registrar: FlutterPluginRegistrar) {
    if let _ = messageListener {
      NERoomKit.shared().messageChannelService
        .removeMessageChannelListener(listener: messageListener!)
    }
    messageListener = nil
  }
}

class FLTMessageListener: NSObject, NEMessageChannelListener {
  private var eventSink: FLTMessageEventSink
  init(_ eventSink: FLTMessageEventSink) {
    self.eventSink = eventSink
  }

  func onReceiveCustomMessage(message: NECustomMessage) {
    let messageEvent = FLTPassThroughMessageEvent.make(withRoomUuid: message.roomUuid,
                                                       senderUuid: message.senderUuid,
                                                       data: message.data,
                                                       commandId: NSNumber(value: message
                                                         .commandId))
    eventSink.onReceivePassThroughMessageMessageEvent(messageEvent) { _ in }
  }
}

extension String {
  func toDictionary() -> [String: Any]? {
    guard !isEmpty else {
      return nil
    }
    let data = data(using: .utf8)
    guard let dic = try? JSONSerialization.jsonObject(
      with: data!,
      options: .mutableContainers
    ) as? [String: Any] else {
      return nil
    }
    return dic
  }
}
