// Copyright (c) 2022 NetEase, Inc. All rights reserved.
// Use of this source code is governed by a MIT license that can be
// found in the LICENSE file.

import Foundation
import NERoomKit
import Flutter

class FLTRoomService: NSObject, FLTRoomServiceApi {
  var messenger: FlutterBinaryMessenger?
  var listeners = [FLTRoomListener]()
  var previewListeners = [FLTPreviewRoomListener]()
  var seatListeners = [FLTSeatEventListener]()
  var statsListeners = [FLTRoomRtcStatsListener]()

  deinit {
    listeners.removeAll()
    previewListeners.removeAll()
    seatListeners.removeAll()
    statsListeners.removeAll()
  }

  func createRoomParams(_ params: FLTCreateRoomParams?,
                        options: FLTCreateRoomOptions?,
                        completion: @escaping (FLTSimpleResponse?, FlutterError?) -> Void) {
    let param = NECreateRoomParams()
    param.roomUuid = params?.roomUuid ?? ""
    param.roomName = params?.roomName ?? ""
    param.templateId = params?.templateId.intValue ?? 0
    param.password = params?.password
    param.initialProperties = params?.initialProperties ?? [String: String]()
    let option = NECreateRoomOptions()
    option.enableChatroom = options?.enableChatroom.boolValue ?? false
    option.enableRtc = options?.enableRtc.boolValue ?? false
    option.enableLive = options?.enableLive.boolValue ?? false
    option.enableRecord = options?.enableRecord.boolValue ?? false
    option.enableWhiteboard = options?.enableWhiteboard.boolValue ?? false
    let seatInitParams = NESeatInitParams()
    if let seatInitParamsTemp = param.seatInitParams {
      seatInitParams.seatCount = seatInitParamsTemp.seatCount
      seatInitParams.seatRequestApprovalMode = seatInitParamsTemp.seatRequestApprovalMode
      seatInitParams.seatInvitationConfirmMode = seatInitParamsTemp.seatInvitationConfirmMode
      param.seatInitParams = seatInitParams
    }
    NERoomKit.shared().roomService.createRoom(params: param,
                                              options: option) { code, str, _ in
      let response = FLTSimpleResponse.make(
        withCode: NSNumber(value: code),
        msg: str,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(response, nil)
    }
  }

  func joinRoomParams(_ params: FLTJoinRoomParams?,
                      options: FLTJoinRoomOptions?,
                      completion: @escaping (FLTJoinRoomResponse?, FlutterError?) -> Void) {
    let param = NEJoinRoomParams()
    param.roomUuid = params?.roomUuid ?? ""
    param.userName = params?.userName ?? ""
    param.role = params?.role ?? ""
    param.password = params?.password
    param.avatar = params?.avatar
    if let crossAppAuth = params?.crossAppAuthorization {
      var crossAppAuthorization = NECrossAppAuthorization()
      crossAppAuthorization.appKey = crossAppAuth.appKey
      crossAppAuthorization.user = crossAppAuth.user
      crossAppAuthorization.token = crossAppAuth.token
      param.crossAppAuthorization = crossAppAuthorization
    }
    param.initialProperties = params?.initialProperties ?? [String: String]()
    let option = NEJoinRoomOptions()
    option.enableMyAudioDeviceOnJoinRtc = (options!.enableMyAudioDeviceOnJoinRtc) as! Bool
    NERoomKit.shared().roomService
      .joinRoom(params: param, options: option) { code, str, context in
        guard let context = context else {
          let response = FLTJoinRoomResponse.make(
            withCode: NSNumber(value: code),
            msg: str,
            snapshot: nil
          )
          completion(response, nil)
          return
        }
        if let messenger = self.messenger {
          let listener = FLTRoomListener(FLTRoomEventSink(binaryMessenger: messenger))
          self.listeners.append(listener)
          listener.roomUuid = param.roomUuid
          context.addRoomListener(listener: listener)

          let statsListener = FLTRoomRtcStatsListener(param.roomUuid, eventSink: FLTRoomStatsSink(binaryMessenger: messenger))
          self.statsListeners.append(statsListener)
          context.addRtcStatsListener(listener: statsListener)

          let seatListener = FLTSeatEventListener(FLTSeatEventSink(binaryMessenger: messenger))
          self.seatListeners.append(seatListener)
          context.seatController.addSeatListener(seatListener)
        }
        var members = [FLTRoomMember]()
        for member in context.remoteMembers {
          members.append(member.mapToMember())
        }
        let snapshot = FLTRoomSnapshot.make(withRoomUuid: context.roomUuid,
                                            roomName: context.roomName ?? "",
                                            password: context.password,
                                            isRoomLocked: NSNumber(value: context
                                              .isRoomLocked),
                                            rtcStartTime: NSNumber(value: context
                                              .rtcStartTime),
                                            localMember: context.localMember?.mapToMember(),
                                            remoteMembers: members,
                                            properties: context.roomProperties,
                                            isRtcSupported: NSNumber(value: context
                                              .rtcController.isSupported),
                                            isChatroomSupported: NSNumber(value: context
                                              .chatController.isSupported),
                                            isWhiteboardSupported: NSNumber(value: context
                                              .whiteboardController.isSupported),
                                            isCloudRecordSupported: NSNumber(value: false),
                                            isLiveSupported: NSNumber(value: context
                                              .liveController.isSupported),
                                            isSeatSupported: NSNumber(value: context.seatController.isSupported),
                                            sipCid: context.sipCid,
                                            remainingSeconds: NSNumber(value: context
                                              .remainingSeconds ?? 0))
        let response = FLTJoinRoomResponse.make(withCode: NSNumber(value: code),
                                                msg: str,
                                                snapshot: snapshot)
        completion(response, nil)
      }
  }

  func previewRoomParams(_ params: FLTPreviewRoomParams?, options: FLTPreviewRoomOptions?,
                         completion: @escaping (FLTSimpleResponse?, FlutterError?) -> Void) {
    NERoomKit.shared().roomService.previewRoom { code, _, context in
      guard let context = context else {
        let response = FLTSimpleResponse.make(
          withCode: NSNumber(value: -1),
          msg: paramsError,
          data1: nil,
          data2: nil,
          data3: nil
        )
        completion(response, nil)
        return
      }
      if let messenger = self.messenger {
        let listener = FLTPreviewRoomListener(FLTPreviewRoomEventSink(binaryMessenger: messenger))
        self.previewListeners.append(listener)
        context.addPreviewRoomListener(listener: listener)
      }
      let response = FLTSimpleResponse.make(
        withCode: NSNumber(value: 0),
        msg: nil,
        data1: nil,
        data2: nil,
        data3: nil
      )
      completion(response, nil)
    }
  }
}

// MARK: - ------------------------- NERoomKit extension --------------------------

extension NERoomMember {
  func mapToMember() -> FLTRoomMember {
    FLTRoomMember.make(withUuid: uuid,
                       name: name,
                       avatar: avatar,
                       role: role.mapToRole(),
                       isAudioOn: NSNumber(value: isAudioOn),
                       isVideoOn: NSNumber(value: isVideoOn),
                       isInRtcChannel: NSNumber(value: isInRtcChannel),
                       isInChatroom: NSNumber(value: isInChatroom),
                       isSharingScreen: NSNumber(value: isSharingScreen),
                       isSharingWhiteboard: NSNumber(value: isSharingWhiteboard),
                       properties: properties,
                       clientType: clientType.toString())
  }
}

extension NERoomRole {
  func mapToRole() -> FLTRoomRole {
    FLTRoomRole.make(withName: name,
                     limit: NSNumber(value: limit),
                     params: params?.mapToParam())
  }
}

extension NERoleParams {
  func mapToParam() -> FLTRoleParams {
    FLTRoleParams.make(withVideo: video?.mapToVideoParam(),
                       audio: audio?.mapToAudioParam(),
                       screen: screen?.mapToScreenParam())
  }
}

extension NERoleVideoParams {
  func mapToVideoParam() -> FLTRoleVideoParams {
    FLTRoleVideoParams.make(withWidth: NSNumber(value: width),
                            height: NSNumber(value: height),
                            fps: NSNumber(value: fps))
  }
}

extension NERoleAudioParams {
  func mapToAudioParam() -> FLTRoleAudioParams {
    FLTRoleAudioParams.make(withProfile: profile,
                            scenario: scenario)
  }
}

extension NERoleScreenParams {
  func mapToScreenParam() -> FLTRoleScreenParams {
    FLTRoleScreenParams.make(withVideoProfile: videoProfile)
  }
}

// MARK: - ------------------------- FLTAssociativeWrapper --------------------------

extension FLTRoomService: FLTAssociativeWrapper {
  func onAttachedToEngine(_ registrar: FlutterPluginRegistrar) {
    FLTRoomServiceApiSetup(registrar.messenger(), self)
    messenger = registrar.messenger()
  }

  func onDetachedFromEngine(_ registrar: FlutterPluginRegistrar) {
    FLTRoomServiceApiSetup(registrar.messenger(), nil)
    messenger = nil
  }
}

@objcMembers
class FLTRoomListener: NSObject, NERoomListener {
  private var evenSink: FLTRoomEventSink
  var roomUuid: String?
  init(_ eventSink: FLTRoomEventSink) {
    evenSink = eventSink
  }

  // MARK: - ------------------------- NERoomListener --------------------------

  // MARK: - ------------------------- 房间通知 --------------------------

  /// 直播状态变更
  func onRoomLiveStateChanged(state: NERoomLiveState) {
    guard let roomUuid = roomUuid else {
      return
    }
    let commonEvent = FLTCommonEvent.make(withRoomUuid: roomUuid,
                                          userUuid: nil,
                                          arg1: NSNumber(value: state.rawValue),
                                          arg2: nil,
                                          sArg1: nil,
                                          sArg2: nil)
    evenSink.onRoomLiveStateChangedEvent(commonEvent) { _ in }
  }

  /// 房间属性变更
  func onRoomPropertiesChanged(properties: [String: String]) {
    guard let roomUuid = roomUuid else {
      return
    }
    let propertyEvent = FLTPropertyEvent.make(withRoomUuid: roomUuid,
                                              userUuid: nil,
                                              properties: properties)
    evenSink.onRoomPropertiesChangedEvent(propertyEvent) { _ in }
  }

  /// 房间属性删除
  func onRoomPropertiesDeleted(properties: [String: String]) {
    guard let roomUuid = roomUuid else {
      return
    }
    let propertyEvent = FLTPropertyEvent.make(withRoomUuid: roomUuid,
                                              userUuid: nil,
                                              properties: properties)
    evenSink.onRoomPropertiesDeletedEvent(propertyEvent) { _ in }
  }

  /// 结束房间
  func onRoomEnded(reason: NERoomEndReason) {
    guard let roomUuid = roomUuid else {
      return
    }
    evenSink.onRoomEndRoomUuid(roomUuid,
                               reason: reason.toString()) { _ in }
  }

  /// 房间加锁状态
  func onRoomLockStateChanged(isLocked: Bool) {
    guard let roomUuid = roomUuid else {
      return
    }
    let toggleEvent = FLTRoomToggleEvent.make(
      withRoomUuid: roomUuid,
      isOn: NSNumber(value: isLocked)
    )
    evenSink.onRoomLockStateChangedEvent(toggleEvent) { _ in }
  }

  func onRtcChannelError(code: Int) {
    guard let roomUuid = roomUuid else {
      return
    }
    let commonEvent = FLTCommonEvent.make(withRoomUuid: roomUuid,
                                          userUuid: nil,
                                          arg1: NSNumber(value: code),
                                          arg2: nil,
                                          sArg1: nil,
                                          sArg2: nil)
    evenSink.onRtcChannelErrorEvent(commonEvent) { _ in }
  }

  func onRtcReceiveSEIMessage(_ userUuid: String, message: Data) {
    guard let roomUuid = roomUuid else {
      return
    }
    let commonEvent = FLTCommonEvent.make(withRoomUuid: roomUuid,
                                          userUuid: nil,
                                          arg1: nil,
                                          arg2: nil,
                                          sArg1: nil,
                                          sArg2: nil)
    // TODO:
  }

  // MARK: - ------------------------- 成员通知 --------------------------

  /// 成员属性变更
  func onMemberPropertiesChanged(member: NERoomMember, properties: [String: String]) {
    guard let roomUuid = roomUuid else {
      return
    }
    let propertyEvent = FLTPropertyEvent.make(withRoomUuid: roomUuid,
                                              userUuid: member.uuid,
                                              properties: properties)
    evenSink.onMemberPropertiesChangedEvent(propertyEvent) { _ in }
  }

  /// 成员属性删除
  func onMemberPropertiesDeleted(member: NERoomMember, properties: [String: String]) {
    guard let roomUuid = roomUuid else {
      return
    }
    let propertyEvent = FLTPropertyEvent.make(withRoomUuid: roomUuid,
                                              userUuid: member.uuid,
                                              properties: properties)
    evenSink.onMemberPropertiesDeletedEvent(propertyEvent) { _ in }
  }

  func onMemberRoleChanged(member: NERoomMember, oldRole: NERoomRole, newRole: NERoomRole) {
    guard let roomUuid = roomUuid else {
      return
    }
    let changedEvent = FLTMemberRoleChangeEvent.make(withRoomUuid: roomUuid,
                                                     userUuid: member.uuid,
                                                     role: member.role.mapToRole())
    evenSink.onMemberRoleChangedEvent(changedEvent) { _ in }
  }

  func onMemberJoinRoom(members: [NERoomMember]) {
    guard let roomUuid = roomUuid else {
      return
    }
    let roomMembers = members.map { $0.mapToMember() }
    let uuids = members.map { $0.uuid! }
    let memberEvent = FLTMembersEvent.make(withRoomUuid: roomUuid,
                                           members: roomMembers,
                                           uuids: uuids)
    evenSink.onMemberJoinRoomEvent(memberEvent) { _ in }
  }

  func onMemberLeaveRoom(members: [NERoomMember]) {
    guard let roomUuid = roomUuid else {
      return
    }
    let roomMembers = members.map { $0.mapToMember() }
    let uuids = members.map { $0.uuid! }
    let memberEvent = FLTMembersEvent.make(withRoomUuid: roomUuid,
                                           members: roomMembers,
                                           uuids: uuids)
    evenSink.onMemberLeaveRoomEvent(memberEvent) { _ in }
  }

  func onMemberJoinRtcChannel(members: [NERoomMember]) {
    guard let roomUuid = roomUuid else {
      return
    }
    let roomMembers = members.map { $0.mapToMember() }
    let uuids = members.map { $0.uuid! }
    let memberEvent = FLTMembersEvent.make(withRoomUuid: roomUuid,
                                           members: roomMembers,
                                           uuids: uuids)
    evenSink.onMemberJoinRtcChannelEvent(memberEvent) { _ in }
  }

  func onMemberLeaveRtcChannel(members: [NERoomMember]) {
    guard let roomUuid = roomUuid else {
      return
    }
    let roomMembers = members.map { $0.mapToMember() }
    let uuids = members.map { $0.uuid! }
    let memberEvent = FLTMembersEvent.make(withRoomUuid: roomUuid,
                                           members: roomMembers,
                                           uuids: uuids)
    evenSink.onMemberLeaveRtcChannelEvent(memberEvent) { _ in }
  }

  func onMemberJoinChatroom(members: [NERoomMember]) {
    guard let roomUuid = roomUuid else {
      return
    }
    let roomMembers = members.map { $0.mapToMember() }
    let uuids = members.map { $0.uuid! }
    let memberEvent = FLTMembersEvent.make(withRoomUuid: roomUuid,
                                           members: roomMembers,
                                           uuids: uuids)
    evenSink.onMemberJoinChatroomEvent(memberEvent) { _ in }
  }

  func onMemberLeaveChatroom(members: [NERoomMember]) {
    guard let roomUuid = roomUuid else {
      return
    }
    let uuids = members.map { $0.uuid! }
    let memberEvent = FLTMembersEvent.make(withRoomUuid: roomUuid,
                                           members: nil,
                                           uuids: uuids)
    evenSink.onMemberLeaveChatroomEvent(memberEvent) { _ in }
  }

  func onMemberAudioMuteChanged(member: NERoomMember, mute: Bool, operateBy: NERoomMember?) {
    guard let roomUuid = roomUuid else { return }
    let toggleEvent = FLTMemberToggleEvent.make(withRoomUuid: roomUuid,
                                                userUuid: member.uuid,
                                                operateUuid: operateBy?.uuid,
                                                isOn: NSNumber(value: member.isAudioOn))
    evenSink.onMemberAudioMuteChangedEvent(toggleEvent) { _ in }
  }

  func onMemberVideoMuteChanged(member: NERoomMember, mute: Bool, operateBy: NERoomMember?) {
    guard let roomUuid = roomUuid else { return }
    let toggleEvent = FLTMemberToggleEvent.make(withRoomUuid: roomUuid,
                                                userUuid: member.uuid,
                                                operateUuid: operateBy?.uuid,
                                                isOn: NSNumber(value: member.isVideoOn))
    evenSink.onMemberVideoMuteChangedEvent(toggleEvent) { _ in }
  }

  func onMemberScreenShareStateChanged(member: NERoomMember, isSharing: Bool,
                                       operateBy: NERoomMember?) {
    guard let roomUuid = roomUuid else { return }
    let toggleEvent = FLTMemberToggleEvent.make(withRoomUuid: roomUuid,
                                                userUuid: member.uuid,
                                                operateUuid: operateBy?.uuid,
                                                isOn: NSNumber(value: isSharing))
    evenSink.onMemberScreenShareStateChangedEvent(toggleEvent) { _ in }
  }

  func onMemberWhiteboardStateChanged(member: NERoomMember, isSharing: Bool,
                                      operateBy: NERoomMember?) {
    guard let roomUuid = roomUuid else { return }
    let toggleEvent = FLTMemberToggleEvent.make(withRoomUuid: roomUuid,
                                                userUuid: member.uuid,
                                                operateUuid: operateBy?.uuid,
                                                isOn: NSNumber(value: isSharing))
    evenSink.onMemberWhiteboardShareStateChangedEvent(toggleEvent) { _ in }
  }

  func onMemberNameChanged(member: NERoomMember, name: String) {
    guard let roomUuid = roomUuid else { return }
    let commonEvent = FLTCommonEvent.make(withRoomUuid: roomUuid,
                                          userUuid: member.uuid,
                                          arg1: nil,
                                          arg2: nil,
                                          sArg1: name,
                                          sArg2: nil)
    evenSink.onMemberNameChangedEvent(commonEvent) { _ in }
  }

  func onReceiveChatroomMessages(messages: [NERoomChatMessage]) {
    guard let roomUuid = roomUuid else {
      return
    }
    let roomMessages = messages.map { $0.mapToMessage() }
    var array = [FLTRoomMessage]()
    for message in roomMessages {
      if let m = message {
        array.append(m)
      }
    }
    let messageEvent = FLTRoomMessageEvent.make(withRoomUuid: roomUuid,
                                                messages: array)
    evenSink.onReceiveChatroomMessagesMessages(messageEvent) { _ in }
  }

  // TODO: 待RoomKit版本升级

//  func onAudioEffectFinished(effectId: UInt32) {
//    guard let roomUuid = roomUuid else { return }
//    let commonEvent = FLTCommonEvent.make(withRoomUuid: roomUuid,
//                                          userUuid: nil,
//                                          arg1: NSNumber(value: effectId),
//                                          arg2: nil,
//                                          sArg1: nil,
//                                          sArg2: nil)
  ////        evenSink.onMemberNameChangedEvent(commonEvent) { _ in }
//  }
//
//
//  func onAudioEffectTimestampUpdate(effectId: UInt32, timeStampMS: UInt64) {
//    guard let roomUuid = roomUuid else { return }
//    let commonEvent = FLTCommonEvent.make(withRoomUuid: roomUuid,
//                                          userUuid: nil,
//                                          arg1: NSNumber(value: effectId),
//                                          arg2: NSNumber(value: timeStampMS),
//                                          sArg1: nil,
//                                          sArg2: nil)
//  }

  func onAudioMixingStateChanged(reason: Int) {
    guard let roomUuid = roomUuid else { return }
    let commonEvent = FLTCommonEvent.make(withRoomUuid: roomUuid,
                                          userUuid: nil,
                                          arg1: NSNumber(value: reason),
                                          arg2: nil,
                                          sArg1: nil,
                                          sArg2: nil)
    evenSink.onAudioMixingStateChangedEvent(commonEvent, completion: { _ in

    })
  }

  func onChatroomMessageAttachmentProgress(messageUuid: String, transferred: Int64,
                                           total: Int64) {
    guard let roomUuid = roomUuid else {
      return
    }
    let commonEvent = FLTCommonEvent.make(withRoomUuid: roomUuid,
                                          userUuid: nil,
                                          arg1: NSNumber(value: transferred),
                                          arg2: NSNumber(value: total),
                                          sArg1: messageUuid,
                                          sArg2: nil)
    evenSink.onChatroomMessageAttachmentProgressMessages(commonEvent) { _ in }
  }

  func onWhiteboardError(code: Int, message: String) {
    guard let roomUuid = roomUuid else {
      return
    }
    let commonEvent = FLTCommonEvent.make(withRoomUuid: roomUuid,
                                          userUuid: nil,
                                          arg1: NSNumber(value: code),
                                          arg2: nil,
                                          sArg1: message,
                                          sArg2: nil)
    evenSink.onWhiteboardErrorEvent(commonEvent) { _ in }
  }

  func onRtcRemoteAudioVolumeIndication(volumes: [NEMemberVolumeInfo], totalVolume: Int) {
    guard let roomUuid = roomUuid else {
      return
    }
    let volumeInfos = volumes.map { $0.mapToInfo() }
    let volumeEvent = FLTRtcVolumeEvent.make(withRoomUuid: roomUuid,
                                             volumes: volumeInfos,
                                             totalVolume: NSNumber(value: totalVolume))
    evenSink.onRtcAudioVolumeIndicationEvent(volumeEvent) { _ in }
  }

  func onRtcLocalAudioVolumeIndication(volume: Int, enableVad: Bool) {
    guard let roomUuid = roomUuid else {
      return
    }
    let localEvent = FLTLocalVolumeEvent.make(withRoomUuid: roomUuid, volume: NSNumber(value: volume), enableVad: NSNumber(value: enableVad))
    evenSink.onLocalAudioVolumeIndicationwithVadEvent(localEvent) { _ in }
  }

  func onRtcAudioOutputDeviceChanged(device: NEAudioOutputDevice) {
    guard let roomUuid = roomUuid else {
      return
    }
    let commonEvent = FLTCommonEvent.make(withRoomUuid: roomUuid,
                                          userUuid: nil,
                                          arg1: NSNumber(value: device.rawValue),
                                          arg2: nil,
                                          sArg1: nil,
                                          sArg2: nil)
    evenSink.onRtcAudioOutputDeviceChangedEvent(commonEvent) { _ in }
  }

  func onRtcVirtualBackgroundSourceEnabled(enable: Bool,
                                           reason: Int) {
    guard let roomUuid = roomUuid else {
      return
    }
    let sourceEvent = FLTRtcVirtualBackgroundSourceEvent.make(withRoomUuid: roomUuid,
                                                              enabled: NSNumber(value: enable),
                                                              reason: NSNumber(
                                                                value: reason
                                                              ))
    evenSink.onRtcVirtualBackgroundSourceEnabledEvent(sourceEvent) { _ in }
  }
}

@objcMembers
class FLTSeatEventListener: NSObject, NESeatEventListener {
  private var evenSink: FLTSeatEventSink
  init(_ eventSink: FLTSeatEventSink) {
    evenSink = eventSink
  }

  // MARK: - ------------------------- NeSeatEventListener --------------------------

  func onSeatInvitationAccepted(_ seatIndex: Int, user: String, isAutoAgree: Bool) {
    evenSink.onSeatInvitationAcceptedSeatIndex(NSNumber(value: seatIndex), user: user, isAutoAgree: NSNumber(value: isAutoAgree)) { _ in }
  }

  func onSeatInvitationCancelled(_ seatIndex: Int, user: String, operateBy: String) {
    evenSink.onSeatInvitationCancelledSeatIndex(NSNumber(value: seatIndex), user: user, operateBy: operateBy) { _ in
    }
  }

  func onSeatInvitationReceived(_ seatIndex: Int, user: String, operateBy: String) {
    evenSink.onSeatInvitationReceivedSeatIndex(NSNumber(value: seatIndex), user: user, operateBy: operateBy) { _ in
    }
  }

  func onSeatKicked(_ seatIndex: Int, user: String, operateBy: String) {
    evenSink.onSeatKickedSeatIndex(NSNumber(value: seatIndex), user: user, operateBy: operateBy) { _ in
    }
  }

  func onSeatLeave(_ seatIndex: Int, user: String) {
    evenSink.onSeatLeaveSeatIndex(NSNumber(value: seatIndex), user: user) { _ in
    }
  }

  func onSeatInvitationRejected(_ seatIndex: Int, user: String) {
    evenSink.onSeatInvitationRejectedSeatIndex(NSNumber(value: seatIndex), user: user) { _ in
    }
  }

  func onSeatManagerAdded(_ managers: [String]) {
    evenSink.onSeatManagerAddedManagers(managers) { _ in
    }
  }

  func onSeatManagerRemoved(_ managers: [String]) {
    evenSink.onSeatManagerRemovedManagers(managers) { _ in
    }
  }

  func onSeatRequestApproved(_ seatIndex: Int, user: String, operateBy: String, isAutoAgree: Bool) {
    evenSink.onSeatRequestApprovedSeatIndex(NSNumber(value: seatIndex), user: user, operateBy: operateBy, isAutoAgree: NSNumber(value: isAutoAgree)) { _ in
    }
  }

  func onSeatRequestCancelled(_ seatIndex: Int, user: String) {
    evenSink.onSeatRequestCancelledSeatIndex(NSNumber(value: seatIndex), user: user) { _ in
    }
  }

  func onSeatRequestSubmitted(_ seatIndex: Int, user: String) {
    evenSink.onSeatRequestSubmittedSeatIndex(NSNumber(value: seatIndex), user: user) { _ in
    }
  }

  func onSeatListChanged(_ seatItems: [NESeatItem]) {
    evenSink.onSeatListChangedSeatItems(seatItems.map { seatItem in
      FLTSeatItem.make(withIndex: NSNumber(value: seatItem.index), status: NSNumber(value: seatItem.status.rawValue), user: seatItem.user, userName: seatItem.userName, icon: seatItem.icon, onSeatType: NSNumber(value: seatItem.onSeatType.rawValue), updated: NSNumber(value: seatItem.updated))
    }) { _ in
    }
  }
}

@objcMembers
class FLTPreviewRoomListener: NSObject, NEPreviewRoomListener {
  private var evenSink: FLTPreviewRoomEventSink
  init(_ eventSink: FLTPreviewRoomEventSink) {
    evenSink = eventSink
  }

  // MARK: - ------------------------- NEPreviewRoomListener --------------------------

  func onRtcVirtualBackgroundSourceEnabled(enable: Bool,
                                           reason: Int) {
    let event = FLTRtcVirtualBackgroundSourceEvent.make(withRoomUuid: "",
                                                        enabled: NSNumber(value: enable),
                                                        reason: NSNumber(value: reason))
    evenSink.onRtcVirtualBackgroundSourceEnabledEvent(event) { _ in }
  }
}

@objcMembers
class FLTRoomRtcStatsListener: NSObject, NERtcStatsListener {
  private var eventSink: FLTRoomStatsSink
  private var roomUuid: String
  init(_ roomUuid: String, eventSink: FLTRoomStatsSink) {
    self.roomUuid = roomUuid
    self.eventSink = eventSink
  }

  func onRtcStats(stats: NERoomRtcStats) {
    let event = FLTRoomStatsEvent.make(withTxBytes: NSNumber(value: stats.txBytes),
                                       rxBytes: NSNumber(value: stats.rxBytes),
                                       cpuAppUsage: NSNumber(value: stats.cpuAppUsage),
                                       cpuTotalUsage: NSNumber(value: stats.cpuTotalUsage),
                                       memoryAppUsageRatio: NSNumber(value: stats.memoryAppUsageRatio),
                                       memoryTotalUsageRatio: NSNumber(value: stats.memoryTotalUsageRatio),
                                       memoryAppUsageInKBytes: NSNumber(value: stats.memoryAppUsageInKBytes),
                                       totalDuration: NSNumber(value: stats.totalDuration),
                                       txAudioBytes: NSNumber(value: stats.txAudioBytes),
                                       txVideoBytes: NSNumber(value: stats.txVideoBytes),
                                       rxAudioBytes: NSNumber(value: stats.rxAudioBytes),
                                       rxVideoBytes: NSNumber(value: stats.rxVideoBytes),
                                       rxAudioKBitRate: NSNumber(value: stats.rxAudioKBitRate),
                                       rxVideoKBitRate: NSNumber(value: stats.rxVideoKBitRate),
                                       txAudioKBitRate: NSNumber(value: stats.txAudioKBitRate),
                                       txVideoKBitRate: NSNumber(value: stats.txVideoKBitRate),
                                       upRtt: NSNumber(value: stats.upRtt),
                                       downRtt: NSNumber(value: stats.downRtt),
                                       txAudioPacketLossRate: NSNumber(value: stats.txAudioPacketLossRate),
                                       txVideoPacketLossRate: NSNumber(value: stats.txVideoPacketLossRate),
                                       txAudioPacketLossSum: NSNumber(value: stats.txAudioPacketLossSum),
                                       txVideoPacketLossSum: NSNumber(value: stats.txVideoPacketLossSum),
                                       txAudioJitter: NSNumber(value: stats.txAudioJitter),
                                       txVideoJitter: NSNumber(value: stats.txVideoJitter),
                                       rxAudioPacketLossRate: NSNumber(value: stats.rxAudioPacketLossRate),
                                       rxVideoPacketLossRate: NSNumber(value: stats.rxVideoPacketLossRate),
                                       rxAudioPacketLossSum: NSNumber(value: stats.rxAudioPacketLossSum),
                                       rxVideoPacketLossSum: NSNumber(value: stats.rxVideoPacketLossSum),
                                       rxAudioJitter: NSNumber(value: stats.rxAudioJitter),
                                       rxVideoJitter: NSNumber(value: stats.rxVideoJitter))
    eventSink.onRtcStatsRoomUuid(roomUuid, event: event) { _ in }
  }

  func onNetworkQuality(statsArray: [NERoomRtcNetworkQualityInfo]) {
    let array = statsArray.map {
      FLTRoomNetworkQuality.make(withUserId: $0.userUuid,
                                 upStatus: NSNumber(value: $0.upStatus.rawValue),
                                 downStatus: NSNumber(value: $0.downStatus.rawValue))
    }
    eventSink.onNetworkQualityRoomUuid(roomUuid, qualities: array) { _ in }
  }
}

extension NEMemberVolumeInfo {
  func mapToInfo() -> FLTMemberVolumeInfo {
    FLTMemberVolumeInfo.make(withUserUuid: userUuid,
                             volume: NSNumber(value: volume))
  }
}

extension NERoomChatMessage {
  func mapToMessage() -> FLTRoomMessage? {
    var uuids: [String]?
    if let members = toUserUuidList {
      uuids = members.map { $0 }
    }
    if messageType == .custom,
       let m = self as? NERoomChatCustomMessage {
      return FLTRoomMessage.make(withUuid: messageUuid,
                                 type: messageType.toString(),
                                 fromAccount: fromUserUuid,
                                 fromNick: fromNick ?? "",
                                 content: m.attachStr,
                                 toAccounts: uuids,
                                 time: NSNumber(value: Int64(time ?? 0)),
                                 attachment: nil)
    } else if messageType == .text,
              let m = self as? NERoomChatTextMessage {
      return FLTRoomMessage.make(withUuid: messageUuid,
                                 type: messageType.toString(),
                                 fromAccount: fromUserUuid,
                                 fromNick: fromNick ?? "",
                                 content: m.text,
                                 toAccounts: uuids,
                                 time: NSNumber(value: Int64(time ?? 0)),
                                 attachment: nil)
    } else if messageType == .file,
              let m = self as? NERoomChatFileMessage {
      let attachment = FLTRoomMessageAttachment.make(withDisplayName: m.displayName,
                                                     extension: m.suffix,
                                                     md5: m.md5,
                                                     url: m.url,
                                                     size: NSNumber(value: m.size),
                                                     thumbPath: nil,
                                                     path: m.path,
                                                     width: nil,
                                                     height: nil)
      return FLTRoomMessage.make(withUuid: messageUuid,
                                 type: messageType.toString(),
                                 fromAccount: fromUserUuid,
                                 fromNick: fromNick ?? "",
                                 content: nil,
                                 toAccounts: uuids,
                                 time: NSNumber(value: Int64(time ?? 0)),
                                 attachment: attachment)
    } else if messageType == .image,
              let m = self as? NERoomChatImageMessage {
      let attachment = FLTRoomMessageAttachment.make(withDisplayName: m.displayName,
                                                     extension: m.suffix,
                                                     md5: m.md5,
                                                     url: m.url,
                                                     size: NSNumber(value: m.size),
                                                     thumbPath: m.thumbPath,
                                                     path: m.path,
                                                     width: NSNumber(value: m.width),
                                                     height: NSNumber(value: m.height))
      return FLTRoomMessage.make(withUuid: messageUuid,
                                 type: messageType.toString(),
                                 fromAccount: fromUserUuid,
                                 fromNick: fromNick ?? "",
                                 content: nil,
                                 toAccounts: uuids,
                                 time: NSNumber(value: Int64(time ?? 0)),
                                 attachment: attachment)
    }
    return nil
  }
}

extension NERoomChatMessageType {
  func toString() -> String {
    switch self {
    case .text:
      return "text"
    case .custom:
      return "custom"
    case .image:
      return "image"
    case .file:
      return "file"
    default:
      return "text"
    }
  }
}

extension NERoomEndReason {
  func toString() -> String {
    switch self {
    case .leaveBySelf:
      return "leaveBySelf"
    case .syncDataError:
      return "syncDataError"
    case .kickOut:
      return "kickOut"
    case .kickBySelf:
      return "selfKick"
    case .closeByMember:
      return "closeByMember"
    case .endOfLife:
      return "endOfLife"
    case .allMembersOut:
      return "allMemberOut"
    case .closeByBackend:
      return "closeByBackend"
    case .loginStateError:
      return "loginStateError"
    case .endOfRtc:
      return "endOfRtc"
    case .unknow:
      return "unknown"
    default:
      return "unknown"
    }
  }
}
