// Copyright (c) 2022 NetEase, Inc. All rights reserved.
// Use of this source code is governed by a MIT license that can be
// found in the LICENSE file.
part of netease_roomkit;

/// 当前房间的上下文
@HawkEntryPoint()
abstract class NERoomContext with AttachmentMixin {
  ///
  /// 房间唯一标识
  ///
  String get roomUuid;

  ///
  /// 房间成员
  ///
  String get roomName;

  ///
  /// 房间密码
  ///
  String? get password;

  ///
  /// 房间 rtc 开始时间, 单位ms
  ///
  int get rtcStartTime;

  ///
  /// 查询房间当前锁定状态
  ///
  bool get isRoomLocked;

  ///
  /// 查询房间当前锁定状态
  ///
  String? get sipCid;

  ///
  /// 查询房间当前剩余时间
  ///
  double? get remainingSeconds;

  ///
  /// 本端成员
  ///
  NERoomMember get localMember;

  ///
  /// 远端成员（不包括本端）
  ///
  List<NERoomMember> get remoteMembers;

  ///
  /// 通过 uuid 获取成员对象
  ///
  NERoomMember? getMember(String? uuid);

  /// 获取聊天控制器
  NERoomChatController get chatController;

  /// 获取rtc控制器
  NERoomRtcController get rtcController;

  /// 获取白板控制器
  NERoomWhiteboardController get whiteboardController;

  /// 获取直播控制器
  NERoomLiveController get liveController;

  NERoomSeatController get seatController;

  /// 获取当前房间属性
  Map<String, String> get roomProperties;

  ///
  /// 添加房间事件监听
  ///
  void addEventCallback(NERoomEventCallback callback);

  /// 移除房间事件监听
  void removeEventCallback(NERoomEventCallback callback);

  ///
  /// 添加统计数据监听
  ///
  void addRtcStatsCallback(NERoomRtcStatsCallback callback);

  /// 删除统计数据监听
  void removeRtcStatsCallback(NERoomRtcStatsCallback callback);

  ///
  /// 修改成员角色
  ///
  /// [userUuid] 用户id
  ///
  /// [role] 角色
  ///
  Future<VoidResult> changeMemberRole(String userUuid, String role);

  ///
  /// 修改昵称
  ///
  /// [nickname] 昵称
  ///
  Future<VoidResult> changeMyName(String name);

  ///
  /// 将自身当前的角色转移给[userUuid]对应的用户，自身会恢复到默认的房间角色。
  /// 只有授权角色才能执行该操作。
  ///
  /// [userUuid] 目标用户id
  ///
  Future<VoidResult> handOverMyRole(String userUuid);

  ///
  /// 移除成员
  /// [userUuid] 用户id
  ///
  Future<VoidResult> kickMemberOut(String userUuid);

  ///
  /// 离开房间
  ///
  Future<VoidResult> leaveRoom();

  ///
  /// 关闭房间
  ///
  /// [isForce] 如果结束房间失败，本地是否强制退出房间。
  ///
  Future<VoidResult> endRoom([bool isForce = false]);

  ///
  /// 锁定房间。锁定后成员无法加入
  ///
  Future<VoidResult> lockRoom();

  ///
  /// 解除锁定房间。解除锁定后成员可以加入该房间
  ///
  Future<VoidResult> unlockRoom();

  ///
  /// 更新房间属性
  ///
  /// [key] 房间属性唯一key
  /// [value] 房间属性值
  /// [associatedUuid] 声明该属性关联的用户uuid，如果该用户退出房间，该属性会被删除
  Future<VoidResult> updateRoomProperty(String key, String value,
      {String? associatedUuid});

  ///
  /// 删除房间属性
  ///
  Future<VoidResult> deleteRoomProperty(String key);

  ///
  /// 更新房间成员属性
  ///
  Future<VoidResult> updateMemberProperty(
      String userUuid, String key, String value);

  ///
  /// 删除房间成员属性
  ///
  Future<VoidResult> deleteMemberProperty(String userUuid, String key);
}

typedef RoomEndCallback = void Function(NERoomEndReason reason);

typedef RoomPropertiesChangedCallback = void Function(
    Map<String, String> properties);

typedef RoomPropertiesDeletedCallback = void Function(
    Map<String, String> properties);

typedef MemberPropertiesChangedCallback = void Function(
    NERoomMember member, Map<String, String> properties);

typedef MemberPropertiesDeletedCallback = void Function(
    NERoomMember member, Map<String, String> properties);

typedef MemberRoleChangedCallback = void Function(
    NERoomMember member, NERoomRole oldRole, NERoomRole newRole);

typedef MemberNameChangedCallback = void Function(
    NERoomMember member, String name);

typedef MemberJoinRoomCallback = void Function(List<NERoomMember> members);

typedef MemberLeaveRoomCallback = void Function(List<NERoomMember> members);

typedef MemberJoinRtcChannelCallback = void Function(
    List<NERoomMember> members);

typedef MemberLeaveRtcChannelCallback = void Function(
    List<NERoomMember> members);

typedef RtcChannelErrorCallback = void Function(int code);

typedef MemberJoinChatroomCallback = void Function(List<NERoomMember> members);

typedef MemberLeaveChatroomCallback = void Function(List<NERoomMember> members);

typedef MemberAudioMuteChangedCallback = void Function(
    NERoomMember member, bool mute, NERoomMember? operateBy);

typedef MemberVideoMuteChangedCallback = void Function(
    NERoomMember member, bool mute, NERoomMember? operateBy);

typedef MemberScreenShareStateChangedCallback = void Function(
    NERoomMember member, bool isSharing, NERoomMember? operateBy);

typedef ReceiveChatroomMessagesCallback = void Function(
    List<NERoomChatMessage> messages);

typedef ChatroomMessageAttachmentProgressCallback = void Function(
    String messageUuid, int transferred, int total);

typedef MemberWhiteboardShareStateChangedCallback = void Function(
    NERoomMember member, bool isSharing, NERoomMember? operateBy);

typedef WhiteboardErrorCallback = void Function(int code, String? msg);

typedef RoomLockStateChangedCallback = void Function(bool isLocked);

typedef RoomLiveStateChangedCallback = void Function(NERoomLiveState state);

typedef RtcRemoteAudioVolumeIndicationCallback = void Function(
    List<NEMemberVolumeInfo> volumes, int totalVolume);

typedef RtcLocalAudioVolumeIndicationCallback = void Function(
    int volume, bool enableVad);

typedef RtcAudioOutputDeviceChangedCallback = void Function(
    NEAudioOutputDevice device);

typedef RtcAudioMixingStateChangedCallback = void Function(int reason);

typedef RtcVirtualBackgroundSourceEnabledCallback = void Function(
    bool enabled, int reason);

typedef WhiteboardShowFileChooserCallback = Future<String?> Function(
    List<String?> types);

/// 房间事件监听器
class NERoomEventCallback {
  /// 房间结束回调
  final RoomEndCallback? roomEnd;

  /// 房间属性变更回调
  final RoomPropertiesChangedCallback? roomPropertiesChanged;

  /// 房间属性删除回调
  final RoomPropertiesDeletedCallback? roomPropertiesDeleted;

  /// 成员属性变更回调
  final MemberPropertiesChangedCallback? memberPropertiesChanged;

  /// 成员属性删除回调
  final MemberPropertiesDeletedCallback? memberPropertiesDeleted;

  /// 成员角色变更回调
  final MemberRoleChangedCallback? memberRoleChanged;

  /// 成员基础信息变更回调
  final MemberNameChangedCallback? memberNameChanged;

  /// 成员进入房间回调
  final MemberJoinRoomCallback? memberJoinRoom;

  ///  成员离开房间回调
  final MemberLeaveRoomCallback? memberLeaveRoom;

  /// 成员进入RTC频道回调
  final MemberJoinRtcChannelCallback? memberJoinRtcChannel;

  /// 成员离开RTC频道回调
  final MemberLeaveRtcChannelCallback? memberLeaveRtcChannel;

  /// RTC频道错误回调
  final RtcChannelErrorCallback? rtcChannelError;

  /// 成员进入聊天室回调
  final MemberJoinChatroomCallback? memberJoinChatroom;

  /// 成员离开聊天室回调
  final MemberLeaveChatroomCallback? memberLeaveChatroom;

  /// 成员音频静音状态变更回调
  final MemberAudioMuteChangedCallback? memberAudioMuteChanged;

  /// 成员视频开关状态变更回调
  final MemberVideoMuteChangedCallback? memberVideoMuteChanged;

  /// 成员屏幕共享状态变更回调
  final MemberScreenShareStateChangedCallback? memberScreenShareStateChanged;

  /// 聊天室消息回调
  final ReceiveChatroomMessagesCallback? chatroomMessagesReceived;

  /// 聊天室消息附件上传/下载进度回调
  final ChatroomMessageAttachmentProgressCallback?
      chatroomMessageAttachmentProgress;

  /// 成员白板共享状态变更回调
  final MemberWhiteboardShareStateChangedCallback?
      memberWhiteboardShareStateChanged;

  /// 白板错误回调
  final WhiteboardErrorCallback? whiteboardError;

  /// 房间锁状态回调
  final RoomLockStateChangedCallback? roomLockStateChanged;

  /// 直播状态回调
  final RoomLiveStateChangedCallback? liveStateChanged;

  /// 音量回调
  final RtcRemoteAudioVolumeIndicationCallback? rtcRemoteAudioVolumeIndication;

  // 本地瞬时音量回调
  final RtcLocalAudioVolumeIndicationCallback? rtcLocalAudioVolumeIndication;

  /// 音频输出设备类型变更
  final RtcAudioOutputDeviceChangedCallback? rtcAudioOutputDeviceChanged;

  /// 伴音错误状态
  final RtcAudioMixingStateChangedCallback? rtcAudioMixingStateChanged;

  /// 虚拟背景开关是否生效 [NERoomVirtualBackgroundSourceStateReason]
  final RtcVirtualBackgroundSourceEnabledCallback?
      rtcVirtualBackgroundSourceEnabled;

  /// 白板图片选择
  final WhiteboardShowFileChooserCallback? whiteboardShowFileChooser;

  NERoomEventCallback({
    this.roomEnd,
    this.roomPropertiesChanged,
    this.roomPropertiesDeleted,
    this.memberPropertiesChanged,
    this.memberPropertiesDeleted,
    this.memberRoleChanged,
    this.memberNameChanged,
    this.memberJoinRoom,
    this.memberLeaveRoom,
    this.memberJoinRtcChannel,
    this.memberLeaveRtcChannel,
    this.rtcChannelError,
    this.memberJoinChatroom,
    this.memberLeaveChatroom,
    this.memberAudioMuteChanged,
    this.memberVideoMuteChanged,
    this.memberScreenShareStateChanged,
    this.chatroomMessagesReceived,
    this.chatroomMessageAttachmentProgress,
    this.memberWhiteboardShareStateChanged,
    this.whiteboardError,
    this.roomLockStateChanged,
    this.liveStateChanged,
    this.rtcRemoteAudioVolumeIndication,
    this.rtcLocalAudioVolumeIndication,
    this.rtcAudioOutputDeviceChanged,
    this.rtcAudioMixingStateChanged,
    this.rtcVirtualBackgroundSourceEnabled,
    this.whiteboardShowFileChooser,
  });
}

///
/// 当前通话统计回调,SDK 定期向 App 报告当前通话的统计信息，每 2 秒触发一次
/// [stats] 通话相关的统计信息
///
typedef RtcStatsCallback = void Function(NERoomRtcStats stats);

///
/// 通话中所有用户的网络状态回调
/// [stats] 网络质量统计信息
///
typedef NetworkQualityCallback = void Function(
    List<NERoomRtcNetworkQualityInfo> stats);

///
/// 统计数据监听器
///
class NERoomRtcStatsCallback {
  final RtcStatsCallback? rtcStats;

  final NetworkQualityCallback? networkQuality;

  NERoomRtcStatsCallback({this.rtcStats, this.networkQuality});
}

///
/// 房间结束原因枚举
///
enum NERoomEndReason {
  ///
  /// 成员主动离开房间
  ///
  kLeaveBySelf,

  ///
  /// 数据同步错误
  ///
  kSyncDataError,

  ///
  /// 被管理员踢出房间
  ///
  kKickOut,

  ///
  /// 多端同时加入同一房间被踢
  ///
  kKickBySelf,

  ///
  /// 房间被关闭
  ///
  kCloseByMember,

  ///
  /// 房间到期关闭
  ///
  kEndOfLife,

  ///
  /// 所有成员退出
  ///
  kAllMemberOut,

  ///
  /// 后台关闭
  ///
  kCloseByBackend,

  ///
  /// 账号异常
  ///
  kLoginStateError,

  ///
  /// RTC 异常
  ///
  kEndOfRtc,

  ///
  /// 未知异常
  ///
  kUnknown,
}

///
/// 房间消息类型枚举
///
enum NERoomChatMessageType {
  ///
  /// 文本类型消息
  ///
  kText,

  ///
  /// 文件类型消息
  ///
  kFile,

  ///
  /// 图片类型消息
  ///
  kImage,

  ///
  /// 自定义类型消息
  ///
  kCustom,
}

///
/// 房间消息基类
///
abstract class NERoomChatMessage {
  ///
  /// 消息唯一ID
  ///
  String get messageUuid;

  ///
  /// 消息类型
  ///
  NERoomChatMessageType get messageType;

  ///
  /// 发送端，可能为空。如果为空，说明发送端可能并未加入到房间中
  ///
  String? get fromUserUuid;

  ///
  /// 发送端昵称
  ///
  String get fromNick;

  ///
  /// 接收端; 为空表示聊天室全体成员
  ///
  List<String>? get toUserUuidList;

  ///
  /// 消息时间戳，单位为毫秒
  ///
  int get time;

  Map toJson() => {
        'messageUuid': messageUuid,
        'messageType': messageType,
        'fromUserUuid': fromUserUuid,
        'fromNick': fromNick,
        'toUserUuidList': toUserUuidList,
        'time': time,
      };
}

///
/// 房间文本消息
///
abstract class NERoomChatTextMessage extends NERoomChatMessage {
  ///
  /// 消息内容
  ///
  String get text;

  Map toJson() => {
        ...super.toJson(),
        'text': text,
      };
}

///
/// 房间文本消息
///
abstract class NERoomChatCustomMessage extends NERoomChatMessage {
  ///
  /// 消息内容
  ///
  String get attachStr;

  Map toJson() => {
        ...super.toJson(),
        'attachStr': attachStr,
      };
}

///
/// 房间文件消息
///
abstract class NERoomChatFileMessage extends NERoomChatMessage {
  ///
  /// 文件显示名
  ///
  String? get displayName;

  ///
  /// 文件后缀名
  ///
  String? get extension;

  ///
  /// 文件md5
  ///
  String? get md5;

  ///
  /// 文件下载地址
  ///
  String get url;

  ///
  /// 文件大小
  ///
  int get size;

  ///
  /// 缩略图地址
  ///
  String? get thumbPath;

  ///
  /// 文件路径
  ///
  String? get path;

  Map toJson() => {
        ...super.toJson(),
        'displayName': displayName,
        'extension': extension,
        'md5': md5,
        'url': url,
        'size': size,
        'thumbPath': thumbPath,
        'path': path,
      };
}

///
/// 房间图片消息
///
abstract class NERoomChatImageMessage extends NERoomChatFileMessage {
  ///
  /// 图片宽度
  ///
  int get width;

  ///
  /// 图片高度
  ///
  int get height;

  Map toJson() => {
        ...super.toJson(),
        'width': width,
        'height': height,
      };
}
