// Copyright (c) 2022 NetEase, Inc. All rights reserved.
// Use of this source code is governed by a MIT license that can be
// found in the LICENSE file.

part of netease_roomkit;

@HawkEntryPoint()
abstract class NERoomBaseRtcController extends NERoomBaseController {
  ///切换摄像头
  Future<VoidResult> switchCamera();

  ///
  /// example: https://doc.yunxin.163.com/docs/jcyOTA0ODM/DQ1OTY0NjI?platformId=50002#示例代码
  /// 启用美颜模块。
  /// 在开启视频之前，调用 startBeauty 开启美颜功能模块。开启美颜功能之后才能使用全局美颜、滤镜等美颜效果。
  /// 调用此接口后，开启美颜引擎。如果后续不再需要使用美颜功能，可以调用 stopBeauty 结束美颜功能模块，销毁美颜引擎并释放资源。
  /// 开启美颜功能模块后，默认开启全局美颜效果，您也可以通过 setBeautyEffect 或其他滤镜、贴纸相关接口设置美颜、滤镜等效果。
  /// @return callback 0: 方法调用成功。其他: 方法调用失败。
  ////
  Future<VoidResult> startBeauty();

  ///
  /// 结束美颜功能模块。
  /// 如果后续不再需要使用美颜功能，可以调用 stopBeauty 结束美颜功能模块，SDK 会自动销毁美颜引擎并释放资源。
  ////
  Future<VoidResult> stopBeauty();

  ///
  /// 打开美颜功能，leave channel不会重置
  /// 接口目前主要用在需要临时关闭美颜，比如对比美颜前后的效果时才需要用到
  /// @param isOpenBeauty true为打开美颜功能，false为关闭
  /// @return callback 0: 方法调用成功。其他: 方法调用失败。
  ////
  Future<VoidResult> enableBeauty(bool isOpenBeauty);

  ///
  /// https://doc.yunxin.163.com/docs/jcyOTA0ODM/DQ1OTY0NjI?platformId=50002#全局美颜效果
  /// 调用 setBeautyEffect 开启指定美颜效果，并设置美颜强度。
  /// @param beautyType 参数用于指定美颜效果。目前支持美白、磨皮等美颜效果，具体枚举值请参考下表。  [NERoomBeautyEffectType]
  /// @param level 参数用于指定美颜强度。取值范围为 [0,1]，各种美颜效果的默认值不同。取值越大，美颜强度越大，开发者可以根据业务需要自定义设置美颜强度。
  /// @return callback 0: 方法调用成功。其他: 方法调用失败。
  ////
  Future<VoidResult> setBeautyEffect(
      NERoomBeautyEffectType beautyType, double level);

  ///
  /// 调用 addBeautyFilter 添加滤镜效果，并通过 setBeautyFilterLevel 设置滤镜强度。
  /// 滤镜效果可以和全局美颜效果互相叠加，但是不支持叠加多个滤镜。需要更换滤镜时，重复调用此接口使用新的滤镜资源即可。
  /// 添加滤镜时，需要通过 path 指定滤镜模型所在路径，支持 SD 卡上的绝对路径，或 asset 目录下的相对路径。
  /// 参数 level 用于指定滤镜强度，取值范围为 [0,1]，默认为 0.5。取值越大，滤镜越明显，开发者可以根据业务需要自定义设置滤镜强度。
  /// @param path	滤镜资源或模型所在路径。支持 SD 卡上的绝对路径，或 asset 目录下的相对路径。
  /// SD卡："/storage/emulated/0/Android/data/com.xxx/files/filter_portrait/filter_style_FN1"。
  /// asset: "2D/bunny"。
  /// @return callback 0: 方法调用成功。其他: 方法调用失败。
  ////
  Future<VoidResult> addBeautyFilter(String path);

  /// 不需要使用滤镜效果时，调用 removeBeautyFilter 移除滤镜。
  ////
  Future<VoidResult> removeBeautyFilter();

  ///
  /// 设置滤镜强度。
  /// 取值越大，滤镜强度越大，开发者可以根据业务需求自定义设置滤镜强度。
  /// 滤镜强度设置实时生效，更换滤镜后滤镜强度不变，如需调整，可以再次调用此接口重新设置滤镜强度。
  /// @param level	滤镜强度。取值范围为 [0 - 1]，默认值为 0.5
  /// @return 0: 方法调用成功。其他: 方法调用失败。
  ////
  Future<VoidResult> setBeautyFilterLevel(double level);

  ///
  /// 添加贴纸效果。
  /// 此接口用于加载贴纸资源，添加对应的贴纸效果。需要更换贴纸时，重复调用此接口使用新的贴纸资源即可。
  /// 使用滤镜、贴纸和美妆等自定义美颜效果之前，需要先准备好对应的美颜资源或模型。
  /// 贴纸效果可以和全局美颜、滤镜、美妆等效果互相叠加，但是不支持叠加多个贴纸。
  /// @param path 	贴纸资源所在路径。支持 SD 卡上的绝对路径，或 asset 目录下的相对路径。
  /// SD卡："/storage/emulated/0/Android/data/com.xxx/files/filter_portrait/filter_style_FN1"
  /// asset: "2D/bunny"
  /// @return 0: 方法调用成功。其他: 方法调用失败。
  ////
  Future<VoidResult> addBeautySticker(String path);

  /// 取消贴纸效果。
  /// @return 0: 方法调用成功。其他: 方法调用失败。
  Future<VoidResult> removeBeautySticker();

  Future<VoidResult> enableVirtualBackground(
      bool enabled, NERoomVirtualBackgroundSource backgroundSource);
}

@HawkEntryPoint()
abstract class NEPreviewRoomRtcController extends NERoomBaseRtcController {
  ///开启预览
  Future<VoidResult> startPreview();

  /// 关闭预览
  Future<VoidResult> stopPreview({bool releaseRtc = true});
}

@HawkEntryPoint()
abstract class NERoomRtcController extends NERoomBaseRtcController {
  /// 加入音视频房间
  Future<VoidResult> joinRtcChannel();

  /// 离开音视频房间
  Future<VoidResult> leaveRtcChannel();

  /// 关闭本端音频
  ///
  /// - [enableMediaPub] 开启静音后是否允许发送音频流，默认为false，即不发送。
  /// 如果设置为true，则本地开启静音后，会继续发送本地音频流(此时为静音包，其他端听不到本端声音)。
  Future<VoidResult> muteMyAudio([bool enableMediaPub = false]);

  ///打开本端音频
  ///
  /// - [enableMediaPub] 关闭静音后是否允许发送音频流，默认为true，即发送。
  /// 如果设置为false，则本地关闭静音后，不会发送本地音频流(此时其他端听不到本端声音)。
  Future<VoidResult> unmuteMyAudio([bool enableMediaPub = true]);

  ///
  /// 开启或关闭本地媒体流（主流）的发送。 该方法用于开始或停止向网络发送本地音频或视频数据。
  /// 该方法不影响接收或播放远端媒体流，也不会影响本地音频或视频的采集状态。
  /// 该方法在加入房间前后均可调用。
  ///
  /// @param mediaType 媒体发布类型，暂时仅支持音频。取值参考 [NERoomRtcMediaPublishType]
  /// @param enable 是否发布本地媒体流。true（默认）：发布本地媒体流。false：不发布本地媒体流。
  ///
  Future<VoidResult> enableMediaPub(int mediaType, bool enable);

  ///
  /// 暂停本地麦克风采集，调用后远端用户听不到本端声音。
  /// 该方法仅修改本地媒体状态，不会改变用户的音频开关状态[NERoomMember.isAudioOn]
  ///
  Future<VoidResult> pauseLocalAudioRecording();

  ///
  /// 恢复本地麦克风采集，调用后远端用户可以听到本端声音。
  /// 该方法仅修改本地媒体状态，不会改变成员的音频开关状态[NERoomMember.isAudioOn]
  ///
  Future<VoidResult> resumeLocalAudioRecording();

  ///关闭本端视频
  Future<VoidResult> muteMyVideo();

  ///打开本端视频
  Future<VoidResult> unmuteMyVideo();

  ///
  /// 暂停本地视频采集。
  /// 该方法仅修改本地媒体状态，不会改变用户的视频开关状态[NERoomMember.isVideoOn]
  ///
  ///
  Future<VoidResult> pauseLocalVideoCapture();

  ///
  /// 恢复本地视频采集。
  /// 该方法仅修改本地媒体状态，不会改变成员的视频开关状态[NERoomMember.isVideoOn]
  ///
  ///
  Future<VoidResult> resumeLocalVideoCapture();

  /// 尝试关闭成员音频，会进行权限校验
  Future<VoidResult> muteMemberAudio(String uuid);

  ///尝试打开成员音频，会进行权限校验
  Future<VoidResult> unmuteMemberAudio(String uuid);

  /// 尝试关闭成员视频，会进行权限校验
  Future<VoidResult> muteMemberVideo(String uuid);

  /// 尝试打开成员视频，会进行权限校验
  Future<VoidResult> unmuteMemberVideo(String uuid);

  ///关闭屏幕共享
  Future<VoidResult> stopScreenShare();

  ///尝试关闭房间内成员的屏幕共享，会进行权限校验。
  Future<VoidResult> stopMemberScreenShare(String uuid);

  ///开启屏幕共享
  Future<VoidResult> startScreenShare({String? iosAppGroup});

  ///
  /// 获取当前正在进行屏幕共享的成员
  ///
  String? getScreenSharingUserUuid();

  /// 订阅指定远端用户的视频流
  /// [userUuid] 用户id
  /// [subscribe] 是否取消订阅视频流
  Future<VoidResult> subscribeRemoteVideoStream(
      String userUuid, NEVideoStreamType streamType);

  /// 设置远端成员播放音量
  /// [userUuid] 用户id
  /// [volume] 音量，0-100
  Future<VoidResult> adjustUserPlaybackSignalVolume(
      String userUuid, int volume);

  /// 调节本地播放的所有远端用户的信号音量。 通过此接口可以实现调节所有远端用户在本地播放的混音音量。
  /// [volume] 音量，取值范围为 [0, 400]。0：静音。100：（默认）原始音量。400：最大可为原始音量的 4 倍（自带溢出保护）
  Future<VoidResult> adjustPlaybackSignalVolume(int volume);

  ///订阅或取消订阅指定远端用户辅流视频
  /// [userUuid] 用户id
  /// [subscribe] 是否取消订阅视频流
  Future<VoidResult> unsubscribeRemoteVideoStream(
      String userUuid, NEVideoStreamType streamType);

  /// 订阅指定远端用户辅流视频
  /// [userUuid] 用户id
  Future<VoidResult> subscribeRemoteVideoSubStream(String userUuid);

  /// 取消订阅指定远端用户辅流视频
  /// [userUuid] 用户id
  /// [subscribe] 是否取消订阅视频流
  Future<VoidResult> unsubscribeRemoteVideoSubStream(String userUuid);

  ///
  /// 打开或关闭扬声器
  ///
  Future<VoidResult> setSpeakerphoneOn(bool on);

  ///
  /// 查询扬声器是否开启
  ///
  Future<bool> isSpeakerphoneOn();

  ///
  /// 停止音频dump
  ///
  Future<VoidResult> stopAudioDump();

  ///
  /// 打开音频dump
  ///
  Future<VoidResult> startAudioDump(NEAudioDumpType type);

  ///
  /// 启用说话者音量提示。
  /// 该方法允许 SDK 定期向 App 反馈本地发流用户和瞬时音量最高的远端用户（最多 3 位）的音量相关信息，
  /// 即当前谁在说话以及说话者的音量。启用该方法后，只要房间内有发流用户，无论是否有人说话，
  /// SDK 都会在加入房间后根据预设的时间间隔触发 onRemoteAudioVolumeIndication 回调
  /// [enable] 是否启用说话者音量提示。
  /// [interval] 指定音量提示的时间间隔。单位为毫秒。必须设置为 100 毫秒的整数倍值，建议设置为 200 毫秒以上。
  ///
  Future<VoidResult> enableAudioVolumeIndication(bool enable, int interval,
      [bool vad = false]);

  ///
  /// 开始跨房间媒体流转发
  ///
  Future<VoidResult> startRtcChannelMediaRelay();

  ///
  /// 停止跨房间媒体流转发
  ///
  Future<VoidResult> stopRtcChannelMediaRelay();

  ///
  /// 开启音频混音
  /// [option] 混音参数
  ///
  Future<VoidResult> startAudioMixing(NECreateAudioMixingOption option);

  ///
  /// 开启音效
  /// [option] 音效参数
  ///
  Future<VoidResult> playEffect(int id, NECreateAudioEffectOption option);

  ///
  /// 停止音频混音
  ///
  Future<VoidResult> stopAudioMixing();

  ///
  /// 停止音效
  /// [id] 音效id
  ///
  Future<VoidResult> stopEffect(int id);

  ///
  /// 设置混音发送音量大小
  /// [volume] 音量大小，0-100
  ///
  Future<VoidResult> setAudioMixingSendVolume(int volume);

  ///
  /// 设置混音播放音量大小
  /// [volume] 音量大小，0-100
  ///
  Future<VoidResult> setAudioMixingPlaybackVolume(int volume);

  ///
  /// 设置音效发送音量大小
  /// [id] 音效id
  /// [volume] 音量大小，0-100
  ///
  Future<VoidResult> setEffectSendVolume(int id, int volume);

  ///
  /// 设置音效播放音量大小
  /// [id] 音效id
  /// [volume] 音量大小，0-100
  ///
  Future<VoidResult> setEffectPlaybackVolume(int id, int volume);

  ///
  /// 停止所有音效
  ///
  Future<VoidResult> stopAllEffects();

  ///
  /// 打开耳返
  /// [volume] 耳返音量大小，0-100
  ///
  Future<VoidResult> enableEarBack(int volume);

  ///
  /// 关闭耳返
  ///
  Future<VoidResult> disableEarBack();

  ///
  /// 暂停播放音乐文件及混音
  ///
  Future<VoidResult> pauseAudioMixing();

  ///
  /// 调节人声音量
  /// [volume] 采集信号音量，取值范围为 [0, 400] 0：静音。100：（默认）原始音量。400：最大可为原始音量的 4 倍（自带溢出保护）
  ///
  Future<VoidResult> adjustRecordingSignalVolume(int volume);

  ///
  /// 恢复音频混音
  ///
  Future<VoidResult> resumeAudioMixing();

  ///
  /// 关闭mic，不关闭音效
  /// [mute] true 关闭 false 打开
  ///
  Future<VoidResult> setRecordDeviceMute(bool mute);

  ///
  ///  设置角色
  ///  [role] 参考[NERoomRtcClientRole]
  ///
  Future<VoidResult> setClientRole(int role);

  ///
  ///设置音频编码属性
  ///@param profile 设置采样率、码率、编码模式和声道数 参考 [NERoomRtcAudioProfile]
  ///@param scenario 设置音频应用场景  参考[NERoomRtcAudioScenario]
  ///@return 0 成功 其他 失败
  ///
  Future<VoidResult> setAudioProfile(int profile, int scenario);

  ///
  ///设置房间场景
  ///房间场景可设置为通话（默认）或直播场景。针对不同场景采取的优化策略不同，
  ///如通话场景侧重语音流畅度，直播场景侧重视频清晰度。
  ///该方法必须在调用 joinChannel 前使用，进入房间后无法再设置房间场景。
  ///参考[NERoomRtcChannelProfile]
  ///

  Future<VoidResult> setChannelProfile(int value);

  ///
  /// 开启/关闭 AI 降噪功能，默认不会开启。开启 AI 降噪之后，在嘈杂的环境中可以针对背景人声、键盘声等
  /// 非稳态噪声进行定向降噪，同时也会提升对于环境稳态噪声的抑制，保留更纯粹的人声。
  /// [enable] - 开启/关闭 AI 降噪功能。
  Future<VoidResult> enableAudioAINS(bool enable);

  ///
  /// 开启/关闭音频共享。
  /// [enable] - 开启/关闭音频共享。
  ///
  Future<VoidResult> enableLoopbackRecording(bool enable);

  ///
  /// 调整共享音频音量。
  /// [volume] - 音量大小，取值范围为 [0, 100] 0：静音。100：（默认）原始音量。
  ///
  // Future<VoidResult> adjustLoopbackRecordingSignalVolume(int volume);
}

///
/// 用户音量信息
///
class NEMemberVolumeInfo {
  /// 用户id
  final String userUuid;

  /// 音量大小 [0-100]
  final int volume;

  NEMemberVolumeInfo(this.userUuid, this.volume);

  Map toJson() => {
        'userUuid': userUuid,
        'volume': volume,
      };
}

class NERoomVirtualBackgroundSourceStateReason {
  /// 虚拟背景开启成功。
  static const kSuccess = 0;

  /// 自定义背景图片不存在。 请检查 VirtualBackgroundSource 中 source 的值。
  static const kImageNotExist = 1;

  /// 自定义背景图片的图片格式无效。 请检查 VirtualBackgroundSource 中 source 的值
  static const kImageFormatNotSupported = 2;

  /// 自定义背景图片的颜色格式无效。 请检查 VirtualBackgroundSource 中 color 的值。
  static const kColorFormatNotSupported = 3;

  /// 该设备不支持使用虚拟背景。
  static const kDeviceNotSupported = 4;
}
