// Copyright (c) 2022 NetEase, Inc. All rights reserved.
// Use of this source code is governed by a MIT license that can be
// found in the LICENSE file.

part of netease_roomkit;

@HawkEntryPoint()
abstract class NERoomSeatController extends NERoomBaseController {
  /// 成员申请指定位置为[seatIndex]的麦位，位置从**1**开始。
  /// 如果当前成员为管理员，则会自动通过申请。
  /// @param seatIndex 麦位位置。
  /// @param exclusive 是否独占麦位。如果麦位为独占模式，则在管理员处理该请求之前，其他人无法申请该麦位。
  ///
  Future<VoidResult> submitSeatRequest(int seatIndex, bool exclusive);

  /// 成员取消申请麦位。
  ///
  Future<VoidResult> cancelSeatRequest();

  /// 成员主动下麦。
  ///
  Future<VoidResult> leaveSeat();

  /// 管理员同意成员[user]的上麦申请，非管理员执行该操作会失败。
  /// @param user 申请人的用户ID。
  ///
  Future<VoidResult> approveSeatRequest(String user);

  /// 管理员拒绝成员[user]的上麦请求，非管理员执行该操作会失败。
  /// @param user 申请人的用户ID。
  ///
  Future<VoidResult> rejectSeatRequest(String user);

  /// 管理员将成员[user]踢下麦，非管理员执行该操作会失败。
  /// @param user 麦上的用户ID。
  ///
  Future<VoidResult> kickSeat(String user);

  /// 管理员向成员[user]发送上麦邀请，指定位置为[seatIndex]，非管理员执行该操作会失败。
  /// @param seatIndex 麦位位置。
  /// @param user 麦上的用户ID。
  ///
  Future<VoidResult> sendSeatInvitation(int seatIndex, String user);

  /// 管理员取消成员[user]的上麦邀请，非管理员执行该操作会失败。
  /// @param user 麦上的用户ID。
  ///
  Future<VoidResult> cancelSeatInvitation(String user);

  /// 成员同意管理员的上麦邀请。
  ///
  Future<VoidResult> acceptSeatInvitation();

  /// 成员拒绝管理员的上麦邀请。
  ///
  Future<VoidResult> rejectSeatInvitation();

  /// 管理员关闭指定位置的麦位，非管理员执行该操作会失败。
  /// @param seatIndices 待关闭的麦位位置列表。
  ///
  Future<VoidResult> closeSeats(List<int> seatIndices);

  /// 管理员打开指定位置的麦位。非管理员执行该操作会失败。
  /// @param seatIndices 待打开的麦位位置列表。
  ///
  Future<VoidResult> openSeats(List<int> seatIndices);

  /// 同步麦位信息。
  ///
  Future<NEResult<NESeatInfo>> getSeatInfo();

  /// 获取麦位申请列表。按照申请时间正序排序，先申请的成员排在列表前面。
  ///
  Future<NEResult<List<NESeatRequestItem>>> getSeatRequestList();

  /// 获取麦位邀请列表。按照邀请时间正序排序，先邀请的成员排在列表前面。
  ///
  Future<NEResult<List<NESeatInvitationItem>>> getSeatInvitationList();

  void addEventCallback(NESeatEventCallback callback);
  void removeEventCallback(NESeatEventCallback callback);
}

typedef SeatManagerAddedCallback = void Function(List<String?> managers);
typedef SeatManagerRemovedCallback = void Function(List<String?> managers);
typedef SeatRequestSubmittedCallback = void Function(
    int seatIndex, String user);
typedef SeatRequestCancelledCallback = void Function(
    int seatIndex, String user);
typedef SeatRequestApprovedCallback = void Function(
    int seatIndex, String user, String operateBy, bool isAutoAgree);
typedef SeatRequestRejectedCallback = void Function(
    int seatIndex, String user, String operateBy);
typedef SeatInvitationReceivedCallback = void Function(
    int seatIndex, String user, String operateBy);
typedef SeatInvitationCancelledCallback = void Function(
    int seatIndex, String user, String operateBy);
typedef SeatInvitationAcceptedCallback = void Function(
    int seatIndex, String user, bool isAutoAgree);
typedef SeatInvitationRejectedCallback = void Function(
    int seatIndex, String user);
typedef SeatLeaveCallback = void Function(int seatIndex, String user);
typedef SeatKickedCallback = void Function(
    int seatIndex, String user, String operateBy);
typedef SeatListChangedCallback = void Function(List<NESeatItem> seatItems);

/// 麦位监听器
class NESeatEventCallback {
  final SeatManagerAddedCallback? seatManagerAddedCallback;

  final SeatManagerRemovedCallback? seatManagerRemovedCallback;

  final SeatRequestSubmittedCallback? seatRequestSubmittedCallback;

  final SeatRequestCancelledCallback? seatRequestCancelledCallback;

  final SeatRequestApprovedCallback? seatRequestApprovedCallback;

  final SeatRequestRejectedCallback? seatRequestRejectedCallback;

  final SeatInvitationReceivedCallback? seatInvitationReceivedCallback;

  final SeatInvitationCancelledCallback? seatInvitationCancelledCallback;

  final SeatInvitationAcceptedCallback? seatInvitationAcceptedCallback;

  final SeatInvitationRejectedCallback? seatInvitationRejectedCallback;

  final SeatLeaveCallback? seatLeaveCallback;

  final SeatKickedCallback? seatKickedCallback;

  final SeatListChangedCallback? seatListChangedCallback;

  NESeatEventCallback({
    this.seatManagerAddedCallback,
    this.seatManagerRemovedCallback,
    this.seatRequestSubmittedCallback,
    this.seatRequestCancelledCallback,
    this.seatRequestApprovedCallback,
    this.seatRequestRejectedCallback,
    this.seatInvitationReceivedCallback,
    this.seatInvitationCancelledCallback,
    this.seatInvitationAcceptedCallback,
    this.seatInvitationRejectedCallback,
    this.seatLeaveCallback,
    this.seatKickedCallback,
    this.seatListChangedCallback,
  });
}
