// Copyright (c) 2022 NetEase, Inc. All rights reserved.
// Use of this source code is governed by a MIT license that can be
// found in the LICENSE file.

part of netease_roomkit;

/// 最大缓存render数量
/// 低端手机最多只能创建 16 个 EGLContext
const int _maxRendererCacheSize = 6;

/// 最多允许同时存在的视频视图数量
// const int _maxActiveRendererSize = 10;

/// 延迟解除订阅
/// 解决频繁订阅问题
const int _unsubscribeDelayInSeconds = 10;

const int _unsubscribeDelayShortInSeconds = 2;

class _RoomUserCreateRenderRequest {
  final String userId;
  final bool subStream;

  _RoomUserCreateRenderRequest(this.userId, this.subStream);

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
      other is _RoomUserCreateRenderRequest &&
          runtimeType == other.runtimeType &&
          userId == other.userId &&
          subStream == other.subStream;

  @override
  int get hashCode => userId.hashCode ^ subStream.hashCode;
}

class _RoomUserVideoStreamSubscription {
  final String userId;
  final NEVideoStreamType streamType;
  final Set<dynamic> owners = {};
  final NERtcVideoRenderer renderer;

  _RoomUserVideoStreamSubscription(this.userId, this.streamType, this.renderer);
}

int _defaultRenderCacheSize(int? requestedSize) {
  if (requestedSize != null && requestedSize > 0) {
    return requestedSize;
  }
  return Platform.isAndroid ? _maxRendererCacheSize : -1;
}

/// An InheritedWidget which provides [NERoomContext] instance
class NERoomContextProvider extends InheritedWidget {
  final NERoomContext roomContext;

  NERoomContextProvider({
    Key? key,
    required Widget child,
    required this.roomContext,
    int? renderCacheSize,
  }) : super(
          key: key,
          child: _RoomUserVideoStreamSubscriberWrapper(
            child: child,
            renderCacheSize: renderCacheSize,
          ),
        );

  static NERoomContextProvider of(BuildContext context) {
    final result =
        context.dependOnInheritedWidgetOfExactType<NERoomContextProvider>();
    assert(result != null,
        'No NERoomUserVideoControllerProvider found in context');
    return result!;
  }

  @override
  bool updateShouldNotify(NERoomContextProvider old) {
    return roomContext != old.roomContext;
  }
}

class _RoomUserVideoStreamSubscriberWrapper extends StatefulWidget {
  final Widget child;
  final int? renderCacheSize;

  const _RoomUserVideoStreamSubscriberWrapper({
    Key? key,
    required this.child,
    this.renderCacheSize,
  }) : super(key: key);

  @override
  __RoomUserVideoStreamSubscriberWrapperState createState() =>
      __RoomUserVideoStreamSubscriberWrapperState();
}

class __RoomUserVideoStreamSubscriberWrapperState
    extends State<_RoomUserVideoStreamSubscriberWrapper> {
  NERoomUserVideoStreamSubscriber? subscriber;

  @override
  void didChangeDependencies() {
    super.didChangeDependencies();
    subscriber?.dispose();
    subscriber = NERoomUserVideoStreamSubscriber(
      NERoomContextProvider.of(context).roomContext,
      widget.renderCacheSize,
    );
  }

  @override
  Widget build(BuildContext context) {
    return NERoomUserVideoStreamSubscriberProvider(
      child: widget.child,
      subscriber: subscriber!,
    );
  }

  @override
  void dispose() {
    subscriber?.dispose();
    super.dispose();
  }
}

class NERoomUserVideoStreamSubscriberProvider extends InheritedWidget {
  final NERoomUserVideoStreamSubscriber subscriber;

  const NERoomUserVideoStreamSubscriberProvider({
    Key? key,
    required Widget child,
    required this.subscriber,
  }) : super(key: key, child: child);

  static NERoomUserVideoStreamSubscriberProvider of(BuildContext context) {
    final result = context.dependOnInheritedWidgetOfExactType<
        NERoomUserVideoStreamSubscriberProvider>();
    assert(result != null,
        'No _RoomUserVideoStreamSubscriberProvider found in context');
    return result!;
  }

  @override
  bool updateShouldNotify(NERoomUserVideoStreamSubscriberProvider old) {
    return subscriber != old.subscriber;
  }
}

/// be responsible for subscribe/unsubscribe video (sub)stream
class NERoomUserVideoStreamSubscriber {
  final NERoomContext roomContext;
  final int renderCacheSize;

  late final NERoomRtcController rtcController;
  bool _disposed = false;

  NERoomUserVideoStreamSubscriber(this.roomContext, [int? _renderCacheSize])
      : renderCacheSize = _defaultRenderCacheSize(_renderCacheSize) {
    rtcController = roomContext.rtcController;
  }

  // 主流订阅信息
  final _mainStreamSubscriptions = <String, _RoomUserVideoStreamSubscription>{};
  // 辅流订阅信息
  final _subStreamSubscriptions = <String, _RoomUserVideoStreamSubscription>{};

  int _pendingUnsubscribeVideoCount = 0;

  int _adjustUnsubscribeVideoDelay() {
    if (_pendingUnsubscribeVideoCount > 8) {
      return 0;
    } else if (_pendingUnsubscribeVideoCount > 4) {
      return _unsubscribeDelayShortInSeconds;
    }
    return _unsubscribeDelayInSeconds;
  }

  final _renderCache = ListQueue<NERtcVideoRenderer>();
  final _renderAvailableEvent = StreamController<int>.broadcast();
  int _curRenderSize = 0;

  Future<void> _recycleRenderer(
      NERtcVideoRenderer renderer, String? debugName) async {
    Alog.i(
      tag: 'RoomUserVideoStreamSubscriber',
      content:
          'recycle render: $debugName $_curRenderSize ${_renderCache.length}',
    );
    renderer.rendererEventLister = null;
    if (_renderCache.length >= renderCacheSize ||
        renderCacheSize == -1 ||
        _disposed) {
      await renderer.dispose();
    } else {
      await renderer.detach();
      if (!_disposed) {
        _renderCache.add(renderer);
        _renderAvailableEvent.add(0);
      } else {
        await renderer.dispose();
      }
    }
  }

  final _renderCreateRequests =
      <_RoomUserCreateRenderRequest, Future<NERtcVideoRenderer>>{};

  dynamic _dequeueRenderer(dynamic owner, String userId, bool subStream,
      [String? debugName]) async {
    final request = _RoomUserCreateRenderRequest(userId, subStream);

    dynamic result;
    String? type;
    final subInfo =
        subStream ? _subStreamSubscriptions : _mainStreamSubscriptions;
    // 复用render
    if (subInfo.containsKey(userId)) {
      subInfo[userId]!.owners.add(owner);
      result = subInfo[userId]!.renderer;
      type = 'use binding render';
    } else if (_renderCreateRequests.containsKey(request)) {
      result = _renderCreateRequests[request]!;
      type = 'duplicate create request';
    } else if (_renderCache.isNotEmpty) {
      final renderer = _renderCache.removeFirst();
      result = renderer;
      type = 'use cache render';
    } else if (_curRenderSize < renderCacheSize || renderCacheSize == -1) {
      _curRenderSize++;
      result =
          NERtcVideoRendererFactory.createVideoRenderer(roomContext.roomUuid);
      _renderCreateRequests[request] = result;
      result.whenComplete(() {
        _renderCreateRequests.remove(request);
      });
      type = 'create render';
    } else {
      final rendererCompleter = Completer<NERtcVideoRenderer>();
      final subscription = _renderAvailableEvent.stream.listen(null);
      subscription.onData((data) {
        if (_renderCache.isNotEmpty) {
          final renderer = _renderCache.removeFirst();
          subscription.cancel();
          Alog.i(
            tag: 'RoomUserVideoStreamSubscriber',
            content:
                'render available for ${debugName ?? userId}: cur=$_curRenderSize, cache=${_renderCache.length}',
          );
          rendererCompleter.complete(renderer);
        }
      });
      result = rendererCompleter.future;
      _renderCreateRequests[request] = result;
      result.whenComplete(() {
        _renderCreateRequests.remove(request);
      });
      type = 'waiting for render';
    }
    Alog.i(
      tag: 'RoomUserVideoStreamSubscriber',
      content:
          '$type for ${debugName ?? userId}: cur=$_curRenderSize, cache=${_renderCache.length}',
    );
    return result;
  }

  // 是否需要重新订阅
  bool _isReSubscribeNecessary(
      NEVideoStreamType oldStreamType, NEVideoStreamType newStreamType) {
    return oldStreamType != newStreamType &&
        newStreamType == NEVideoStreamType.kHigh;
  }

  bool _isLocalUser(String userId) => roomContext.localMember.uuid == userId;

  void _subscribeVideoStream(dynamic owner, NERtcVideoRenderer renderer,
      String userId, NEVideoStreamType streamType, bool subStream,
      [String? debugName]) {
    final isLocal = _isLocalUser(userId);
    // 1. 先绑定画布
    if (subStream) {
      isLocal
          ? renderer.attachToLocalSubStreamVideo()
          : renderer.attachToRemoteSubStreamVideo(userId);
    } else {
      isLocal
          ? renderer.attachToLocalVideo()
          : renderer.attachToRemoteVideo(userId);
    }

    // 2. 检查是否需要订阅
    final subInfo =
        subStream ? _subStreamSubscriptions : _mainStreamSubscriptions;
    var subscription = subInfo[userId];
    if (subscription == null ||
        _isReSubscribeNecessary(subscription.streamType, streamType)) {
      Alog.i(
        tag: 'RoomUserVideoStreamSubscriber',
        content:
            'subscribeVideoStream: ${debugName ?? userId} $streamType $subStream',
      );
      final updated =
          _RoomUserVideoStreamSubscription(userId, streamType, renderer);
      subInfo[userId] = updated;
      updated.owners.addAll(subscription?.owners ?? []);
      subscription = updated;
      if (!isLocal) {
        if (subStream) {
          rtcController.subscribeRemoteVideoSubStream(userId);
        } else {
          rtcController.subscribeRemoteVideoStream(userId, streamType);
        }
      }
    }
    subscription.owners.add(owner);
  }

  Future<void> _unsubscribeVideoStream(
      dynamic owner,
      NERtcVideoRenderer? renderer,
      String userId,
      NEVideoStreamType? streamType,
      bool subStream,
      [String? debugName]) async {
    // recycle render if possible
    var subInfo =
        subStream ? _subStreamSubscriptions : _mainStreamSubscriptions;
    var exist = subInfo[userId];
    exist?.owners.remove(owner);
    if (exist?.renderer != renderer) {
      if (renderer != null) {
        _recycleRenderer(renderer, debugName);
      }
      return;
    }

    if (exist != null && exist.owners.isEmpty) {
      subInfo.remove(userId);
      if (renderer != null) {
        Alog.i(
          tag: 'RoomUserVideoStreamSubscriber',
          content:
              'dispose render: ${debugName ?? userId} $streamType $subStream; pendingUnsub=$_pendingUnsubscribeVideoCount',
        );
        await _recycleRenderer(renderer, debugName);
      }
    } else if (renderer != null && exist?.renderer != renderer) {
      await _recycleRenderer(renderer, debugName);
    }
    if (_disposed) return;

    // delay unsubscribe
    final delay = _adjustUnsubscribeVideoDelay();
    if (delay > 0) {
      _pendingUnsubscribeVideoCount++;
      await Future.delayed(Duration(seconds: delay));
      _pendingUnsubscribeVideoCount--;
    }

    if (_disposed) return;
    subInfo = subStream ? _subStreamSubscriptions : _mainStreamSubscriptions;
    exist = subInfo[userId];
    if (exist == null) {
      Alog.i(
        tag: 'RoomUserVideoStreamSubscriber',
        content:
            'unsubscribeVideoStream: ${debugName ?? userId} $streamType $subStream',
      );
      if (!_isLocalUser(userId)) {
        if (subStream) {
          await rtcController.unsubscribeRemoteVideoSubStream(userId);
        } else {
          assert(streamType != null);
          await rtcController.unsubscribeRemoteVideoStream(userId, streamType!);
        }
      }
    }
  }

  void dispose() {
    Alog.i(
      tag: 'RoomUserVideoStreamSubscriber',
      content: 'dispose',
    );
    _disposed = true;
    _subStreamSubscriptions.clear();
    _mainStreamSubscriptions.clear();
    for (var element in _renderCache) {
      element.dispose();
    }
    _renderCache.clear();
    _renderAvailableEvent.close();
  }
}

class _UserVideoKey extends LocalKey {
  const _UserVideoKey(this.userId, this.subStream);

  final String userId;
  final bool subStream;

  @override
  bool operator ==(Object other) {
    if (other.runtimeType != runtimeType) {
      return false;
    }
    return other is _UserVideoKey &&
        userId == other.userId &&
        subStream == other.subStream;
  }

  @override
  int get hashCode => Object.hash(runtimeType, userId, subStream);
}

/// user video view widget
/// it automatically subscribe and unsubscribe video stream for user
class NERoomUserVideoView extends StatefulWidget {
  final NERoomUserVideoViewListener? listener;
  final NERtcVideoViewFitType fitType;
  final Color backgroundColor;
  final bool mirror;
  final String userId;
  final bool subStream;
  final NEVideoStreamType streamType;
  final String? debugName;
  final WidgetBuilder? builder;
  final bool keepAlive;

  /// 正常流
  NERoomUserVideoView(
    this.userId, {
    Key? key,
    this.debugName,
    this.listener,
    this.mirror = false,
    NEVideoStreamType? streamType,
    this.fitType = NERtcVideoViewFitType.contain,
    this.builder,
    this.keepAlive = false,
    Color? backgroundColor,
  })  : subStream = false,
        streamType = streamType ?? NEVideoStreamType.kLow,
        this.backgroundColor = backgroundColor ?? const Color(0xFF292933),
        super(key: key ?? _UserVideoKey(userId, false));

  /// 辅流
  NERoomUserVideoView.subStream(
    this.userId, {
    Key? key,
    this.debugName,
    this.listener,
    this.mirror = false,
    this.keepAlive = false,
    this.builder,
    this.fitType = NERtcVideoViewFitType.contain,
    Color? backgroundColor,
  })  : streamType = NEVideoStreamType.kHigh,
        subStream = true,
        this.backgroundColor = backgroundColor ?? const Color(0xFF292933),
        super(key: key ?? _UserVideoKey(userId, true));

  @override
  _RoomUserVideoViewState createState() => _RoomUserVideoViewState();
}

class _RoomUserVideoViewState extends State<NERoomUserVideoView>
    with NERtcVideoRendererEventListener, AutomaticKeepAliveClientMixin {
  NERtcVideoRenderer? renderer;
  late NERoomUserVideoStreamSubscriber subscriber;
  late String debugName;

  @override
  bool get wantKeepAlive => widget.keepAlive;

  void setupRenderer() {
    final renderer = this.renderer;
    assert(renderer != null, 'setup renderer error');
    print(
        'RoomUserVideoViewState setupRenderer: ${renderer?.textureId} $debugName');
    if (renderer != null) {
      renderer.setMirror(widget.mirror);
      renderer.rendererEventLister = this;
      subscriber._subscribeVideoStream(this, renderer, widget.userId,
          widget.streamType, widget.subStream, debugName);
    }
  }

  @override
  void initState() {
    super.initState();
    debugName = '${widget.debugName}@${widget.userId}';
    Alog.i(
      tag: 'RoomUserVideoViewState',
      content:
          'init for $debugName ${widget.mirror} ${widget.subStream} ${widget.streamType}',
    );
  }

  @override
  void didUpdateWidget(NERoomUserVideoView oldWidget) {
    super.didUpdateWidget(oldWidget);
    if (widget.mirror != oldWidget.mirror ||
        widget.streamType != oldWidget.streamType) {
      setupRenderer();
    }
    if (oldWidget.listener != widget.listener && widget.listener != null) {
      _maybeNotifyListener();
    }
    if (oldWidget.keepAlive != widget.keepAlive) {
      updateKeepAlive();
    }
  }

  @override
  void didChangeDependencies() {
    super.didChangeDependencies();
    if (renderer == null) {
      subscriber =
          NERoomUserVideoStreamSubscriberProvider.of(context).subscriber;
      final renderOrFuture = subscriber._dequeueRenderer(
          this, widget.userId, widget.subStream, debugName);
      assert(renderOrFuture is NERtcVideoRenderer || renderOrFuture is Future);
      if (renderOrFuture is NERtcVideoRenderer) {
        renderer = renderOrFuture;
        setupRenderer();
      } else {
        final future = renderOrFuture as Future;
        future.then((value) {
          if (mounted) {
            renderer = value as NERtcVideoRenderer;
            setState(setupRenderer);
          } else {
            subscriber._unsubscribeVideoStream(this, value, widget.userId,
                widget.streamType, widget.subStream, debugName);
          }
        });
      }
    }
  }

  @override
  void dispose() {
    Alog.i(
      tag: 'RoomUserVideoViewState',
      content: 'dispose for $debugName ${widget.subStream}',
    );
    subscriber._unsubscribeVideoStream(this, renderer, widget.userId,
        widget.streamType, widget.subStream, debugName);
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    super.build(context);
    final renderer = this.renderer;
    if (renderer == null) {
      if (widget.builder == null) {
        return Container(
          color: widget.backgroundColor,
        );
      } else {
        return widget.builder!(context);
      }
    }
    return NERtcVideoView(
      renderer,
      fitType: widget.fitType,
      backgroundColor: widget.backgroundColor,
    );
  }

  @override
  void onFirstFrameRendered() {
    Alog.i(
      tag: 'RoomUserVideoViewState',
      content:
          'onFirstFrameRendered for $debugName ${widget.streamType} ${widget.subStream}',
    );
    widget.listener?.onFirstFrameRendered(widget.userId);
  }

  @override
  void onFrameResolutionChanged(int width, int height, int rotation) {
    Alog.i(
      tag: 'RoomUserVideoViewState',
      content:
          'onFrameResolutionChanged for $debugName ${widget.streamType} ${widget.subStream}',
    );
    widget.listener
        ?.onFrameResolutionChanged(widget.userId, width, height, rotation);
  }

  void _maybeNotifyListener() {
    final renderer = this.renderer;
    if (renderer != null) {
      final value = renderer.value;
      if (value.width != 0 && value.height != 0) {
        widget.listener!.onFirstFrameRendered(widget.userId);
        widget.listener!.onFrameResolutionChanged(
            widget.userId, value.width, value.height, value.rotation);
      }
    }
  }
}

/// 视频渲染事件监听器
abstract class NERoomUserVideoViewListener {
  void onFrameResolutionChanged(
      String userId, int width, int height, int rotation) {}

  void onFirstFrameRendered(String userId) {}
}
