// Copyright (c) 2022 NetEase, Inc. All rights reserved.
// Use of this source code is governed by a MIT license that can be
// found in the LICENSE file.

part of netease_roomkit;

class _NERoomContextImpl extends NERoomContext
    with _AloggerMixin, RoomEventSink, RoomStatsSink {
  final _platform = RoomContextPlatform.instance;
  final _roomStatsPlatform = RoomStatsPlatform.instance;

  bool isConnected = false;

  late String _roomUuid;
  late String _roomName;
  String? _password;
  late final int _rtcStartTime;
  late bool _isRoomLocked;
  late String? _sipCid;
  late double? _remainingSeconds;
  final _properties = <String, String>{};
  late _NERoomMemberImpl _local;
  late final String localUuid;
  final Map<String, _NERoomMemberImpl> _uid2Member = {};
  final _eventCallbacks = <NERoomEventCallback>{};
  final _statsCallbacks = <NERoomRtcStatsCallback>{};

  late _NERoomChatControllerImpl _chatController;
  late _NERoomWhiteboardControllerImpl _whiteboardController;
  late _NERoomRtcControllerImpl _rtcController;
  late _NELRoomLiveControllerImpl _liveController;
  late _NERoomSeatControllerImpl _seatController;

  _NERoomMemberImpl? screenSharingMember;
  _NERoomMemberImpl? whiteboardSharingMember;

  VoidCallback? onDispose;

  _NERoomContextImpl(String roomUuid) {
    _roomUuid = roomUuid;
  }

  void initState(RoomSnapshot snapshot, {required VoidCallback onDispose}) {
    commonLogger.i('initState');
    isConnected = true;
    _platform.registerEventListener(roomUuid, this);
    _roomStatsPlatform.registerStatsListener(roomUuid, this);
    this.onDispose = onDispose;
    _roomUuid = snapshot.roomUuid;
    _roomName = snapshot.roomName;
    _password = snapshot.password;
    _rtcStartTime = snapshot.rtcStartTime;
    _isRoomLocked = snapshot.isRoomLocked;
    _sipCid = snapshot.sipCid;
    _remainingSeconds = snapshot.remainingSeconds;
    _properties.addAll(snapshot.properties?.cast() ?? {});
    _local = createMember(snapshot.localMember!);
    localUuid = _local.uuid;
    _uid2Member[localUuid] = _local;
    _chatController =
        _NERoomChatControllerImpl(snapshot.isChatroomSupported, _roomUuid);
    _whiteboardController =
        _NERoomWhiteboardControllerImpl(snapshot.isWhiteboardSupported, this);
    _rtcController = _NERoomRtcControllerImpl(snapshot.isRtcSupported, this);
    _liveController =
        _NELRoomLiveControllerImpl(snapshot.isLiveSupported, roomUuid);
    _seatController =
        _NERoomSeatControllerImpl(snapshot.isSeatSupported, _roomUuid);
    for (var element in snapshot.remoteMembers) {
      if (element != null) {
        _uid2Member[element.uuid] = createMember(element);
      }
    }
    _uid2Member[_local.uuid] = _local;
  }

  void closeRoomInner(NERoomEndReason reason) {
    for (var callback in _eventCallbacks.copy()) {
      callback.roomEnd?.call(reason);
    }
    Timer.run(dispose);
  }

  void dispose() {
    commonLogger.i('dispose');
    if (!isConnected) return;
    isConnected = false;
    _eventCallbacks.clear();
    _statsCallbacks.clear();
    _platform.unregisterEventListener(roomUuid, this);
    _roomStatsPlatform.unregisterStatsListener(roomUuid, this);

    _seatController.dispose();

    onDispose?.call();
    onDispose = null;
    clearAttachments();
  }

  @override
  String get roomUuid => _roomUuid;

  @override
  String get roomName => _roomName;

  @override
  String? get password => _password;

  @override
  int get rtcStartTime => _rtcStartTime;

  @override
  bool get isRoomLocked => _isRoomLocked;

  @override
  String? get sipCid => _sipCid;

  @override
  double? get remainingSeconds => _remainingSeconds;

  @override
  NERoomMember get localMember => _local;

  @override
  List<NERoomMember> get remoteMembers {
    final dup = Map<String, _NERoomMemberImpl>.from(_uid2Member)
      ..remove(localUuid);
    return dup.values.toList();
  }

  @override
  NERoomMember? getMember(String? uuid) => _uid2Member[uuid];

  @override
  void addEventCallback(NERoomEventCallback callback) {
    _eventCallbacks.add(callback);
  }

  @override
  void removeEventCallback(NERoomEventCallback callback) {
    _eventCallbacks.remove(callback);
  }

  @override
  void addRtcStatsCallback(NERoomRtcStatsCallback callback) {
    _statsCallbacks.add(callback);
  }

  @override
  void removeRtcStatsCallback(NERoomRtcStatsCallback callback) {
    _statsCallbacks.remove(callback);
  }

  @override
  Future<VoidResult> changeMyName(String name) {
    apiLogger.i('changeMyName: $roomUuid, $name');
    return _platform.changeMyName(roomUuid, name);
  }

  @override
  Future<VoidResult> handOverMyRole(String userUuid) {
    apiLogger.i('handOverMyRole: $roomUuid, $userUuid');
    return _platform.handOverMyRole(roomUuid, userUuid);
  }

  @override
  Future<VoidResult> changeMemberRole(String userUuid, String role) {
    apiLogger.i('changeMemberRole: $roomUuid, $userUuid, $role');
    return _platform.changeMemberRole(roomUuid, userUuid, role);
  }

  @override
  Future<VoidResult> kickMemberOut(String userUuid) {
    apiLogger.i('kickMemberOut: $roomUuid, $userUuid');
    return _platform.kickMemberOut(roomUuid, userUuid);
  }

  @override
  Future<VoidResult> leaveRoom() {
    apiLogger.i('leaveRoom: $roomUuid');
    Timer.run(() => closeRoomInner(NERoomEndReason.kLeaveBySelf));
    return _platform.leaveRoom(roomUuid);
  }

  @override
  Future<VoidResult> endRoom([bool isForce = false]) {
    apiLogger.i('endRoom: $roomUuid');
    return _platform.endRoom(roomUuid, isForce);
  }

  @override
  Future<VoidResult> lockRoom() {
    apiLogger.i('lockRoom: $roomUuid');
    return _platform.lockRoom(roomUuid);
  }

  @override
  Future<VoidResult> unlockRoom() {
    apiLogger.i('unlockRoom: $roomUuid');
    return _platform.unlockRoom(roomUuid);
  }

  @override
  Future<VoidResult> updateRoomProperty(String key, String value,
      {String? associatedUuid}) {
    apiLogger.i('updateRoomProperty: $key, $value, $associatedUuid');
    return _platform.updateRoomProperty(roomUuid, key, value,
        associatedUuid: associatedUuid);
  }

  @override
  Future<VoidResult> deleteRoomProperty(String key) {
    apiLogger.i('deleteRoomProperty: $key');
    return _platform.deleteRoomProperty(roomUuid, key);
  }

  @override
  Future<VoidResult> updateMemberProperty(
      String userUuid, String key, String value) {
    apiLogger.i('updateMemberProperty: $userUuid, $key, $value');
    return _platform.updateMemberProperty(roomUuid, userUuid, key, value);
  }

  @override
  Future<VoidResult> deleteMemberProperty(String userUuid, String key) {
    apiLogger.i('deleteMemberProperty: $userUuid, $key');
    return _platform.deleteMemberProperty(roomUuid, userUuid, key);
  }

  @override
  NERoomChatController get chatController => _chatController;

  @override
  NERoomRtcController get rtcController => _rtcController;

  @override
  NERoomWhiteboardController get whiteboardController => _whiteboardController;

  @override
  NERoomLiveController get liveController => _liveController;

  @override
  NERoomSeatController get seatController => _seatController;

  @override
  Map<String, String> get roomProperties => Map.from(_properties);

  // **************************************************************************

  _NERoomMemberImpl createMember(RoomMember member) {
    final _member = _NERoomMemberImpl(
      uuid: member.uuid,
      name: member.name,
      avatar: member.avatar,
      role: _createRole(member.role!),
      isAudioOn: member.isAudioOn,
      isVideoOn: member.isVideoOn,
      isInRtcChannel: member.isInRtcChannel,
      isInChatroom: member.isInChatroom,
      isSharingScreen: member.isSharingScreen,
      isSharingWhiteboard: member.isSharingWhiteboard,
      clientType: NEClientTypeExtension.fromType(member.clientType),

      ///dart枚举接口规范，全部使用小写
      props: member.properties?.cast() ?? {},
    );
    if (_member.isSharingScreen) {
      screenSharingMember = _member;
    }
    if (_member.isSharingWhiteboard) {
      whiteboardSharingMember = _member;
    }
    return _member;
  }

  NERoomRole _createRole(RoomRole role) {
    return NERoomRole(
      name: role.name,
      limit: role.limit,
    );
  }

  _NERoomMemberImpl? removeMember(String? uuid) {
    if (uuid == screenSharingMember?.uuid) {
      screenSharingMember = null;
    }
    if (uuid == whiteboardSharingMember?.uuid) {
      whiteboardSharingMember = null;
    }
    final member = _uid2Member.remove(uuid);
    if (member != null) {
      member.clearAttachments();
    }
    return member;
  }

  _NERoomMemberImpl? updateMemberIfExists(
      String? userUuid, void Function(_NERoomMemberImpl) action) {
    final member = _uid2Member[userUuid];
    if (member != null) {
      action(member);
    }
    return member;
  }

  List<_NERoomMemberImpl> updateMembersIfExists(
      List<String?> userUuids, void Function(_NERoomMemberImpl) action) {
    final members = <_NERoomMemberImpl>[];
    for (var userUuid in userUuids) {
      if (userUuid == null) continue;
      final member = _uid2Member[userUuid];
      if (member != null) {
        action(member);
        members.add(member);
      }
    }
    return members;
  }

  // **************************************************************************

  @override
  void onRoomPropertiesChanged(PropertyEvent event) {
    if (!isConnected) return;
    commonLogger.i('onRoomPropertiesChanged: ${event.properties}');
    _properties.addAll(event.properties.cast());
    for (var callback in _eventCallbacks.copy()) {
      callback.roomPropertiesChanged?.call(event.properties.cast());
    }
  }

  @override
  void onRoomPropertiesDeleted(PropertyEvent event) {
    if (!isConnected) return;
    commonLogger.i('onRoomPropertiesDeleted: ${event.properties}');
    _properties.removeWhere((key, value) => event.properties.containsKey(key));
    for (var callback in _eventCallbacks.copy()) {
      callback.roomPropertiesDeleted?.call(event.properties.cast());
    }
  }

  @override
  void onMemberPropertiesChanged(PropertyEvent event) {
    if (!isConnected) return;
    assert(event.userUuid != null);
    commonLogger
        .i('onMemberPropertiesChanged: ${event.userUuid} ${event.properties}');
    final member = updateMemberIfExists(event.userUuid, (member) {
      member.props.addAll(event.properties.cast());
    });
    if (member != null) {
      for (var callback in _eventCallbacks.copy()) {
        callback.memberPropertiesChanged?.call(member, event.properties.cast());
      }
    }
  }

  @override
  void onMemberPropertiesDeleted(PropertyEvent event) {
    if (!isConnected) return;
    assert(event.userUuid != null);
    commonLogger
        .i('onMemberPropertiesDeleted: ${event.userUuid} ${event.properties}');
    final member = updateMemberIfExists(event.userUuid, (member) {
      member.props
          .removeWhere((key, value) => event.properties.containsKey(key));
    });
    if (member != null) {
      for (var callback in _eventCallbacks.copy()) {
        callback.memberPropertiesDeleted?.call(member, event.properties.cast());
      }
    }
  }

  @override
  void onMemberRoleChanged(MemberRoleChangeEvent event) {
    assert(event.roomUuid == roomUuid);
    assert(event.role != null);
    if (!isConnected) return;
    commonLogger.i('onMemberRoleChanged');
    final member = _uid2Member[event.userUuid];
    if (member != null) {
      final beforeRole = member.role;
      member.role = _createRole(event.role!);
      for (var callback in _eventCallbacks.copy()) {
        callback.memberRoleChanged?.call(member, beforeRole, member.role);
      }
    }
  }

  @override
  void onMemberNameChanged(CommonEvent event) {
    assert(event.roomUuid == roomUuid);
    assert(event.userUuid != null);
    if (!isConnected) return;
    commonLogger.i('onMemberNameChanged: ${event.userUuid} ${event.sArg1}');
    final member = updateMemberIfExists(event.userUuid, (mem) {
      mem.name = event.sArg1!;
    });
    if (member != null) {
      for (var callback in _eventCallbacks.copy()) {
        callback.memberNameChanged?.call(member, member.name);
      }
    }
  }

  @override
  void onMemberJoinRoom(MembersEvent event) {
    assert(event.members != null);
    if (!isConnected) return;
    commonLogger
        .i('onMemberJoinRoom: ${event.members!.map((e) => e?.uuid).toList()}');
    final adds = <NERoomMember>[];
    for (var element in event.members!) {
      if (element != null) {
        final member = createMember(element);
        _uid2Member[element.uuid] = member;
        adds.add(member);
      }
    }
    for (var callback in _eventCallbacks.copy()) {
      callback.memberJoinRoom?.call(adds);
    }
  }

  @override
  void onMemberLeaveRoom(MembersEvent event) {
    assert(event.uuids != null);
    if (!isConnected) return;
    commonLogger.i('onMemberLeaveRoom: ${event.uuids}');
    final leaves = event.uuids!
        .map((uuid) => removeMember(uuid))
        .whereType<NERoomMember>()
        .toList();
    if (leaves.isNotEmpty) {
      for (var callback in _eventCallbacks.copy()) {
        callback.memberLeaveRoom?.call(leaves);
      }
    }
  }

  @override
  void onMemberJoinRtcChannel(MembersEvent event) {
    assert(event.uuids != null);
    if (!isConnected) return;
    commonLogger.i('onMemberJoinRtcChannel: ${event.uuids}');
    final joins = updateMembersIfExists(event.uuids!, (member) {
      member.isInRtcChannel = true;
    });
    if (joins.isNotEmpty) {
      for (var callback in _eventCallbacks.copy()) {
        callback.memberJoinRtcChannel?.call(joins);
      }
    }
  }

  @override
  void onMemberLeaveRtcChannel(MembersEvent event) {
    assert(event.uuids != null);
    if (!isConnected) return;
    commonLogger.i('onMemberLeaveRtcChannel: ${event.uuids}');
    final leaves = updateMembersIfExists(event.uuids!, (member) {
      member.isInRtcChannel = false;
    });
    if (leaves.isNotEmpty) {
      for (var callback in _eventCallbacks.copy()) {
        callback.memberLeaveRtcChannel?.call(leaves);
      }
    }
  }

  @override
  void onRtcChannelError(CommonEvent event) {
    assert(event.arg1 != null);
    if (!isConnected) return;
    commonLogger.i('onRtcChannelError: ${event.arg1}');
    for (var callback in _eventCallbacks.copy()) {
      callback.rtcChannelError?.call(event.arg1!);
    }
  }

  @override
  void onMemberJoinChatroom(MembersEvent event) {
    assert(event.uuids != null);
    if (!isConnected) return;
    commonLogger.i('onMemberJoinChatroom');
    final joins = updateMembersIfExists(event.uuids!, (member) {
      member.isInChatroom = true;
    });
    for (var callback in _eventCallbacks.copy()) {
      callback.memberJoinChatroom?.call(joins);
    }
  }

  @override
  void onMemberLeaveChatroom(MembersEvent event) {
    assert(event.uuids != null);
    if (!isConnected) return;
    commonLogger.i('onMemberLeaveChatroom');
    final leaves = updateMembersIfExists(event.uuids!, (member) {
      member.isInChatroom = false;
    });
    for (var callback in _eventCallbacks.copy()) {
      callback.memberLeaveChatroom?.call(leaves);
    }
  }

  @override
  void onMemberAudioMuteChanged(MemberToggleEvent event) {
    if (!isConnected) return;
    commonLogger.i('onMemberAudioMuteChanged: ${event.userUuid} ${event.isOn}');
    final member = updateMemberIfExists(event.userUuid, (member) {
      member.isAudioOn = event.isOn;
    });
    if (member != null) {
      for (var callback in _eventCallbacks.copy()) {
        callback.memberAudioMuteChanged
            ?.call(member, !member.isAudioOn, _uid2Member[event.operateUuid]);
      }
    }
  }

  @override
  void onMemberVideoMuteChanged(MemberToggleEvent event) {
    if (!isConnected) return;
    commonLogger.i('onMemberVideoMuteChanged: ${event.userUuid} ${event.isOn}');
    final member = updateMemberIfExists(event.userUuid, (member) {
      member.isVideoOn = event.isOn;
    });
    if (member != null) {
      for (var callback in _eventCallbacks.copy()) {
        callback.memberVideoMuteChanged
            ?.call(member, !member.isVideoOn, _uid2Member[event.operateUuid]);
      }
    }
  }

  @override
  void onMemberScreenShareStateChanged(MemberToggleEvent event) {
    if (!isConnected) return;
    commonLogger
        .i('onMemberScreenShareStateChanged: ${event.userUuid} ${event.isOn}');
    var member = _uid2Member[event.userUuid];
    if (event.isOn) {
      screenSharingMember = member;
    } else if (screenSharingMember?.uuid == event.userUuid) {
      member = screenSharingMember;
      screenSharingMember = null;
    }
    if (member != null) {
      member.isSharingScreen = event.isOn;
      for (var callback in _eventCallbacks.copy()) {
        callback.memberScreenShareStateChanged?.call(
            member, member.isSharingScreen, _uid2Member[event.operateUuid]);
      }
    }
  }

  @override
  void onReceiveChatroomMessages(RoomMessageEvent event) {
    if (!isConnected) return;
    commonLogger.i('onReceiveChatroomMessages');
    final messages = event.messages
        .map((e) => createMessage(e))
        .whereType<NERoomChatMessage>()
        .toList();
    if (messages.isNotEmpty) {
      for (var callback in _eventCallbacks.copy()) {
        callback.chatroomMessagesReceived?.call(messages);
      }
    }
  }

  @override
  void onChatroomMessageAttachmentProgress(CommonEvent event) {
    if (!isConnected) return;
    final uuid = event.sArg1 as String;
    final transferred = event.arg1 as int;
    final total = event.arg2 as int;
    assert(() {
      commonLogger
          .i('onChatroomMessageAttachmentProgress: $uuid $transferred $total');
      return true;
    }());
    for (var callback in _eventCallbacks.copy()) {
      callback.chatroomMessageAttachmentProgress?.call(
        event.sArg1 as String,
        event.arg1 as int,
        event.arg2 as int,
      );
    }
  }

  @override
  void onRoomEnd(String roomUuid, String reason) {
    if (!isConnected) return;
    commonLogger.i('onRoomEnd: $roomUuid $reason');
    for (var callback in _eventCallbacks.copy()) {
      callback.roomEnd?.call(stringToReason(reason));
    }
    Timer.run(dispose);
  }

  @override
  void onMemberWhiteboardShareStateChanged(MemberToggleEvent event) {
    if (!isConnected) return;
    commonLogger.i(
        'onMemberWhiteboardShareStateChanged: ${event.userUuid} ${event.isOn}');
    var member = _uid2Member[event.userUuid];
    if (event.isOn) {
      whiteboardSharingMember = member;
    } else if (whiteboardSharingMember?.uuid == event.userUuid) {
      member = whiteboardSharingMember;
      whiteboardSharingMember = null;
    }
    if (member != null) {
      member.isSharingWhiteboard = event.isOn;
      for (var callback in _eventCallbacks.copy()) {
        callback.memberWhiteboardShareStateChanged?.call(
            member, member.isSharingWhiteboard, _uid2Member[event.operateUuid]);
      }
    }
  }

  @override
  void onWhiteboardError(CommonEvent event) {
    if (!isConnected) return;
    commonLogger
        .i('onWhiteboardError: ${event.arg1} ${event.arg2} ${event.sArg1}');
    for (var callback in _eventCallbacks.copy()) {
      callback.whiteboardError?.call(event.arg1 ?? -1, event.sArg1);
    }
  }

  @override
  void onRoomLockStateChanged(RoomToggleEvent event) {
    if (!isConnected) return;
    _isRoomLocked = event.isOn;
    commonLogger.i('onRoomLockStateChanged: ${event.isOn} ');
    for (var callback in _eventCallbacks.copy()) {
      callback.roomLockStateChanged?.call(event.isOn);
    }
  }

  @override
  void onRoomLiveStateChanged(CommonEvent event) {
    if (!isConnected) return;
    commonLogger.i('onRoomLiveStateChanged: ${event.arg1} ');
    for (var callback in _eventCallbacks.copy()) {
      var state = event.arg1 ?? 0;
      callback.liveStateChanged?.call(NERoomLiveState.values[state]);
    }
  }

  @override
  void onRtcAudioVolumeIndication(RtcVolumeEvent event) {
    if (!isConnected || event.volumes == null) return;
    final volumes = event.volumes!
        .where((element) => element != null && element.userUuid.isNotEmpty)
        .map((e) => NEMemberVolumeInfo(e!.userUuid, e.volume))
        .toList();
    for (var callback in _eventCallbacks.copy()) {
      callback.rtcRemoteAudioVolumeIndication?.call(volumes, event.totalVolume);
    }
  }

  @override
  void onLocalAudioVolumeIndicationwithVad(LocalVolumeEvent event) {
    if (!isConnected) return;
    for (var callback in _eventCallbacks.copy()) {
      callback.rtcLocalAudioVolumeIndication
          ?.call(event.volume, event.enableVad);
    }
  }

  @override
  void onRtcAudioOutputDeviceChanged(CommonEvent event) {
    if (!isConnected) return;
    final selected = NEAudioOutputDeviceTypeExtension.fromType(event.arg1!);
    commonLogger.i('onRtcAudioOutputDeviceChanged: $selected');
    for (var callback in _eventCallbacks.copy()) {
      callback.rtcAudioOutputDeviceChanged?.call(selected);
    }
  }

  @override
  void onAudioMixingStateChanged(CommonEvent event) {
    if (!isConnected) return;
    commonLogger.i('onAudioMixingStateChanged: ${event.arg1}');
    for (var callback in _eventCallbacks.copy()) {
      callback.rtcAudioMixingStateChanged?.call(event.arg1!);
    }
  }

  @override
  void onRtcVirtualBackgroundSourceEnabled(
      RtcVirtualBackgroundSourceEvent event) {
    if (!isConnected) return;
    final enabled = event.enabled;
    final reason = event.reason;
    commonLogger.i(
        'onRtcVirtualBackgroundSourceEnabled: enabled:$enabled,reason:$reason');
    for (var callback in _eventCallbacks.copy()) {
      callback.rtcVirtualBackgroundSourceEnabled?.call(enabled, reason);
    }
  }

  Future<String?> onWhiteboardShowFileChooser(
      WhiteboardShowFileChooserTypeEvent event) async {
    String? result;
    if (!isConnected) return Future.value(result);
    final types = event.types;
    commonLogger.i('onWhiteboardShowFileChooser: types:$types');
    for (var callback in _eventCallbacks.copy()) {
      var path = await callback.whiteboardShowFileChooser?.call(types);
      if (path.isNotEmpty) {
        result = path;
      }
    }
    return Future.value(result);
  }

  @override
  void onRtcStats(String roomUuid, RoomStatsEvent event) {
    if (!isConnected) return;
    var stats = NERoomRtcStats(
        txBytes: event.txBytes,
        cpuAppUsage: event.cpuAppUsage,
        cpuTotalUsage: event.cpuTotalUsage,
        memoryAppUsageRatio: event.memoryAppUsageRatio,
        memoryTotalUsageRatio: event.memoryTotalUsageRatio,
        memoryAppUsageInKBytes: event.memoryAppUsageInKBytes,
        totalDuration: event.totalDuration,
        txAudioBytes: event.txAudioBytes,
        txVideoBytes: event.txVideoBytes,
        rxAudioBytes: event.rxAudioBytes,
        rxVideoBytes: event.rxVideoBytes,
        rxAudioKBitRate: event.rxAudioKBitRate,
        rxVideoKBitRate: event.rxVideoKBitRate,
        txAudioKBitRate: event.txAudioKBitRate,
        txVideoKBitRate: event.txVideoKBitRate,
        upRtt: event.upRtt,
        downRtt: event.downRtt,
        txAudioPacketLossRate: event.txAudioPacketLossRate,
        txVideoPacketLossRate: event.txVideoPacketLossRate,
        txAudioPacketLossSum: event.txAudioPacketLossSum,
        txVideoPacketLossSum: event.txVideoPacketLossSum,
        rxAudioPacketLossRate: event.rxAudioPacketLossRate,
        rxVideoPacketLossRate: event.rxVideoPacketLossRate,
        rxAudioPacketLossSum: event.rxAudioPacketLossSum,
        rxVideoPacketLossSum: event.rxVideoPacketLossSum,
        rxAudioJitter: event.rxAudioJitter,
        rxVideoJitter: event.rxVideoJitter);
    for (var callback in _statsCallbacks.copy()) {
      callback.rtcStats?.call(stats);
    }
  }

  @override
  void onNetworkQuality(String roomUuid, List<RoomNetworkQuality?> qualities) {
    if (!isConnected) return;
    // assert(() {
    //   debugPrint('onNetworkQuality@${DateTime.now()}:' +
    //       qualities
    //           .map((e) =>
    //               e != null ? '${e.userId}:${e.upStatus}:${e.downStatus}' : '')
    //           .join('#'));
    //   return true;
    // }());
    var array = qualities
        .whereType<RoomNetworkQuality>()
        .map((q) => NERoomRtcNetworkQualityInfo(q.userId,
            upStatus:
                NERoomRtcNetworkStatusTypeExtension.getTypeValue(q.upStatus),
            downStatus:
                NERoomRtcNetworkStatusTypeExtension.getTypeValue(q.downStatus)))
        .toList();

    for (var callback in _statsCallbacks.copy()) {
      callback.networkQuality?.call(array);
    }
  }
}

class _NERoomMemberImpl extends NERoomMember {
  @override
  final String uuid;

  @override
  String name;

  @override
  String? avatar;

  @override
  NERoomRole role;

  @override
  bool isAudioOn;

  @override
  bool isVideoOn;

  @override
  bool isInRtcChannel;

  @override
  bool isInChatroom;

  @override
  bool isSharingScreen;

  Map<String, String> props;

  @override
  Map<String, String> get properties => Map.from(props);

  @override
  bool isSharingWhiteboard;

  @override
  NEClientType clientType;

  _NERoomMemberImpl({
    required this.uuid,
    required this.name,
    required this.avatar,
    required this.role,
    required this.isAudioOn,
    required this.isVideoOn,
    required this.isInRtcChannel,
    required this.isInChatroom,
    required this.isSharingScreen,
    required this.isSharingWhiteboard,
    required this.props,
    required this.clientType,
  });

  @override
  String toString() {
    return '_NERoomMemberImpl{uuid: $uuid, name: $name, avatar: $avatar, role: $role, isAudioOn: $isAudioOn, isVideoOn: $isVideoOn, isInRtcChannel: $isInRtcChannel, isInChatroom: $isInChatroom, isSharingScreen: $isSharingScreen, props: $props, isSharingWhiteboard: $isSharingWhiteboard, clientType: $clientType}';
  }
}

class _NERoomTextMessageImpl extends NERoomChatTextMessage {
  @override
  final String messageUuid;

  @override
  final String? fromUserUuid;

  @override
  final String fromNick;

  @override
  final messageType = NERoomChatMessageType.kText;

  @override
  final String text;

  @override
  final int time;

  @override
  final List<String>? toUserUuidList;

  _NERoomTextMessageImpl({
    required this.messageUuid,
    required this.fromUserUuid,
    required this.fromNick,
    required this.text,
    required this.time,
    required this.toUserUuidList,
  });
}

class _NERoomCustomMessageImpl extends NERoomChatCustomMessage {
  @override
  final String messageUuid;

  @override
  final String? fromUserUuid;

  @override
  final String fromNick;

  @override
  final messageType = NERoomChatMessageType.kCustom;

  @override
  final String attachStr;

  @override
  final int time;

  @override
  final List<String>? toUserUuidList;

  _NERoomCustomMessageImpl({
    required this.messageUuid,
    required this.fromUserUuid,
    required this.fromNick,
    required this.attachStr,
    required this.time,
    required this.toUserUuidList,
  });
}

class _NERoomFileMessageImpl extends NERoomChatFileMessage {
  @override
  final String messageUuid;

  @override
  final String? fromUserUuid;

  @override
  final String fromNick;

  @override
  final messageType = NERoomChatMessageType.kFile;

  @override
  final int time;

  @override
  final List<String>? toUserUuidList;

  @override
  final String? displayName;

  @override
  final String? extension;

  @override
  final String? md5;

  @override
  final String url;

  @override
  final int size;

  @override
  final String? thumbPath;

  @override
  final String? path;

  _NERoomFileMessageImpl({
    required this.messageUuid,
    required this.fromUserUuid,
    required this.fromNick,
    required this.time,
    required this.toUserUuidList,
    required this.displayName,
    required this.extension,
    required this.md5,
    required this.url,
    required this.size,
    required this.thumbPath,
    required this.path,
  });
}

class _NERoomImageMessageImpl extends NERoomChatImageMessage {
  @override
  final String messageUuid;

  @override
  final String? fromUserUuid;

  @override
  final String fromNick;

  @override
  final messageType = NERoomChatMessageType.kImage;

  @override
  final int time;

  @override
  final List<String>? toUserUuidList;

  @override
  final String? displayName;

  @override
  final String? extension;

  @override
  final String? md5;

  @override
  final String url;

  @override
  final int size;

  @override
  final String? thumbPath;

  @override
  final String? path;

  @override
  final int width;

  @override
  final int height;

  _NERoomImageMessageImpl({
    required this.messageUuid,
    required this.fromUserUuid,
    required this.fromNick,
    required this.time,
    required this.toUserUuidList,
    required this.displayName,
    required this.extension,
    required this.md5,
    required this.url,
    required this.size,
    required this.thumbPath,
    required this.path,
    required this.width,
    required this.height,
  });
}

NERoomChatMessage? createMessage(RoomMessage? message) {
  if (message != null) {
    switch (message.type) {
      case 'text':
        {
          return _NERoomTextMessageImpl(
            messageUuid: message.uuid,
            fromUserUuid: message.fromAccount,
            fromNick: message.fromNick,
            text: message.content ?? '',
            time: message.time,
            toUserUuidList: message.toAccounts?.whereType<String>().toList(),
          );
        }
      case 'file':
        {
          final attachment = message.attachment as RoomMessageAttachment;
          return _NERoomFileMessageImpl(
              messageUuid: message.uuid,
              fromUserUuid: message.fromAccount,
              fromNick: message.fromNick,
              time: message.time,
              toUserUuidList: message.toAccounts?.whereType<String>().toList(),
              displayName: attachment.displayName,
              extension: attachment.extension,
              md5: attachment.md5,
              url: attachment.url,
              size: attachment.size,
              thumbPath: attachment.thumbPath,
              path: attachment.path);
        }
      case 'image':
        {
          final attachment = message.attachment as RoomMessageAttachment;
          return _NERoomImageMessageImpl(
            messageUuid: message.uuid,
            fromUserUuid: message.fromAccount,
            fromNick: message.fromNick,
            time: message.time,
            toUserUuidList: message.toAccounts?.whereType<String>().toList(),
            displayName: attachment.displayName,
            extension: attachment.extension,
            md5: attachment.md5,
            url: attachment.url,
            size: attachment.size,
            thumbPath: attachment.thumbPath,
            path: attachment.path,
            width: attachment.width ?? 0,
            height: attachment.height ?? 0,
          );
        }
      case 'custom':
        {
          return _NERoomCustomMessageImpl(
            messageUuid: message.uuid,
            fromUserUuid: message.fromAccount,
            fromNick: message.fromNick,
            attachStr: message.content ?? '',
            time: message.time,
            toUserUuidList: message.toAccounts?.whereType<String>().toList(),
          );
        }
    }
  }
  return null;
}

NERoomEndReason stringToReason(String reason) {
  switch (reason) {
    case 'leaveBySelf':
      return NERoomEndReason.kLeaveBySelf;
    case 'syncDataError':
      return NERoomEndReason.kSyncDataError;
    case 'kickOut':
      return NERoomEndReason.kKickOut;
    case 'selfKick':
      return NERoomEndReason.kKickBySelf;
    case 'closeByMember':
      return NERoomEndReason.kCloseByMember;
    case 'endOfLife':
      return NERoomEndReason.kEndOfLife;
    case 'allMemberOut':
      return NERoomEndReason.kAllMemberOut;
    case 'closeByBackend':
      return NERoomEndReason.kCloseByBackend;
    case 'loginStateError':
      return NERoomEndReason.kLoginStateError;
    case 'endOfRtc':
      return NERoomEndReason.kEndOfRtc;
  }
  return NERoomEndReason.kUnknown;
}
