// Copyright (c) 2022 NetEase, Inc. All rights reserved.
// Use of this source code is governed by a MIT license that can be
// found in the LICENSE file.

part of netease_roomkit;

class _NERoomSeatControllerImpl extends NERoomSeatController
    with SeatEventSink, _AloggerMixin {
  final _platform = RoomSeatControllerPlatform.instance;
  final String roomUuid;
  final _eventCallbacks = <NESeatEventCallback>{};

  List<NESeatEventCallback> get eventCallbacks => _eventCallbacks.toList();

  @override
  final bool isSupported;

  bool isConnected = true;

  _NERoomSeatControllerImpl(this.isSupported, this.roomUuid) {
    _platform.registerEventListener(roomUuid, this);
  }

  void dispose() {
    isConnected = false;
    _eventCallbacks.clear();
    _platform.unregisterEventListener(roomUuid, this);
  }

  @override
  Future<VoidResult> acceptSeatInvitation() {
    apiLogger.i('acceptSeatInvitation');
    return _platform.acceptSeatInvitation(roomUuid);
  }

  @override
  Future<VoidResult> approveSeatRequest(String user) {
    apiLogger.i('acceptSeatInvitation,user:$user');
    return _platform.approveSeatRequest(roomUuid, user);
  }

  @override
  Future<VoidResult> cancelSeatInvitation(String user) {
    apiLogger.i('cancelSeatInvitation,user:$user');
    return _platform.cancelSeatInvitation(roomUuid, user);
  }

  @override
  Future<VoidResult> cancelSeatRequest() {
    apiLogger.i('cancelSeatRequest');
    return _platform.cancelSeatRequest(roomUuid);
  }

  @override
  Future<VoidResult> closeSeats(List<int> seatIndices) {
    apiLogger.i('closeSeats,seatIndices:$seatIndices');
    return _platform.closeSeats(roomUuid, seatIndices);
  }

  @override
  Future<VoidResult> kickSeat(String user) {
    apiLogger.i('kickSeat,user:$user');
    return _platform.kickSeat(roomUuid, user);
  }

  @override
  Future<VoidResult> leaveSeat() {
    apiLogger.i('leaveSeat');
    return _platform.leaveSeat(roomUuid);
  }

  @override
  Future<VoidResult> openSeats(List<int> seatIndices) {
    apiLogger.i('openSeats,seatIndices:$seatIndices');
    return _platform.openSeats(roomUuid, seatIndices);
  }

  @override
  Future<VoidResult> rejectSeatInvitation() {
    apiLogger.i('rejectSeatInvitation');
    return _platform.rejectSeatInvitation(roomUuid);
  }

  @override
  Future<VoidResult> rejectSeatRequest(String user) {
    apiLogger.i('rejectSeatRequest,user:$user');
    return _platform.rejectSeatRequest(roomUuid, user);
  }

  @override
  Future<VoidResult> sendSeatInvitation(int seatIndex, String user) {
    apiLogger.i('sendSeatInvitation,seatIndex:$seatIndex,user:$user');
    return _platform.sendSeatInvitation(roomUuid, seatIndex, user);
  }

  @override
  Future<VoidResult> submitSeatRequest(int seatIndex, bool exclusive) {
    apiLogger.i('submitSeatRequest,seatIndex:$seatIndex,exclusive:$exclusive');
    return _platform.submitSeatRequest(roomUuid, seatIndex, exclusive);
  }

  @override
  Future<NEResult<NESeatInfo>> getSeatInfo() {
    apiLogger.i('getSeatInfo');
    return _platform.getSeatInfo(roomUuid);
  }

  @override
  Future<NEResult<List<NESeatInvitationItem>>> getSeatInvitationList() {
    apiLogger.i('getSeatInvitationList');
    return _platform.getSeatInvitationList(roomUuid);
  }

  @override
  Future<NEResult<List<NESeatRequestItem>>> getSeatRequestList() {
    apiLogger.i('getSeatRequestList');
    return _platform.getSeatRequestList(roomUuid);
  }

  @override
  void addEventCallback(NESeatEventCallback callback) {
    _eventCallbacks.add(callback);
  }

  @override
  void removeEventCallback(NESeatEventCallback callback) {
    _eventCallbacks.remove(callback);
  }

  @override
  void onSeatManagerAdded(List<String?> managers) {
    if (!isConnected) return;
    commonLogger.i('onSeatManagerAdded,managers:$managers');
    eventCallbacks.copy().forEach((element) {
      element.seatManagerAddedCallback?.call(managers);
    });
  }

  @override
  void onSeatManagerRemoved(List<String?> managers) {
    if (!isConnected) return;
    commonLogger.i('onSeatManagerRemoved,managers:$managers');
    eventCallbacks.copy().forEach((element) {
      element.seatManagerRemovedCallback?.call(managers);
    });
  }

  @override
  void onSeatRequestSubmitted(int seatIndex, String user) {
    if (!isConnected) return;
    commonLogger.i('onSeatRequestSubmitted,seatIndex:$seatIndex,user:$user');
    eventCallbacks.copy().forEach((element) {
      element.seatRequestSubmittedCallback?.call(seatIndex, user);
    });
  }

  @override
  void onSeatRequestCancelled(int seatIndex, String user) {
    if (!isConnected) return;
    commonLogger.i('onSeatRequestCancelled,seatIndex:$seatIndex,user:$user');
    eventCallbacks.copy().forEach((element) {
      element.seatRequestCancelledCallback?.call(seatIndex, user);
    });
  }

  @override
  void onSeatRequestApproved(
      int seatIndex, String user, String operateBy, bool isAutoAgree) {
    if (!isConnected) return;
    commonLogger.i(
        'onSeatRequestApproved,seatIndex:$seatIndex,user:$user,operateBy:$operateBy,isAutoAgree:$isAutoAgree');
    eventCallbacks.copy().forEach((element) {
      element.seatRequestApprovedCallback
          ?.call(seatIndex, user, operateBy, isAutoAgree);
    });
  }

  @override
  void onSeatRequestRejected(int seatIndex, String user, String operateBy) {
    if (!isConnected) return;
    commonLogger.i(
        'onSeatRequestRejected,seatIndex:$seatIndex,user:$user,operateBy:$operateBy');
    eventCallbacks.copy().forEach((element) {
      element.seatRequestRejectedCallback?.call(seatIndex, user, operateBy);
    });
  }

  @override
  void onSeatInvitationReceived(int seatIndex, String user, String operateBy) {
    if (!isConnected) return;
    commonLogger.i(
        'onSeatInvitationReceived,seatIndex:$seatIndex,user:$user,operateBy:$operateBy');
    eventCallbacks.copy().forEach((element) {
      element.seatInvitationReceivedCallback?.call(seatIndex, user, operateBy);
    });
  }

  @override
  void onSeatInvitationCancelled(int seatIndex, String user, String operateBy) {
    if (!isConnected) return;
    commonLogger.i(
        'onSeatInvitationCancelled,seatIndex:$seatIndex,user:$user,operateBy:$operateBy');
    eventCallbacks.copy().forEach((element) {
      element.seatInvitationCancelledCallback?.call(seatIndex, user, operateBy);
    });
  }

  @override
  void onSeatInvitationAccepted(int seatIndex, String user, bool isAutoAgree) {
    if (!isConnected) return;
    commonLogger.i(
        'onSeatInvitationAccepted,seatIndex:$seatIndex,user:$user,isAutoAgree:$isAutoAgree');
    eventCallbacks.copy().forEach((element) {
      element.seatInvitationAcceptedCallback
          ?.call(seatIndex, user, isAutoAgree);
    });
  }

  @override
  void onSeatInvitationRejected(int seatIndex, String user) {
    if (!isConnected) return;
    commonLogger.i('onSeatInvitationRejected,seatIndex:$seatIndex,user:$user');
    eventCallbacks.copy().forEach((element) {
      element.seatInvitationRejectedCallback?.call(seatIndex, user);
    });
  }

  @override
  void onSeatLeave(int seatIndex, String user) {
    if (!isConnected) return;
    commonLogger.i('onSeatLeave,seatIndex:$seatIndex,user:$user');
    eventCallbacks.copy().forEach((element) {
      element.seatLeaveCallback?.call(seatIndex, user);
    });
  }

  @override
  void onSeatKicked(int seatIndex, String user, String operateBy) {
    if (!isConnected) return;
    commonLogger
        .i('onSeatKicked,seatIndex:$seatIndex,user:$user,operateBy:$operateBy');
    eventCallbacks.copy().forEach((element) {
      element.seatKickedCallback?.call(seatIndex, user, operateBy);
    });
  }

  @override
  void onSeatListChanged(List<SeatItem?> seatItems) {
    if (!isConnected) return;
    commonLogger.i('onSeatListChanged,seatItems:$seatItems');
    eventCallbacks.copy().forEach((element) {
      List<NESeatItem> list = [];
      seatItems.forEach((element) {
        list.add(NESeatItem(
            element?.index,
            element?.status,
            element?.user,
            element?.userName,
            element?.icon,
            element?.onSeatType,
            element?.updated));
      });
      element.seatListChangedCallback?.call(list);
    });
  }
}
