// Copyright (c) 2022 NetEase, Inc. All rights reserved.
// Use of this source code is governed by a MIT license that can be
// found in the LICENSE file.

part of netease_roomkit;

class _NERtcVideoValue {
  const _NERtcVideoValue({
    this.width = 0,
    this.height = 0,
    this.rotation = 0,
  });

  final int width;
  final int height;
  final int rotation;

  const _NERtcVideoValue.uninitialized()
      : this(
          width: 0,
          height: 0,
          rotation: 0,
        );

  _NERtcVideoValue copyWith({
    int? width,
    int? height,
    int? rotation,
  }) {
    return _NERtcVideoValue(
      width: width ?? this.width,
      height: height ?? this.height,
      rotation: rotation ?? this.rotation,
    );
  }

  double get aspectRatio {
    return width * height == 0
        ? 1.0
        : (rotation == 90 || rotation == 270)
            ? height / width
            : width / height;
  }

  @override
  String toString() {
    return 'NERtcVideoValue{width: $width, height: $height, rotation: $rotation}';
  }
}

class _RenderAttachInfo {
  final String userUuid;
  final bool isSubStream;

  _RenderAttachInfo(this.userUuid, this.isSubStream);

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
      other is _RenderAttachInfo &&
          runtimeType == other.runtimeType &&
          userUuid == other.userUuid &&
          isSubStream == other.isSubStream;

  @override
  int get hashCode => userUuid.hashCode ^ isSubStream.hashCode;
}

class _NERtcVideoRendererImpl extends NERtcVideoRenderer with _AloggerMixin {
  static final _renderAttachInfo = <_RenderAttachInfo, NERtcVideoRenderer>{};

  final _platform = RoomRtcRenderPlatform.instance;
  final String roomUuid;
  int? _textureId;
  bool _attached = false;
  bool _disposed = false;

  _NERtcVideoRendererImpl(this.roomUuid,
      {NERtcVideoRendererEventListener? listener})
      : super(rendererEventLister: listener);

  bool _mirror = false;
  StreamSubscription<dynamic>? _rendererEventSubscription;

  bool _local = false;
  bool _subStream = false;

  String? _userUuid;

  @override
  Future<bool> initialize() async {
    if (_disposed) return false;
    if (_textureId != null) return true;
    final reply = await _platform.createVideoRenderer(roomUuid);
    apiLogger.e('_NERtcVideoRendererImpl initialize:${reply.code}');
    if (reply.code < 0) {
      return false;
    }
    _textureId = reply.code;
    _rendererEventSubscription =
        EventChannel('NERoomKitFlutterVideoRenderer/Texture$_textureId')
            .receiveBroadcastStream()
            .listen(_onRendererEvent);
    return true;
  }

  void _onRendererEvent(dynamic event) {
    final Map<dynamic, dynamic> map = event;
    switch (map['event']) {
      case 'onFrameResolutionChanged':
        value = value.copyWith(
            width: map['width'],
            height: map['height'],
            rotation: map['rotation']);
        rendererEventLister?.onFrameResolutionChanged(
            value.width, value.height, value.rotation);
        if (kDebugMode) {
          print('onFrameResolutionChanged: ${value.toString()}');
        }
        break;
      case 'onFirstFrameRendered':
        rendererEventLister?.onFirstFrameRendered();
        if (kDebugMode) {
          print('onFirstFrameRendered');
        }
        break;
    }
  }

  @override
  bool get canRender =>
      value.width * value.height != 0 && (_userUuid != null || _local);

  @override
  int? get textureId => _textureId;

  @override
  Future<int> attachToLocalVideo() async {
    return _attach(true, false, null, _textureId);
  }

  @override
  Future<int> attachToLocalSubStreamVideo() async {
    return _attach(true, true, null, _textureId);
  }

  @override
  Future<int> attachToRemoteVideo(String userUuid) {
    return _attach(false, false, userUuid, _textureId);
  }

  @override
  Future<int> attachToRemoteSubStreamVideo(String userUuid) {
    return _attach(false, true, userUuid, _textureId);
  }

  @override
  Future<int> detach() async {
    final id = _textureId;
    if (!_attached || id == null) return -1;
    VoidResult reply;
    final attachedUuid = _userUuid ?? '0';
    if (!hasAttachInfo(attachedUuid, _subStream, this)) {
      reply = NEResult.success();
    } else {
      if (_local) {
        reply = await _platform.detachLocalVideoRenderer(_subStream, id);
      } else {
        if (_userUuid == null) return -1;
        reply = await _platform.detachRemoteVideoRenderer(
            _userUuid!, _subStream, id);
      }
    }
    assert(() {
      print(
          '_NERtcVideoRendererImpl: detach $id from $_userUuid = ${reply.code}');
      return true;
    }());
    if (reply.code == 0) {
      _removeAttachInfo(attachedUuid, _subStream, this);
      _attached = false;
      _local = false;
      _userUuid = null;
      _subStream = false;
      value = const _NERtcVideoValue.uninitialized();
    }
    return reply.code;
  }

  Future<int> _attach(
      bool local, bool subStream, String? userUuid, int? textureId) async {
    if (textureId == null) return -1;
    VoidResult reply;
    if (local) {
      reply = subStream
          ? await _platform.setupLocalSubStreamVideoRenderer(textureId)
          : await _platform.setupLocalVideoRenderer(textureId);
    } else {
      reply = subStream
          ? await _platform.setupRemoteSubStreamVideoRenderer(
              userUuid!, textureId)
          : await _platform.setupRemoteVideoRenderer(userUuid!, textureId);
    }
    assert(() {
      print(
          '_NERtcVideoRendererImpl: attach $textureId to $_userUuid = ${reply.code}');
      return true;
    }());
    if (reply.code == 0) {
      _saveAttachInfo(userUuid ?? '0', subStream, this);
      _attached = true;
      _local = local;
      _subStream = subStream;
      _userUuid = userUuid;
      if (local) {
        _platform.setMirror(textureId, _mirror);
      }
    }
    return reply.code;
  }

  @override
  Future<int> setMirror(bool mirror) async {
    if (_textureId == null) {
      _mirror = mirror;
      return 0;
    }
    final result = await _platform.setMirror(_textureId!, mirror);
    if (result.isSuccess()) {
      _mirror = mirror;
    }
    return result.code;
  }

  @override
  bool isMirror() => _mirror;

  @override
  Future<int> dispose() async {
    if (_disposed) return 0;
    _disposed = true;
    await _rendererEventSubscription?.cancel();
    await detach();
    if (_textureId != null) {
      await _platform.disposeVideoRenderer(_textureId!);
      _textureId = null;
    }
    return super.dispose();
  }

  _saveAttachInfo(String uuid, bool isSubStream, NERtcVideoRenderer renderer) {
    final attachInfo = _RenderAttachInfo(uuid, isSubStream);
    _renderAttachInfo[attachInfo].guard((value) {
      if (value != renderer) {
        commonLogger.w(
            'Render attachment is overwritten: $uuid, $isSubStream, $value -> $renderer');
      }
    });
    _renderAttachInfo[attachInfo] = renderer;
  }

  void _removeAttachInfo(
      String uuid, bool isSubStream, NERtcVideoRenderer renderer) {
    _renderAttachInfo.removeWhere((key, value) {
      return key == _RenderAttachInfo(uuid, isSubStream) && value == renderer;
    });
  }

  bool hasAttachInfo(
      String uuid, bool isSubStream, NERtcVideoRenderer renderer) {
    return _renderAttachInfo[_RenderAttachInfo(uuid, isSubStream)] == renderer;
  }
}
